!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2014,2017 Yoshihide Yoshimoto
!
!
#include "fftgrid.h"
!
program atmcdist
  use iounit
  use bsplpot
  use util
  use xcf_types
  use mpi
  use replica
  use chkinput
  use subr_rddata
  use subr_kpoint
  use meshgrid
  use chain_atom_cg
  use converter
  use atomicchrg
  implicit none
#include "config90.h"
  !
  integer pg(3,nsymq),rg(3,3,nsymq),srg(nsymq)
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  real(kind=8):: bb0(3,3),bb(3,3),ax,bx,va,vb
  integer:: ng,nnp,nsv,ismode
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: nsi,nkd
  integer:: nrx,nry,nrz,nhx,nhy,nhz,nrxyz,ntt &
       ,nrcell,nspin,ngx,ngy,ngz,ngxyz &
       ,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz
  integer:: ndens
  real(kind=8):: qf,rf,qm,qc,qwf,qbtp(3)
  real(kind=8):: qbsp
  integer:: nbsf
  integer:: nkr,nkfxa,nkc,nng,nband,nw,nb1,nb2
  !
  integer:: iopt(nopt)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer nsk,nks,nk,nkmesh
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  type (irdkpt_struct),pointer:: irdkpt(:)
  !
  real(kind=8):: clk
  integer:: mssg,ierr
  real(kind=8):: ekbt
  real(kind=8):: fwqcom
  integer:: xctp
  real(kind=8),pointer:: zoin(:),znin(:)
  !
  integer:: nproc,myrank,comm
  type (replica_struct):: repls
  !
  real(kind=8),pointer:: udirchr(:,:)
  !
  type (meshgrid_struct):: mgs
  !
  real(kind=8),allocatable:: aff(:)
  real(kind=8),allocatable:: ztmp(:,:,:,:)
  real(kind=8),allocatable:: rho(:,:,:,:),arho(:,:,:)
  !
  character(len=256):: fname
  integer:: ikd,nfrhi

  character(len=256):: argv(1)
  integer:: argc
  !
  call MPI_INIT(ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD,nproc,ierr)
  if (nproc.eq.1) then
     call MPI_COMM_DUP(MPI_COMM_WORLD,comm,ierr)
     call MPI_COMM_DUP(MPI_COMM_WORLD,repls%comt,ierr)
     repls%ifreplm = .false.
  else
     call setup_replica(NFINPUT,MPI_COMM_WORLD,comm,repls)
  end if
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)
  !
  ! atmcdist should be performed by single MPI proc for each replica.
  !
  if (myrank.ne.0) then
     call MPI_FINALIZE(ierr)
     stop
  end if

  clk=xetime(0)
  write(6,*)'ATMCDIST'
  mssg=1
  !
  call rdinidata(ax,bx,aa0,bb0,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz,ntt,ngx,ngy,ngz,ngxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,NFINPUT,mssg &
       ,qbsp,nbsf,fwqcom,xctp,zoin,znin)

  allocate(udirchr(3,nsi))
  udirchr = 0.0d0
  call chkall(ng,rg,pg,srg,nnp,aa,nsi,kd,asi,ndens,udirchr,MPI_COMM_SELF)
  deallocate(udirchr)

  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)

  allocate(bspots(nkd),stat=ierr)
  call chkalloc('atmcdist','bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     write(6,*)
     call tellbspot(nkd,bspots)
  end if

  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)
  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)
  vb = cellvol(bb)

  allocate(aff(nkfxa),stat=ierr)
  call chkalloc('atmcdist','aff',ierr)
  call kf2aff(bb,nkfxa,mgs%kf,aff)

  allocate(FFT_A3D_DEFSZ(ztmp,nrx*2,nry*2,nrz*2),stat=ierr)
  call chkalloc('atmcdist','ztmp',ierr)
  call mkionchrg(bspotfc,nkd,bspots,nkfxa,mgs%kf,aff,va, &
       nsi,asi,kd,nrx*2,nry*2,nrz*2,ztmp)

  allocate(arho(nrx*2,nry*2,nrz*2),stat=ierr)
  call chkalloc('atmcdist','arho',ierr)
  
  arho = 0.0d0
  argc = command_argument_count()
  if (argc.eq.1) then
     call get_command_argument(1, argv(1))
     if (trim(argv(1)) .eq. '-a') then
        allocate(rho(nrx*2,nry*2,nrz*2,nspin*ndens),stat=ierr)
        call xrewind(NFCHR)
        read(NFCHR) ! magic 'TAPP:rho'
        read(NFCHR) ! aa, nspin
        read(NFCHR) ! 2*nrx, 2*nry, 2*nrz
        read(NFCHR) rho
        close(NFCHR)

        write(6,*) 
        write(6,*) 'append the valence charge to the atomic charge'
        arho = rho(:,:,:,1)
     end if
  end if

  arho(:,:,:) = arho(:,:,:) + FFT_A3D_RL(ztmp,:,:,:)

  call xrewind(NFATMCHRD)
  write(NFATMCHRD) 'TAPP:rho'
  write(NFATMCHRD) aa, 1
  write(NFATMCHRD) 2*nrx, 2*nry, 2*nrz
  write(NFATMCHRD) arho
  close(NFATMCHRD)

  deallocate(arho,ztmp)
  call dealloc_mesh(mgs)
  deallocate(kd,asi,mmm,zoin,znin)

  clk=xetime(1)
  write(6,*)
  write(6,*) 'TOTAL TIME=',clk
  call MPI_FINALIZE(ierr)
  stop
end program atmcdist
