!
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
module chain_atom_cg
contains
  subroutine store_atom_cg_data(icy,dif,etotal,setotm,osetotm, &
       asi,cdm,dform,dfold,h,hcdm)
    use iounit
    implicit none
    integer:: icy
    real(kind=8):: dif, etotal
    real(kind=8),dimension(:,:):: asi, dform, dfold, h
    real(kind=8):: setotm(6),osetotm(6),cdm(6),hcdm(6)
    !
    call xrewind(NFACGD)
    write(NFACGD) icy,dif,etotal,asi,dform,dfold,h
    write(NFACGD) cdm,hcdm,setotm,osetotm
    return
  end subroutine store_atom_cg_data

  subroutine load_atom_cg_data(icy,dif,etotal,setotm,osetotm, &
       asi,cdm,dform,dfold,h,hcdm,myrank,comm)
    use iounit
    use mpi
    implicit none
    integer:: icy
    real(kind=8):: dif, etotal
    real(kind=8),dimension(:,:):: asi, dform, dfold, h
    real(kind=8):: setotm(6),osetotm(6),cdm(6),hcdm(6)
    integer,intent(in):: myrank,comm

    integer:: n,ierr
    !
    if (myrank.eq.0) then
       call xrewind(NFACGD)
       read(NFACGD,iostat=ierr) icy,dif,etotal,asi,dform,dfold,h
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening atom cg data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       read(NFACGD) cdm,hcdm,setotm,osetotm
    end if
    call MPI_BCAST(icy,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(dif,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(etotal,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    n = size(asi,1)*size(asi,2)
    call MPI_BCAST(asi,n,MPI_DOUBLE_PRECISION,0,comm,ierr)
    n = size(dform,1)*size(dform,2)
    call MPI_BCAST(dform,n,MPI_DOUBLE_PRECISION,0,comm,ierr)
    n = size(dfold,1)*size(dfold,2)
    call MPI_BCAST(dfold,n,MPI_DOUBLE_PRECISION,0,comm,ierr)
    n = size(h,1)*size(h,2)
    call MPI_BCAST(h,n,MPI_DOUBLE_PRECISION,0,comm,ierr)

    call MPI_BCAST(cdm,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(hcdm,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(setotm,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(osetotm,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    
    return
  end subroutine load_atom_cg_data

  subroutine load_atom_pos_data1(nsi,asi,cdm,ierr)
    use iounit
    use mpi
    implicit none
    integer,intent(in):: nsi
    integer:: ierr
    real(kind=8):: asi(3,nsi),cdm(6)

    integer:: icy
    real(kind=8):: dif, etotal
    real(kind=8),dimension(3,nsi):: dform, dfold, h
    real(kind=8):: setotm(6),osetotm(6),hcdm(6)
    !
    call xrewind(NFACGD)
    read(NFACGD,iostat=ierr) icy,dif,etotal,asi,dform,dfold,h
    if (ierr.ne.0) then
       return
    end if
    read(NFACGD) cdm,hcdm,setotm,osetotm
    
    return
  end subroutine load_atom_pos_data1

  subroutine load_atom_pos_data(nsi,asi,cdm,myrank,comm)
    use iounit
    use mpi
    implicit none
    integer,intent(in):: nsi,myrank,comm
    real(kind=8):: asi(3,nsi),cdm(6)

    integer:: n,ierr
    !
    if (myrank.eq.0) then
       call load_atom_pos_data1(nsi,asi,cdm,ierr)
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening atom cg data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
    end if
    n = size(asi,1)*size(asi,2)
    call MPI_BCAST(asi,n,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(cdm,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    
    return
  end subroutine load_atom_pos_data
end module chain_atom_cg

module chain_linmin
contains
  subroutine store_linmin_data(ifar, itlin, alph1, alph2, etotal, okstep, &
       all, al, ar, arr)
    use iounit
    implicit none
    integer:: ifar, itlin
    real(kind=8):: alph1, alph2, etotal, okstep(2)
    real(kind=8),dimension(:):: all, al, ar, arr
    !
    call xrewind(NFLMD)
    write(NFLMD) ifar, itlin, alph1, alph2, etotal, okstep, all, al, ar, arr
    return
  end subroutine store_linmin_data

  subroutine load_linmin_data(ifar, itlin, alph1, alph2, etotal, okstep, &
       all, al, ar, arr, myrank, comm)
    use iounit
    use mpi
    implicit none
    integer:: ifar, itlin
    real(kind=8):: alph1, alph2, etotal, okstep(2)
    real(kind=8),dimension(:):: all, al, ar, arr
    integer,intent(in):: myrank,comm
    integer:: ierr
    !
    if (myrank.eq.0) then
       call xrewind(NFLMD)
       read(NFLMD,iostat=ierr) ifar, itlin, alph1, alph2, etotal, okstep, all, al, ar, arr
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening linear minimization data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
    end if
    call MPI_BCAST(ifar,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(itlin,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(alph1,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(alph2,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(etotal,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(okstep,2,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(all,size(all,1),MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(al,size(al,1),MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(ar,size(ar,1),MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(arr,size(arr,1),MPI_DOUBLE_PRECISION,0,comm,ierr)
    
    return
  end subroutine load_linmin_data
end module chain_linmin

module chain_scf
contains
  subroutine store_scf_data(ndens, iter, etsave, dd, xs, xs2, r2hst, &
       myrank, comm)
    use iounit
    use zextrapolate
    implicit none
    integer:: ndens, iter
    real(kind=8):: etsave, r2hst(:)
    complex(kind=8):: dd(:,:,:,:)
    type (zxtrapstruct) :: xs, xs2
    integer:: niter
    integer,intent(in):: myrank, comm
    !
    if (myrank.eq.0) then
       call xrewind(NFSCFD)
       write(NFSCFD) iter, etsave, dd
    end if

    call zxtrap_store(NFSCFD,xs,comm)
    if (ndens.eq.4) then
       call zxtrap_store(NFSCFD,xs2,comm)
    end if

    niter = size(r2hst,1)
    if (myrank.eq.0) then
       write(NFSCFD) niter
       write(NFSCFD) r2hst
    end if
    return
  end subroutine store_scf_data

  logical function probe_scf_data()
    use iounit
    implicit none
    integer:: ierr, idum

    call xrewind(NFSCFD)
    read(NFSCFD,iostat=ierr) idum
    if (ierr.eq.0) then
       probe_scf_data = .true.
    else
       probe_scf_data = .false.
    end if
    return
  end function probe_scf_data
  
  subroutine load_scf_data(ndens, iter, etsave, dd, xs, xs2, r2hst, comm)
    use iounit
    use zextrapolate
    use mpi
    implicit none
    integer:: ndens, iter
    real(kind=8):: etsave, r2hst(:)
    complex(kind=8):: dd(:,:,:,:)
    type (zxtrapstruct) :: xs, xs2
    integer,intent(in):: comm
    !
    integer:: niter, ios
    integer:: myrank, ierr
    !
    call MPI_COMM_RANK(comm,myrank,ierr)

    if (myrank.eq.0) then
       call xrewind(NFSCFD)
       read(NFSCFD,iostat=ierr) iter, etsave, dd
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening scf extrapolation data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
    end if
    call MPI_BCAST(iter,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(etsave,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(dd,size(dd),MPI_DOUBLE_COMPLEX,0,comm,ierr)

    call zxtrap_load(NFSCFD,xs,comm)
    if (ndens.eq.4) then
       call zxtrap_load(NFSCFD,xs2,comm)
    end if

    if (myrank.eq.0) then
       read(NFSCFD,iostat=ios) niter
       if (ios.ne.0) then
          write(6,'(1x,a)') 'no r2hst data. r2hst -> huge'
          r2hst = huge(1.0d0)
       else if (niter.ne.size(r2hst,1)) then
          write(6,'(1x,a)') 'size of r2hst is modified.'
          write(6,'(1x,a,i4,1x,i4)') 'niter, size(r2hst,1) =', &
               niter, size(r2hst,1)
          write(6,'(1x,a)') 'r2hst -> huge.'
          r2hst = huge(1.0d0)
       else
          read(NFSCFD) r2hst
       end if
    end if
    call MPI_BCAST(r2hst,size(r2hst,1),MPI_DOUBLE_PRECISION,0,comm,ierr)
    return
  end subroutine load_scf_data
end module chain_scf

module chain_wfn
  private iksnst
  private load_wfn_master3, load_wfn_slave3
contains
  subroutine store_wfn(aa,ncomp,nspin,nw,ifinvsym, &
       dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
       nband,ee,nkd,bspots,myrank,nproc,comm)
    use iounit
    use mpi
    use types
    use meshgrid
    use util
    use bsplpot
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in)::  ncomp,nspin,nw
    type (meshgrid_struct):: mgs
    logical,intent(in):: ifinvsym
    type (wfn_for_k):: dp(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: laprj,nsi,kd(nsi)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    integer,intent(in):: nband
    real(kind=8),intent(in):: ee(nband,mgs%nks,nspin)
    integer:: nkd
    type (bspot_for_atm),intent(in):: bspots(nkd)
    integer:: myrank,nproc,comm
    !     
    real(kind=8):: bb(3,3)
    integer:: ik,is,i,j,ib,nkm
    complex(kind=8),allocatable:: p(:,:)
    integer:: ig,bkm,ierr
    integer:: it
    integer:: ip

    integer:: nkr,nng,nks
!soption contarea(nkmary,kg,g_count,g_displ)
    integer,pointer:: nkmary(:),kg(:,:,:)
    integer,pointer:: g_count(:,:),g_displ(:,:)
    real(kind=8):: w
    integer:: mpi_typ_wfn

    call recp_uvec(aa,bb)
    nks = mgs%nks
    nkr = mgs%nkr
    nng = mgs%nng
    nkmary => mgs%nkm
    kg => mgs%kg
    g_count => mgs%g_count
    g_displ => mgs%g_displ
    !
    if (myrank.eq.0) then
       call xrewind(NFWFO)
       if (ncomp.eq.1 .and. ifinvsym) then
          write(NFWFO) 'TAPP:wfi', 3
       else
          write(NFWFO) 'TAPP:wfn', 3
       end if
       write(NFWFO) aa, bb
       write(NFWFO) nspin, nks, nw, mgs%ng, mgs%nnp, ncomp
       write(NFWFO) mgs%rg, mgs%pg, mgs%srg
       write(NFWFO) (mgs%irdkpt(ik)%sk,ik=1,nks)
       write(NFWFO) nkmary
    end if

    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    do ik=1,nks
       nkm=nkmary(ik)
       if (myrank.eq.0) then
          write(NFWFO) ((kg(i,ig,ik),i=1,3),ig=1,nkm)
       end if
       allocate(p(ncomp,nkm), stat=ierr)
       call chkalloc('store_wfn','p', ierr)
       bkm = g_displ(myrank,ik)+1
       do is=1,nspin
          do ib=1,nw
             if (ib.lt.nb1) then
                if (nspin.eq.1) then
                   w = mgs%irdkpt(ik)%ngir*2.0d0
                else
                   w = mgs%irdkpt(ik)%ngir
                end if
             else if (ib.le.nb2) then
                w = wgtk(ib,ik,is)
             end if

             call MPI_GATHERV(dp(ik,is)%wf(:,:,ib),g_count(myrank,ik), &
                  mpi_typ_wfn,p,g_count(:,ik),g_displ(:,ik), &
                  mpi_typ_wfn,0,comm,ierr)
             if (ncomp.eq.1 .and. ifinvsym) then
                if (myrank.eq.0) then
                   write(NFWFO) (dble(p(1,i)),i=1,nkm),w,ee(ib,ik,is)
                end if
             else
                if (myrank.eq.0) then
                   write(NFWFO) ((p(i,j),i=1,ncomp),j=1,nkm),w,ee(ib,ik,is)
                end if
             end if
             call MPI_BARRIER(comm,ierr)
          end do
       end do
       deallocate(p)
    end do

    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)

    if (myrank .eq. 0) then
       do is=1,nspin
          do ik=1,mgs%nks
             write(NFWFO) (((pby(ip,it,ib,ik,is) &
                  ,ip=1,bspots(kd(it))%nprj) &
                  ,it=1,nsi),ib=1,nw)
          end do
       end do
       close(NFWFO)
     end if
    return
  end subroutine store_wfn
  !
  subroutine load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee,dp, &
       mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
       nband,ee,nkd,bspots,mssg,myrank,nproc,comm)
    use iounit
    use util
    use types
    use meshgrid
    use mpi
    use bsplpot
    implicit none
    integer,intent(in):: ncomp,ndens
    integer,intent(in):: nspin,nw
    real(kind=8),intent(in):: aa(3,3)
    logical:: ifrgpbyee
    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dp(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: laprj,nsi,kd(nsi)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    integer,intent(in):: nband
    real(kind=8):: ee(nband,mgs%nks,nspin)
    integer:: nkd
    type (bspot_for_atm),intent(in):: bspots(nkd)
    integer:: mssg,myrank,nproc,comm
    real(kind=8):: bb(3,3)

    character(len=8):: sign
    integer:: ierr,ver
    integer:: ip,it,ib,ik,is
    !
    wgtk = 0.0d0
    call recp_uvec(aa,bb)

    if (mssg.gt.-1) write(6,*) 'loading wave function...'
    if (myrank.eq.0) then
       call xrewind(NFWFI)
       read(NFWFI,iostat=ierr) sign, ver
       if (ierr.ne.0) then
          ver = 1
          call xrewind(NFWFI)
          read(NFWFI,iostat=ierr) sign
          if (ierr.ne.0) then
             write(6,'(1x,a)') 'error in opening wave function'
             call MPI_ABORT(comm,1,ierr)
             stop
          end if
       end if
    end if

    call MPI_BCAST(sign,len(sign),MPI_CHARACTER,0,comm,ierr)
    call MPI_BCAST(ver,1,MPI_INTEGER,0,comm,ierr)

    if (ver.eq.1) then
       write(6,*) 'version 1 format is not supported'
       call MPI_ABORT(comm,1,ierr)
       stop
    else if (ver.eq.2) then
       write(6,*) 'version 2 format is not supported'
       call MPI_ABORT(comm,1,ierr)
       stop
    else if (ver.eq.3) then
       if (myrank.eq.0) then
          call load_wfn_master3(NFWFI,sign,ncomp,nspin,ndens,nw,bb,ifrgpbyee,dp, &
               mgs,nb1,nb2,wgtk,nband,ee,mssg,myrank,nproc,comm)
       else
          call load_wfn_slave3(ncomp,nspin,nw,ifrgpbyee,dp, &
               mgs,nb1,nb2,wgtk,nband,ee,myrank,nproc,comm)
       end if
    else
       if (mssg.gt.-1) write(6,'(1x,a)') 'invalid wfn format version'
       call MPI_ABORT(comm,1,ierr)
       stop
    end if

    pby = (0.0d0, 0.0d0)
    if (ifrgpbyee) then
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'load wave function:'
          write(6,'(1x,a)') ' <beta|psi> in the file is inappropriate ', &
               'for the current condition'
          write(6,'(1x,a)') ' <beta|psi> is rejected'
       end if
    else
       if (myrank.eq.0) then
          do is=1,nspin
             do ik=1,mgs%nks
                read(NFWFI) (((pby(ip,it,ib,ik,is) &
                     ,ip=1,bspots(kd(it))%nprj),it=1,nsi),ib=1,nw)
             end do
          end do
       end if
       call MPI_BCAST(pby,laprj*nsi*nw*mgs%nks*nspin, &
            MPI_DOUBLE_COMPLEX,0,comm,ierr)

       if (ver.eq.1 ) then
          if (myrank.eq.0) then
             do is=1,nspin
                read(nfwfi)((ee(ib,ik,is),ib=1,nw),ik=1,mgs%nks)
             end do
          end if
          call MPI_BCAST(ee,nband*mgs%nks*nspin, &
               MPI_DOUBLE_PRECISION,0,comm,ierr)
       end if
    end if
    if (myrank.eq.0) close(NFWFI)

    return
  end subroutine load_wfn

  subroutine load_wfn_master3(nfin,sign,ncomp,nspin,ndens,nw,bb,ifrgpbyee,dp, &
       mgs,nb1,nb2,wgtk,nband,ee,mssg,myrank,nproc,comm)
    use util
    use types
    use mpi
    use meshgrid
    use symutil
    implicit none
    integer,intent(in):: nfin,ncomp,nspin,ndens,nw
    character(len=8):: sign
    real(kind=8),intent(in):: bb(3,3)
    logical:: ifrgpbyee
    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dp(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: nband
    real(kind=8):: ee(nband,mgs%nks,nspin)
    integer:: mssg,myrank,nproc,comm
    !
    integer:: ik,iki,is,i,j,ib,ig,irg
    real(kind=8):: aai(3,3), bbi(3,3)
    integer::  nspini,nksi,nwi,nkmmaxi
    integer:: ncompi
    complex(kind=8),allocatable:: pin(:,:)
    real(kind=8),allocatable:: prin(:)
    complex(kind=8),allocatable:: p(:,:)
    real(kind=8),allocatable:: ski(:,:)
    integer,allocatable:: kgi(:,:), nkmaryi(:)
    complex(kind=8),allocatable:: expwf(:,:,:,:)
    integer:: mai1, mai2, mai3, tkg(3), i1, i2, i3, ierr
    integer,allocatable:: ikn(:),irgn(:),itr(:,:)
    logical:: ifinvsymi
    integer:: iseed
    real(kind=8):: rmin
    real(kind=8):: w, eet
    real(kind=8):: ph

    integer:: nks,ldkg
!soption contarea(sk,nkmary,kg,g_count,g_displ)
    real(kind=8),pointer:: sk(:,:)
    integer,pointer:: nkmary(:),kg(:,:,:)
    integer,pointer:: g_count(:,:),g_displ(:,:)
    integer,allocatable:: rgi(:,:,:),pgi(:,:),srgi(:)
    integer:: ngi,nnpi
    logical:: iftrs
    integer:: mpi_typ_wfn
    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    complex(kind=8):: tsu2(2,2),tspr(2)

    nks = mgs%nks
    ldkg = mgs%ldkg
    sk => mgs%sk
    nkmary => mgs%nkm
    kg => mgs%kg
    g_count => mgs%g_count
    g_displ => mgs%g_displ
    !
    iseed = 240549957
    !
    iftrs = ndens.eq.1
    !
    if (sign .eq. 'TAPP:wfn') then
       ifinvsymi = .false.
    else if (sign .eq. 'TAPP:wfi') then
       ifinvsymi = .true.
    else
       write(6,*) 'load_wfn_master: bad sign'
       call MPI_ABORT(comm,1,ierr)
       stop
    end if

    read(nfin) aai, bbi
    read(nfin,iostat=ierr) nspini, nksi, nwi, ngi, nnpi, ncompi
    if (ierr.ne.0) then
       ncompi = 1
       read(nfin) nspini, nksi, nwi, ngi, nnpi
    end if
    if (ncompi.ne.ncomp) then
       write(6,*) ' error: inconsistent ncompi = ',ncompi
       call MPI_ABORT(comm,1,ierr)
       stop
    end if
    if ( (nspini .ne. 1) .and. (nspini .ne. 2) ) then
       write(6,*) ' error: nspini = ',nspini
       call MPI_ABORT(comm,1,ierr)
       stop
    end if
    allocate(nkmaryi(nksi), ski(3,nksi), &
         rgi(3,3,ngi), pgi(3,ngi), srgi(ngi), stat=ierr)
    call chkalloc('load_wfn','tmp', ierr)
    read(nfin) rgi, pgi, srgi
    read(nfin) (ski(:,iki),iki=1,nksi)
    !
    ifrgpbyee = .true.
    if (nks .eq. nksi) then
       ifrgpbyee = .false.
       do ik=1,nks
          if ( (ski(1,ik) .ne. sk(1,ik)) .or. (ski(2,ik) .ne. sk(2,ik)) &
               .or. (ski(3,ik) .ne. sk(3,ik)) ) then
             ifrgpbyee = .true.
             exit
          end if
       end do
    end if
    
    if (nw .gt. nwi) then
       if (mssg.gt.-1) then
          write(6,*) ' nw > nwi = ', nwi
          write(6,*) ' fill ib = [nwi+1,nw] with random wave func.'
       end if
       ifrgpbyee = .true.
    end if
    if (nspin .ne. nspini) then
       ifrgpbyee = .true.
    end if

    call MPI_BCAST(ifinvsymi,1,MPI_LOGICAL,0,comm,ierr)
    call MPI_BCAST(ifrgpbyee,1,MPI_LOGICAL,0,comm,ierr)
    call MPI_BCAST(nwi,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nspini,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nksi,1,MPI_INTEGER,0,comm,ierr)

    read(nfin) nkmaryi
    nkmmaxi=maxval(nkmaryi)
    allocate(kgi(3,nkmmaxi), stat=ierr)
    call chkalloc('load_wfn','kgi', ierr)

    allocate(ikn(nks),irgn(nks),itr(3,nks))
    write(6,'(1x,a)') 'ik, ikn, irgn, rmin, itr(3) = '
    do ik=1,nks
       call iksnst(sk(:,ik),ngi,rgi,srgi,bb,nksi,ski,&
            ikn(ik),irgn(ik),itr(:,ik),rmin)
       if (mssg.gt.-1) then
          write(6,'(2x,i4,1x,i4,1x,i3,e9.2,3(1x,i2))') &
               ik, ikn(ik), irgn(ik), rmin, itr(1,ik), itr(2,ik), itr(3,ik)
       end if
    end do

    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    do iki=1,nksi
       if (mssg.gt.-1) write(6,*) ' iki, nkmi = ', iki, nkmaryi(iki)
       read(nfin) ((kgi(i,ig),i=1,3),ig=1,nkmaryi(iki))
       mai1 = 0
       mai2 = 0
       mai3 = 0
       do ig=1,nkmaryi(iki)
          do irg=1,ngi
             tkg = matmul(rgi(:,:,irg),kgi(:,ig))
             mai1 = max(mai1, abs(tkg(1)))
             mai2 = max(mai2, abs(tkg(2)))
             mai3 = max(mai3, abs(tkg(3)))
          end do
       end do
       if (mssg.gt.-1) write(6,*) '  range of kgi = ', mai1, mai2, mai3
       allocate(expwf(ncomp,-mai1:mai1, -mai2:mai2, -mai3:mai3), stat=ierr)
       call chkalloc('load_wfn','expwf', ierr)

       allocate(pin(ncomp,0:nkmaryi(iki)),prin(0:nkmaryi(iki)),stat=ierr)
       call chkalloc('load_wfn','pin,prin', ierr)
       do is=1,nspini
          do ib=1,max(nwi,nw)
             pin(:,0) = 0.0d0
             pin = 0.0d0
             if (ib.le.nwi) then
                if (ifinvsymi) then
                   read(nfin) (prin(j),j=1,nkmaryi(iki)),w,eet
                   pin(1,:) = prin(:)
                else
                   read(nfin) ((pin(i,j),i=1,ncomp),j=1,nkmaryi(iki)),w,eet
                end if
             else
                do j=1,min(nw*2,nkmaryi(iki))
                   do i=1,ncomp
                      pin(i,j) = ran0(iseed)
                   end do
                end do
             end if

             if (ib.gt.nw .or. is.gt.nspin) cycle

             call MPI_BCAST(w,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
             call MPI_BCAST(eet,1,MPI_DOUBLE_PRECISION,0,comm,ierr)

             do ik=1,nks
                call MPI_BCAST(ikn(ik),1,MPI_INTEGER,0,comm,ierr)
                if (ikn(ik) .ne. iki) then
                   cycle
                end if

                if ((nb1.le.ib) .and. (ib.le.nb2)) then
                   if (nspin.gt.nspini) then
                      wgtk(ib,ik,1) = w*0.5d0
                      wgtk(ib,ik,2) = w*0.5d0
                   else
                      wgtk(ib,ik,is) = w
                   end if
                end if
                if (nspin.gt.nspini) then
                   ee(ib,ik,1) = eet
                   ee(ib,ik,2) = eet
                else
                   ee(ib,ik,is) = eet
                end if

                allocate(p(ncomp,nkmary(ik)),stat=ierr)
                call chkalloc('load_wfn','p', ierr)

                irg = irgn(ik)
                call rg2tsu2(rgi(:,:,irg),aai,tsu2)
                expwf = 0.0d0
                do ig=1,nkmaryi(iki)
                   tkg = srgi(irgn(ik))*matmul(rgi(:,:,irg),kgi(:,ig))

                   ph = dble(pgi(1,irg))*dble(tkg(1)) &
                        +dble(pgi(2,irg))*dble(tkg(2)) &
                        +dble(pgi(3,irg))*dble(tkg(3))
                   ph = 2.0d0*pi/dble(nnpi)*ph
                   if (srgi(irgn(ik)).eq.1) then
                      if (ncomp.eq.1) then
                         expwf(1,tkg(1),tkg(2),tkg(3)) = pin(1,ig) &
                              *dcmplx(cos(ph),-sin(ph))
                      else
                         expwf(1:2,tkg(1),tkg(2),tkg(3)) = &
                              matmul(tsu2,pin(1:2,ig)) &
                              *dcmplx(cos(ph),-sin(ph))
                      end if
                   else
                      if (ncomp.eq.1) then
                         expwf(1,tkg(1),tkg(2),tkg(3)) = conjg(pin(1,ig)) &
                              *dcmplx(cos(ph),-sin(ph))
                      else
                         tspr(1) = -conjg(pin(2,ig))
                         tspr(2) = conjg(pin(1,ig))
                         expwf(1:2,tkg(1),tkg(2),tkg(3)) = &
                              matmul(tsu2,tspr)*dcmplx(cos(ph),-sin(ph))
                      end if
                   end if
                end do

                p = (0.0d0, 0.0d0)
                do ig=1,nkmary(ik)
                   i1 = kg(1,ig,ik)+itr(1,ik)
                   i2 = kg(2,ig,ik)+itr(2,ik)
                   i3 = kg(3,ig,ik)+itr(3,ik)
                   if ( (abs(i1) .le. mai1) .and. (abs(i2) .le. mai2) .and. &
                        (abs(i3) .le. mai3) ) then
                      p(:,ig) = expwf(:,i1,i2,i3)
                   end if
                end do
                call MPI_SCATTERV(p,g_count(:,ik), &
                     g_displ(:,ik),mpi_typ_wfn, &
                     dp(ik,is)%wf(:,:,ib),g_count(myrank,ik), &
                     mpi_typ_wfn,0,comm,ierr)

                if (nspin.gt.nspini) then
                   dp(ik,2)%wf(:,:,ib) = dp(ik,1)%wf(:,:,ib)
                end if

                deallocate(p)
             end do
          end do
       end do

       deallocate(expwf,pin,prin)
    end do
    deallocate(ikn,irgn,itr)

    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)

    !
    ! set if wave funciton is orthonormal wrt S or not
    if (ifrgpbyee) then
       dp(:,:)%ifothnm = .false.
    else
       dp(:,:)%ifothnm = .true.
    end if
    !
    ! wave function is not randomly generated
    dp(:,:)%ifrand = .false.
    
    deallocate(rgi, pgi, srgi)
    deallocate(ski, kgi, nkmaryi)
    return
  end subroutine load_wfn_master3

  subroutine load_wfn_slave3(ncomp,nspin,nw,ifrgpbyee,dp,mgs, &
       nb1,nb2,wgtk,nband,ee,myrank,nproc,comm)
    use types
    use mpi
    use meshgrid
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: nspin,nw
    logical:: ifrgpbyee
    type (meshgrid_struct):: mgs
    type (wfn_for_k):: dp(mgs%nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    integer,intent(in):: nband
    real(kind=8):: ee(nband,mgs%nks,nspin)
    integer:: myrank,nproc,comm
    !
    integer:: ik,iki,is,ikn,ib
    integer:: nspini,nksi,nwi
    integer:: ierr
    logical:: ifinvsymi
    complex(kind=8):: p(2,1)
    real(kind=8):: w,eet
    !
    integer:: nks
!soption contarea(g_count,g_displ)
    integer,pointer:: g_count(:,:),g_displ(:,:)
    integer:: mpi_typ_wfn

    nks = mgs%nks
    g_count => mgs%g_count
    g_displ => mgs%g_displ

    call MPI_BCAST(ifinvsymi,1,MPI_LOGICAL,0,comm,ierr)
    call MPI_BCAST(ifrgpbyee,1,MPI_LOGICAL,0,comm,ierr)
    call MPI_BCAST(nwi,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nspini,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(nksi,1,MPI_INTEGER,0,comm,ierr)
    
    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    do iki=1,nksi
       do is=1,nspini
          do ib=1,max(nwi,nw)
             if (ib.gt.nw .or. is.gt.nspin) cycle

             call MPI_BCAST(w,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
             call MPI_BCAST(eet,1,MPI_DOUBLE_PRECISION,0,comm,ierr)

             do ik=1,nks
                !call iksnst(sk(:,ik),ng,rg,srg,bb,nksi,ski,ikn,irgn,itr,rmin)
                call MPI_BCAST(ikn,1,MPI_INTEGER,0,comm,ierr)
                if (ikn .ne. iki) then
                   cycle
                end if

                if ((nb1.le.ib) .and. (ib.le.nb2)) then
                   if (nspin.gt.nspini) then
                      wgtk(ib,ik,1) = w*0.5d0
                      wgtk(ib,ik,2) = w*0.5d0
                   else
                      wgtk(ib,ik,is) = w
                   end if
                end if

                if (nspin.gt.nspini) then
                   ee(ib,ik,1) = eet
                   ee(ib,ik,2) = eet
                else
                   ee(ib,ik,is) = eet
                end if

                call MPI_SCATTERV(p,g_count(:,ik), &
                     g_displ(:,ik),mpi_typ_wfn, &
                     dp(ik,is)%wf(:,:,ib),g_count(myrank,ik), &
                     mpi_typ_wfn,0,comm,ierr)
                !dp(ik,is)%wf(:,ib) = p(bkm:ekm)
                if (nspin.gt.nspini) then
                   dp(ik,2)%wf(:,:,ib) = dp(ik,1)%wf(:,:,ib)
                end if
             end do
          end do
       end do
    end do

    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)
    !
    ! set if wave funciton is orthonormal wrt S or not
    if (ifrgpbyee) then
       dp(:,:)%ifothnm = .false.
    else
       dp(:,:)%ifothnm = .true.
    end if
    !
    ! wave function is not randomly generated
    dp(:,:)%ifrand = .false.
    return
  end subroutine load_wfn_slave3

  subroutine iksnst(sk,ng,rg,srg,bb,nksi,ski,iksret,irgret,itrret,rmin)
    implicit none
    integer,intent(in):: ng, rg(3,3,ng), srg(ng), nksi
    integer,intent(out):: iksret, irgret, itrret(3)
    real(kind=8),intent(in):: sk(3),bb(3,3),ski(3,nksi)
    real(kind=8),intent(out):: rmin
    
    integer:: ig, iks, irg, itr(3)
    real(kind=8):: r, s, tsk(3), d(3)
    iksret = 0
    rmin = huge(rmin)
    do iks = 1, nksi
       r = huge(r)
       do ig = 1, ng
          tsk = srg(ig)*matmul(rg(:,:,ig),ski(:,iks))
          d = sk - tsk
          d = d + 0.5d0
          d = d - floor(d)
          d = d - 0.5d0
          d = matmul(bb,d)
          s = sqrt(d(1)*d(1) + d(2)*d(2) + d(3)*d(3))
          if (s .lt. r) then
             r = s
             irg = ig
             itr = floor(sk-tsk+0.5d0)
          end if
       end do
       if (r .lt. rmin) then
          iksret = iks
          irgret = irg
          itrret = itr
          rmin = r
       end if
    end do
    if (iksret .eq. 0) then
       write(6,*) 'iksnst: iksret = 0'
       stop
    end if
    return
  end subroutine iksnst

end module chain_wfn

module chain_moldyn
  integer,parameter:: num_mdstat = 6, num_mdparam = 5, num_mdastat = 1
contains
  subroutine store_moldyn(mode,casi,cavl,sysvol,imdstp,mdstat,mdastat)
    use iounit
    implicit none
    character(len=*),intent(in):: mode
    real(kind=8),dimension(:,:),intent(in):: casi, cavl
    real(kind=8),intent(in):: sysvol,mdstat(:),mdastat(:,:,:)
    integer,intent(in):: imdstp
    !
    call xrewind(NFMDCNT)
    write(NFMDCNT) len_trim(mode)
    write(NFMDCNT) trim(mode)
    write(NFMDCNT) casi, cavl, sysvol
    write(NFMDCNT) imdstp
    write(NFMDCNT) mdstat, mdastat
    return
  end subroutine store_moldyn

  subroutine load_moldyn(mode,casi,cavl,sysvol,imdstp,mdstat,mdastat, &
       myrank,comm)
    use iounit
    use mpi
    implicit none
    character(len=*),intent(out):: mode
    real(kind=8),dimension(:,:),intent(out):: casi, cavl
    real(kind=8),intent(out):: sysvol,mdstat(:),mdastat(:,:,:)
    integer,intent(out):: imdstp
    integer,intent(in):: myrank,comm
    !
    integer:: n
    integer:: ns,ierr
    !
    if (myrank.eq.0) then
       call xrewind(NFMDCNT)
       read(NFMDCNT,iostat=ierr) n
       if (ierr.ne.0) then
          write(6,'(1x,a)') 'error in opening molecular dynamics data'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       read(NFMDCNT) mode(:n)
       mode(n+1:) = ' '
       read(NFMDCNT) casi, cavl, sysvol
       read(NFMDCNT) imdstp
       read(NFMDCNT) mdstat, mdastat
    end if
    call MPI_BCAST(n,1,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(mode,len(mode),MPI_CHARACTER,0,comm,ierr)
    ns = size(casi,1)*size(casi,2)
    call MPI_BCAST(casi,ns,MPI_DOUBLE_PRECISION,0,comm,ierr)
    ns = size(cavl,1)*size(cavl,2)
    call MPI_BCAST(cavl,ns,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(sysvol,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(imdstp,1,MPI_INTEGER,0,comm,ierr)
    ns = size(mdstat,1)
    call MPI_BCAST(mdstat,ns,MPI_DOUBLE_PRECISION,0,comm,ierr)
    ns = size(mdastat,1)*size(mdastat,2)*size(mdastat,3)
    call MPI_BCAST(mdastat,ns,MPI_DOUBLE_PRECISION,0,comm,ierr)
    return
  end subroutine load_moldyn

end module chain_moldyn
