!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
!**********************************************************************
!                          1992-07-18
!                          1992-08-22  NW? ARE INTRODUCED
!     last modified at 1995-06-05
!**********************************************************************
!
#include "fftgrid.h"

module diag_hml
  public ps2vcg, scgdgl
contains
  subroutine ps2vcg(dp,ee,dd,nband,nks &
       ,ncomp,nspin,ndens &
       ,dvpj,ldvpj,qbtp &
       ,vg,nvx,nvy,nvz,nw,ndiag,itsb &
       ,bb,va,asi,nsi,nkd,kd,bspots,lpnum,maxr,laprj,pby &
       ,xctp,mgs,nb1,nb2,bzwk,wgtk &
       ,ifinvsym,nproc,comm,mssg)
    use types
    use bsplpot
    use utilmpi
    use converter
    use mpi
    use timeprof
    use xcf_types
    use meshgrid
    use util
    implicit none
    integer,intent(in):: nband,nks,nspin
    integer,intent(in):: ncomp,ndens
    integer,intent(in):: nvx,nvy,nvz,nw,nsi
    integer,intent(in):: lpnum,maxr,laprj
    integer,intent(in):: nkd,kd(nsi)
    integer,intent(in):: ndiag,itsb,mssg
    integer,intent(in):: ldvpj
    real(kind=8),intent(in):: dvpj(ldvpj,nks)
    real(kind=8),intent(in):: bb(3,3),va,asi(3,nsi)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi,nspin)
    type (bspot_for_atm):: bspots(nkd)
    complex(kind=8),intent(in):: &
         vg(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    logical:: ifinvsym
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgs
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk(nks),wgtk(nb1:nb2,nks,nspin)
    integer,intent(in):: nproc,comm
    type (wfn_for_k):: dp(nks,nspin)
    real(kind=8),intent(in):: qbtp(3)
    !
    !     WARNING: if ee != 0.0d0, scgdgl() will determine eps
    !              from the value of ee at the entrance.
    !              The code to chain calculations must be aware of this.
    !
    real(kind=8):: ee(nband,nks,nspin)
    complex(kind=8):: pby(laprj,nsi,nw,nks,nspin)
    !
    real(kind=8),parameter:: epsdef = 1.0d-8
    !
    integer:: nwx,nwy,nwz
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: g_count(:,:),g_displ(:,:)

    real(kind=8),allocatable:: vr(:,:)
    !
    type (wfn_for_k),allocatable:: dpf(:)
    !
    integer:: iks,is,ib,ig,ierr
    integer:: nlx,nly,nlz,nlxyz
    real(kind=8):: clk
    real(kind=8),allocatable:: out(:,:,:,:),epsks(:,:)
    real(kind=8):: omax1,omax2,omax3
    integer,allocatable:: kserr(:,:)
    logical:: ifoeps
    integer:: myrank,bkm,ekm
    integer:: idens

    call tpf_begin(tpf_dhml)
    !
    nwx = mgs%nwx
    nwy = mgs%nwy
    nwz = mgs%nwz
    irdkpt => mgs%irdkpt
    nkm => mgs%nkm
    kg => mgs%kg
    g_count => mgs%g_count
    g_displ => mgs%g_displ

    call MPI_COMM_RANK(comm,myrank,ierr)
    !
    nlx = FFT_PAD_NX(nwx*2)
    nly = FFT_PAD_NY(nwy*2)
    nlz = FFT_PAD_NZ(nwz*2)

    nlxyz=nlx*nly*nlz
    allocate(vr(nlxyz,ndens),stat=ierr)
    call chkalloc('ps2vcg','vr',ierr)
    !
    !
    !  BAND CALCULATION
    !
    ! Initialize pby as 0, because  only some part of pby is  occupied.
    pby = 0.0d0
    allocate(kserr(nks,nspin),epsks(nks,nspin), &
         out(nw,3,nks,nspin),stat=ierr)
    call chkalloc('ps2vcg','kserr,out',ierr)
    if (mssg.gt.0) then
       do iks=1,nks
          write(6,*) 'iks, nkm=',iks,nkm(iks)
       end do
    end if

    epsks = epsdef
    do is=1,nspin

       if (xcf_exch_factor(xctp) .ne. 0.0d0) then
          allocate(dpf(nks),stat=ierr)
          call chkalloc('ps2vcg','dpf',ierr)
          do iks=1,nks
             bkm = g_displ(myrank,iks)+1
             ekm = g_displ(myrank,iks)+g_count(myrank,iks)
             allocate(dpf(iks)%wf(ncomp,bkm:ekm,nw),stat=ierr)
             call chkalloc('ps2vcg','dpf%wf',ierr)

!$omp parallel do private(ib,ig)
             do ib=1,nw
                do ig=bkm,ekm
                   dpf(iks)%wf(:,ig,ib) = dp(iks,is)%wf(:,ig,ib)
                end do
             end do
             dpf(iks)%ifothnm = dp(iks,is)%ifothnm
          end do
       end if
       !
       if (ncomp.eq.1) then
          call vg2vr(vg(-nvx,-nvy,-nvz,is),vr(:,1),nvx,nvy,nvz, &
               nwx,nwy,nwz,nlx,nly,nlz)
       else
          do idens = 1,ndens
             call vg2vr(vg(-nvx,-nvy,-nvz,idens),vr(:,idens),nvx,nvy,nvz, &
                  nwx,nwy,nwz,nlx,nly,nlz)
          end do
       end if
       !
       do iks=1,nks
          !
          if (is.eq.1 .and. iks.eq.1) then
             ifoeps = .not. dp(iks,1)%ifrand
          else
             ifoeps = .true.
          end if

          if (is.eq.1 .and. iks.ne.1 .and. dp(iks,1)%ifrand) then
             call copywfn(ncomp,nw,irdkpt(1)%sk,nkm(1),kg(:,:,1),dp(1,1)%wf, &
                  irdkpt(iks)%sk,nkm(iks),kg(:,:,iks),dp(iks,1)%wf,nproc, &
                  g_count(:,1),g_displ(:,1), &
                  g_count(:,iks),g_displ(:,iks),ifinvsym,comm)
             ee(:,iks,1) = ee(:,1,1)
          else if (is.eq.2) then
             if (dp(iks,2)%ifrand) then
                dp(iks,2)%wf = dp(iks,1)%wf
                ee(:,iks,2) = ee(:,iks,1)
             end if
          end if

          bkm = g_displ(myrank,iks)+1
          ekm = g_displ(myrank,iks)+g_count(myrank,iks)
          call scgdgl(ncomp,ndiag,itsb,nwx,nwy,nwz,nlx,nly,nlz,nlxyz, &
               nkm(iks),bkm,ekm,nw,iks,irdkpt(iks)%sk,kg(:,:,iks),bb, &
               ee(1,iks,is),ndens,vr,out(1,1,iks,is),lpnum,maxr, &
               dvpj(1,iks),laprj,dd(1,1,1,is),ifoeps,epsks(iks,is), &
               kserr(iks,is),va,asi,bspots,nsi,nkd,kd,qbtp, &
               dp(iks,is),pby(1,1,1,iks,is),ifinvsym, &
               xctp,dpf,mgs,nb1,nb2,bzwk,wgtk(:,:,is), &
               nproc,g_count(:,iks),g_displ(:,iks),comm)

          dp(iks,is)%ifrand = .false.
       end do

       if (xcf_exch_factor(xctp) .ne. 0.0d0) then
          do iks=1,nks
             deallocate(dpf(iks)%wf,stat=ierr)
          end do
          deallocate(dpf)
       end if
    end do
    !B-ANNO
    ! now kserr contains itrn
    !E-ANNO
    if (mssg.gt.0) then
       clk=xetime(1)
       write(6,'(1x,a,f9.2)')' ETIME AFTER SCGDGL ',clk
       ! CONV CHECK
       write(6,'(1x,a)') 'IKS,  EIGEN,  RESID,  SIN,  EPS'
       do iks=1,nks
          omax1=0.0d0
          omax2=0.0d0
          omax3=0.0d0
          do is=1,nspin
             do ib=1,nw
                omax1=max(omax1,out(ib,1,iks,is))
                omax2=max(omax2,out(ib,2,iks,is))
                omax3=max(omax3,out(ib,3,iks,is))
             end do
             write(6,'(1x,i3,4(1x,e11.4))') iks,omax1,omax2,omax3, &
                  epsks(iks,is)
          end do
          if (nspin .eq. 2) write(6,'(1x)')
       end do
       do iks=1,nks
          write(6,100) irdkpt(iks)%sk
100       format(' SK=',3e15.7,30('.') )
          do is=1,nspin
             write(6,'(1x,a,i2,1x,i4)') 'IS, IRTN=',is,kserr(iks,is)
             write(6,300) (ee(ib,iks,is),ib=1,nw)
          end do
300       format(' EIGEN VALUE'/(' ',8e15.7))
          if (nspin .eq. 2) write(6,'(1x)')
       end do
    end if
    !
    deallocate(kserr,epsks,out,vr)
    !
    call tpf_end(tpf_dhml)
    return
  end subroutine ps2vcg

  subroutine scgdgl(ncomp,ndiag,itsb,nwx,nwy,nwz,nlx,nly,nlz,nlxyz, &
       nkm,bkm,ekm,nw, &
       iks,sk,kg,bb,ee,ndens,vr,out,lpnum,maxr,dvpj,laprj,dd,ifoeps,eps, &
       ier,va,asi,bspots,nsi,nkd,kd,qbtp,dp,pby,ifinvsym, &
       xctp,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
       nproc,g_count,g_displ,comm)
    !
    use mpi
    use bsplpot
    use types
    use diag_dvdsn
    use xcf_types
    use meshgrid
    implicit none
    integer,intent(in):: ndiag,itsb,nwx,nwy,nwz,nlx,nly,nlz,nlxyz, &
         nkm,bkm,ekm,nw,lpnum,maxr,laprj,nsi,nkd
    integer,intent(in):: ncomp,ndens
    real(kind=8),intent(in):: va
    logical:: ifinvsym,ifoeps
    real(kind=8):: eps
    !
    ! update dpf in diag iteration, if iks > 0
    integer,intent(in):: iks
    real(kind=8),intent(in):: sk(3),bb(3,3)
    real(kind=8),intent(in):: vr(nlxyz,ndens)
    integer,intent(in)::  kg(3,nkm),kd(nsi)
    real(kind=8),intent(in):: &
         asi(3,nsi),dvpj(bkm:ekm,maxr*lpnum*nkd)
    complex(kind=8),intent(in):: dd(laprj,laprj,nsi)
    real(kind=8),intent(in):: qbtp(3)
    type (bspot_for_atm):: bspots(nkd)
    !
    integer:: ier,ig,ib,igh,i
    real(kind=8):: out(nw,3),ee(nw)
    !
    type (wfn_for_k):: dp
    !
    complex(kind=8):: pby(laprj,nsi,nw)
    !
    integer,intent(in):: xctp
    type (meshgrid_struct):: mgsf
    type (wfn_for_k):: dpf(mgsf%nks)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: bzwk(mgsf%nks),wgtk(nb1:nb2,mgsf%nks)
    !
    integer,intent(in):: nproc,comm
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    !
    real(kind=8),allocatable:: dvpjg(:,:),dpi(:,:)
    complex(kind=8),allocatable:: dpg(:,:),dpgf(:,:)
!SOPTION CONTAREA(wfp)
    complex(kind=8),pointer:: wfp(:,:,:)
    integer:: myrank,ierr
    integer:: bkmg,ekmg,nkmg
    integer,allocatable:: kgg(:,:)
    integer:: irnk
    integer,allocatable:: gg_count(:),gg_displ(:)
    logical:: ulxchf
    real(kind=8),allocatable:: ddg(:,:,:)
    !
    call MPI_COMM_RANK(comm,myrank,ierr)
    !
    ulxchf = xcf_exch_factor(xctp) .eq. 0.0d0
    if (ncomp.eq.1 .and. ulxchf .and. ifinvsym) then
       allocate(dpi(bkm:ekm,nw),stat=ierr)
       call chkalloc('scgdgl','dpi',ierr)
       dpi = dble(dp%wf(1,:,:))
       call dvdsn1i(ndiag,itsb,nwx*2,nwy*2,nwz*2,nlx,nly,nlz,nlxyz, &
            nkm,bkm,ekm, &
            nw,sk,bb,kg,ee,vr, &
            out,lpnum,maxr,dvpj,laprj,dd,asi, &
            ifoeps,eps,ier,dpi,va,nsi,nkd,kd,bspots,pby,qbtp,dp%ifothnm, &
            nproc,myrank,g_count,g_displ,comm)
       dp%wf(1,:,:) = dpi
       deallocate(dpi)
    else if ( (.not. ifinvsym) .and. (ulxchf .or. mgsf%nks.eq.1) &
         .and. maxval(abs(sk)).eq.0.0d0 .and. ncomp.eq.1 ) then
       bkmg = bkm/2 + 1
       ekmg = (ekm-1)/2 + 1
       nkmg = (nkm-1)/2 + 1

       allocate(dvpjg(bkmg:ekmg,maxr*lpnum*nkd), &
            dpg(bkmg:ekmg,nw),kgg(3,nkmg), &
            gg_count(0:nproc-1),gg_displ(0:nproc-1),stat=ierr)
       call chkalloc('scgdgl','dvpjg,dpg,kgg,gg_c,gg_d',ierr)

       if (xcf_exch_factor(xctp) .ne. 0.0d0) then
          allocate(dpgf(bkmg:ekmg,nw),stat=ierr)
       else
          allocate(dpgf(1,1),stat=ierr)
       end if
       call chkalloc('scgdgl','dpgf',ierr)

       allocate(ddg(laprj,laprj,nsi),stat=ierr)
       call chkalloc('scgdgl','ddg',ierr)
       ddg = dble(dd)

       gg_count(0) = g_count(0)/2+1
       do irnk=1,nproc-1
          gg_count(irnk) = g_count(irnk)/2
       end do
       gg_displ(0) = 0
       do irnk=0,nproc-2
          gg_displ(irnk+1) = gg_displ(irnk)+gg_count(irnk)
       end do

       kgg(:,1) = kg(:,1)
       igh=2
       do ig=2,nkm,2
          kgg(:,igh) = kg(:,ig)
          igh = igh+1
       end do
       if (igh-1.ne.nkmg) then
          write(6,*) 'igh, nkmg', igh, nkmg
          stop
       end if

       if (bkm.eq.1) then
          do i=1,maxr*lpnum*nkd
             dvpjg(1,i) = dvpj(1,i)
          end do
          do ib=1,nw
             dpg(1,ib) = dble(dp%wf(1,1,ib))
          end do
          if (xcf_exch_factor(xctp) .ne. 0.0d0) then
             do ib=1,nw
                dpgf(1,ib) = dble(dpf(1)%wf(1,1,ib))
             end do
          end if
       end if
       do i=1,maxr*lpnum*nkd
          igh = max(bkm,2)/2+1
          do ig=max(bkm,2),ekm,2
             dvpjg(igh,i) = dvpj(ig,i)
             igh = igh+1
          end do
       end do
       do ib=1,nw
          igh = max(bkm,2)/2+1
          do ig=max(bkm,2),ekm,2
             dpg(igh,ib) = dp%wf(1,ig,ib)
             igh = igh+1
          end do
       end do
       if (xcf_exch_factor(xctp) .ne. 0.0d0) then
          do ib=1,nw
             igh = max(bkm,2)/2+1
             do ig=max(bkm,2),ekm,2
                dpgf(igh,ib) = dpf(1)%wf(1,ig,ib)
                igh = igh+1
             end do
          end do
       end if
       call dvdsn1g(ndiag,itsb,nwx*2,nwy*2,nwz*2,nlx,nly,nlz,nlxyz, &
            nkmg,bkmg,ekmg, &
            nw,iks,sk,bb,kgg,ee,vr, &
            out,lpnum,maxr,dvpjg,laprj,ddg,asi, &
            ifoeps,eps,ier,dpg,va,nsi,nkd,kd,bspots,pby, &
            qbtp,dp%ifothnm, &
            xctp,dpgf,mgsf,nb1,nb2,bzwk(1),wgtk, &
            nproc,myrank,gg_count,gg_displ,comm)
       if (bkm.eq.1) then
          do ib=1,nw
             dp%wf(1,1,ib) = dble(dpg(1,ib))
          end do
       end if
       do ib=1,nw
          igh = max(bkm,2)/2+1
          do ig=max(bkm,2),ekm,2
             dp%wf(1,ig,ib) = dpg(igh,ib)
             dp%wf(1,ig+1,ib) = conjg(dpg(igh,ib))
             igh = igh+1
          end do
       end do

       deallocate(ddg)
       deallocate(dvpjg,dpg,kgg,gg_count,gg_displ)
       deallocate(dpgf)
    else
       wfp => dp%wf
       call dvdsn1(ncomp,ndiag,itsb,nwx*2,nwy*2,nwz*2,nlx,nly,nlz,nlxyz, &
            nkm,bkm,ekm, &
            nw,iks,sk,bb,kg,ee,ndens,vr, &
            out,lpnum,maxr,dvpj,laprj,dd,asi, &
            ifoeps,eps,ier,wfp,va,nsi,nkd,kd,bspots,pby, &
            qbtp,dp%ifothnm, &
            xctp,dpf,mgsf,nb1,nb2,bzwk,wgtk, &
            nproc,myrank,g_count,g_displ,comm)
    end if

    !
    return
  end subroutine scgdgl
end module diag_hml
