!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2014,2017 Yoshihide Yoshimoto
!
!
#include "fftgrid.h"
!
program distpart
  use iounit
  use bsplpot
  use util
  use xcf_types
  use mpi
  use replica
  use chkinput
  use subr_rddata
  use chain_atom_cg
  use converter
  use types
  use subr_engpart
  use subr_distpart
  implicit none
#include "config90.h"
  !
  integer pg(3,nsymq),rg(3,3,nsymq),srg(nsymq)
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  real(kind=8):: bb0(3,3),bb(3,3),ax,bx,va,vb
  integer:: ng,nnp,nsv,ismode
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: nsi,nkd
  integer:: nrx,nry,nrz,nhx,nhy,nhz,nrxyz,ntt &
       ,nrcell,nspin,ngx,ngy,ngz,ngxyz &
       ,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz
  integer:: ndens
  real(kind=8):: qf,rf,qm,qc,qwf,qbtp(3)
  real(kind=8):: qbsp
  integer:: nbsf
  integer:: nkr,nkfxa,nkc,nng,nband,nw,nb1,nb2
  !
  integer:: iopt(nopt)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer nsk,nks,nk,nkmesh
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  !
  real(kind=8):: clk
  integer:: mssg,ierr
  real(kind=8):: ekbt
  real(kind=8):: fwqcom
  integer:: xctp
  real(kind=8),pointer:: zoin(:),znin(:)
  !
  integer:: nproc,myrank,comm
  type (replica_struct):: repls
  !
  real(kind=8),pointer:: udirchr(:,:)
  !
  real(kind=8),allocatable:: rho(:,:,:)
  integer,allocatable:: idxpar(:,:,:)
  real(kind=8),allocatable:: excor(:,:,:),hrtor(:,:,:),ekinor(:,:,:)
  real(kind=8),allocatable:: sekinor(:,:,:,:),shrtor(:,:,:,:)
  real(kind=8),allocatable:: evxcor(:,:,:,:)

  type (energy_per_atom):: engatm
  type (senergy_per_atom):: sengatm
  integer:: inrx2,inry2,inrz2
  !
  integer:: it,isek
  integer:: insi, inkd
  integer,allocatable:: idxcar2xtp(:)
  integer:: argc
  character(len=256):: argv(2)
  type (eng_part_struct):: engparc
  integer:: ikd
  !
  call init_engparc(engparc)
  !
  call MPI_INIT(ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD,nproc,ierr)
  if (nproc.eq.1) then
     call MPI_COMM_DUP(MPI_COMM_WORLD,comm,ierr)
     call MPI_COMM_DUP(MPI_COMM_WORLD,repls%comt,ierr)
     repls%ifreplm = .false.
  else
     call setup_replica(NFINPUT,MPI_COMM_WORLD,comm,repls)
  end if
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)
  !
  ! distpart should be performed by single MPI proc for each replica.
  !
  if (myrank.ne.0) then
     call MPI_FINALIZE(ierr)
     stop
  end if

  clk=xetime(0)
  write(6,*)'DISTPART'
  mssg=1
  !
  call rdinidata(ax,bx,aa0,bb0,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz,ntt,ngx,ngy,ngz,ngxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,NFINPUT,mssg &
       ,qbsp,nbsf,fwqcom,xctp,zoin,znin)

  engparc%qfengpar = qf
  engparc%num_bin_wgt = 10

  argc = command_argument_count()
  call get_command_argument(1, argv(1))
  if (trim(argv(1)).eq.'bader') then
     engparc%mode = 3
     write(6,'(1x,a)') 'partition mode = bader'
     if (argc.eq.2) then
        call get_command_argument(2, argv(2))
        read(argv(2),*) engparc%qfengpar
     end if
     write(6,'(/1x,a,f10.5/)') 'cutoff wave-num for part-wgt = ', &
          engparc%qfengpar
  else if (trim(argv(1)).eq.'voronoi') then
     engparc%mode = 2
     write(6,'(1x,a)') 'partition mode = voronoi'
     if (argc.eq.2) then
        call get_command_argument(2, argv(2))
        read(argv(2),*) engparc%qfengpar
     end if
     write(6,'(/1x,a,f10.5/)') 'cutoff wave-num for part-wgt = ', &
          engparc%qfengpar
     allocate(engparc%refsz(nkd))
     do ikd=1,nkd
        read(5,*) engparc%refsz(ikd)
     end do
  else if (trim(argv(1)).eq.'trinkle') then
     engparc%mode = 4
     write(6,'(1x,a)') 'partition mode = trinkle'
     call get_command_argument(2, argv(2))
  else
     write(6,'(1x,a,a)') 'invalid partition mode = ', trim(argv(1))
     stop
  end if
  
  allocate(udirchr(3,nsi))
  udirchr = 0.0d0
  call chkall(ng,rg,pg,srg,nnp,aa,nsi,kd,asi,ndens,udirchr,MPI_COMM_SELF)
  deallocate(udirchr)

  allocate(bspots(nkd),stat=ierr)
  call chkalloc('distpart','bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     write(6,*)
     call tellbspot(nkd,bspots)
  end if

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)
  vb = cellvol(bb)

  allocate(rho(nrx*2,nry*2,nrz*2), &
       idxpar(nrx*2,nry*2,nrz*2), &
       excor(nrx*2,nry*2,nrz*2), evxcor(nrx*2,nry*2,nrz*2,0:6), &
       hrtor(nrx*2,nry*2,nrz*2), &
       ekinor(nrx*2,nry*2,nrz*2), sekinor(nrx*2,nry*2,nrz*2,6), &
       shrtor(nrx*2,nry*2,nrz*2,6), stat=ierr)
  call chkalloc('distpart','work arrays',ierr)

  allocate(idxcar2xtp(nsi))
  if (engparc%mode .eq. 3 .or. engparc%mode .eq. 4) then
     call xrewind(NFCAR2XTP)
     read(NFCAR2XTP,*) ! skip comment
     read(NFCAR2XTP,*) inkd,insi
     if (nsi.ne.insi .or. nkd.ne.inkd) then
        write(6,*) 'inconsistent car2xtp data'
        stop
     end if
     do it=1,nsi
        read(NFCAR2XTP,*) idxcar2xtp(it)
     end do
     close(NFCAR2XTP)
  else
     do it=1,nsi
        idxcar2xtp(it) = it
     end do
  end if

  call energy_per_atom_load(NFLOCENG,nsi,engatm,sengatm)

  call xrewind(NFCHR)
  read(NFCHR) ! magic 'TAPP:rho'
  read(NFCHR) ! aa, nspin
  read(NFCHR) inrx2,inry2,inrz2
  read(NFCHR) rho
  close(NFCHR)

  if (engparc%mode .eq. 3) then
     call xrewind(NFIDXPAR)
     read(NFIDXPAR,*) inrx2,inry2,inrz2
     read(NFIDXPAR,*) idxpar
  end if

  if (engparc%mode .eq. 2 .or. engparc%mode .eq. 3) then 
     call xrewind(NFEKINOR)
     read(NFEKINOR) ! magic 'TAPP:rho'
     read(NFEKINOR) ! aa, 1
     read(NFEKINOR) inrx2,inry2,inrz2
     read(NFEKINOR) ekinor
     close(NFEKINOR)

     call xrewind(NFSEKINOR)
     do isek=1,6
        read(NFSEKINOR) ! magic 'TAPP:rho'
        read(NFSEKINOR) ! aa, 1
        read(NFSEKINOR) inrx2,inry2,inrz2
        read(NFSEKINOR) sekinor(:,:,:,isek)
     end do
     close(NFSEKINOR)

     call xrewind(NFEXCOR)
     read(NFEXCOR) ! magic 'TAPP:rho'
     read(NFEXCOR) ! aa, 1
     read(NFEXCOR) inrx2,inry2,inrz2
     read(NFEXCOR) excor
     close(NFEXCOR)

     call xrewind(NFEVXCOR)
     do isek=0,6
        read(NFEVXCOR) ! magic 'TAPP:rho'
        read(NFEVXCOR) ! aa, 1
        read(NFEVXCOR) inrx2,inry2,inrz2
        read(NFEVXCOR) evxcor(:,:,:,isek)
     end do
     close(NFEVXCOR)

     call xrewind(NFHRTOR)
     read(NFHRTOR) ! magic 'TAPP:rho'
     read(NFHRTOR) ! aa, 1
     read(NFHRTOR) inrx2,inry2,inrz2
     read(NFHRTOR) hrtor
     close(NFHRTOR)

     call xrewind(NFSHRTOR)
     do isek=1,6
        read(NFSHRTOR) ! magic 'TAPP:rho'
        read(NFSHRTOR) ! aa, 1
        read(NFSHRTOR) inrx2,inry2,inrz2
        read(NFSHRTOR) shrtor(:,:,:,isek)
     end do
     close(NFSHRTOR)
  end if
  
  if (engparc%mode .eq. 3) then
     call calcidxparengpart(aa,nsi,nrx*2,nry*2,nrz*2, &
          ekinor,excor,evxcor,hrtor,sekinor,shrtor,rho, &
          engatm,sengatm,idxpar,idxcar2xtp,engparc,mssg)
  else if (engparc%mode .eq. 2) then
     call calcvoroengpart(aa,nsi,kd,asi,nkd,nrx*2,nry*2,nrz*2, &
       ekinor,excor,evxcor,hrtor,sekinor,shrtor,rho, &
       engatm,sengatm,engparc,mssg)
  else if (engparc%mode .eq. 4) then
     call read_trinkle(argv(2),nsi,idxcar2xtp,engatm,sengatm)
     call calctrinkengpart(nsi,engatm,sengatm,mssg)
  else
     write(6,*) 'invalid mode', engparc%mode
     stop
  end if

  call write_distpart(nkd,nsi,kd,engatm,sengatm,bspots,6)
  
  call energy_per_atom_dealloc(engatm)
  call senergy_per_atom_dealloc(sengatm)

  deallocate(idxcar2xtp)
  deallocate(rho,idxpar,excor,evxcor,hrtor)
  if (associated(engparc%refsz)) deallocate(engparc%refsz)
  deallocate(kd,asi,mmm,zoin,znin)

  clk=xetime(1)
  write(6,*)
  write(6,*) 'TOTAL TIME=',clk
  call MPI_FINALIZE(ierr)
  stop
end program distpart
