! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
module global_esm
implicit none
type esm_prm
  integer              ::ngau,nfit,idebug,gp_s,gp_e,iffix,nz_r,nz_l
  real(kind=8)         ::z1,fix_ef,add_elec,z_r,z_l,v0
  character(len=50)    ::boundary_condition
  character(len=1)     ::normal
  integer              ::bctype
  logical              ::enable
end type esm_prm
type cg2esm
  integer              ::nsi,nkd,nrx,nry,nrz,nrx2,nry2,nrz2,nvx,nvy,nvz &
                        ,ng
!___ANDO___
  integer              ::nkfxa
  integer,pointer      ::kf(:,:)
!___E_ANDO___
  integer,pointer      ::kd(:)
  real(kind=8)         ::aa(3,3),bb(3,3),qf,va
  real(kind=8),pointer ::zv(:),agauss(:,:),bgauss(:,:)
end type cg2esm
type esm_2d
  !!!!! n0   : # of mesh for surface normal direction
  !!!!! n1,n2: # of mesh for surface parallel direction
  integer              ::nkf_2d,naf,n0,n1,n2 
  integer,pointer      ::kkg_2d(:,:)
  real(kind=8)         ::aa_2d(2,2),bb_2d(2,2),sa,sb,L,Lg
  real(kind=8),pointer ::akk_2d(:),asi_z(:),asi_2d(:,:)
end type esm_2d

type(esm_prm) ::esm

contains

  subroutine esm_write_esm_info(cginp,var2d,esm)
    implicit none
    type(cg2esm),intent(in)  ::cginp
    type(esm_2d),intent(in)  ::var2d
    type(esm_prm),intent(in) ::esm
    write(6,'(A)')'**************** ESM information *****************'
    write(6,'(A)')esm%boundary_condition
    if (esm%normal .eq. 'X') then
       write(6,'(A)')' Surface normal direction is x-axis'
       write(6,'(A,F8.4,A,F8.4,A)')' Cell size(Lx)=',var2d%L,' a.u., x1=' &
            ,esm%z1,' a.u.'
    else if (esm%normal .eq. 'Z') then
       write(6,'(A)')' Surface normal direction is z-axis'
       write(6,'(A,F8.4,A,F8.4,A)')' Cell size(Lz)=',var2d%L,' a.u., z1=' &
            ,esm%z1,' a.u.'
    end if
    write(6,'(A,I5,A,F8.4,A)')' # of g_parallel vector: ',var2d%nkf_2d
    write(6,'(A,I5)')' # of real space data points: ',var2d%n0
    write(6,'(A,I4)')' smoothing for g_parallel=0(nfit):',esm%nfit
    if (esm%normal .eq. 'X') then
       write(6,'(A,I5,A,F10.5)')' nrx_left =',esm%nz_l,', x_left =',esm%z_l
       write(6,'(A,I5,A,F10.5)')' nrx_right=',esm%nz_r,', x_right=',esm%z_r
    else if (esm%normal .eq. 'Z') then
       write(6,'(A,I5,A,F10.5)')' nrz_left =',esm%nz_l,', z_left =',esm%z_l
       write(6,'(A,I5,A,F10.5)')' nrz_right=',esm%nz_r,', z_right=',esm%z_r
    end if
    if(esm%iffix.eq.1)then
      write(6,'(A)')' N=const. mode'
      write(6,'(A,F10.5)')' # of added electron:',esm%add_elec
    else
      write(6,'(A)')' Ef=fix mode'
      write(6,'(A,F10.5)')' Fixed fermi energy is ',esm%fix_ef
    endif
if (esm%bctype .eq. 4) then
    write(6,'(A,F10.5,A)')' Electric field=',esm%v0/2.d0/esm%z1,' ht./a.u.'
    write(6,'(A,F10.5,A)')' Bias voltage=',esm%v0*2.d0*13.6,' V'
end if
    write(6,'(A)')'**************************************************'
  end subroutine esm_write_esm_info

  subroutine esm_write_esm_info2file(cginp,var2d,esm)
    implicit none
    type(cg2esm),intent(in)  ::cginp
    type(esm_2d),intent(in)  ::var2d
    type(esm_prm),intent(in) ::esm


  end subroutine esm_write_esm_info2file

end module global_esm
