! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
module m_pot ! 1D Poisson_solvers
  use global_esm
contains
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------------main routine--gp!=0----------------------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_pot_solve_Poisson_equation(va,z1,nz2,L,gp,eht,vg,rho,fftw1,fftw2,fs)
use fft_1d
implicit none
integer,intent(in)          ::nz2
real(kind=8),intent(in)     ::L,z1,gp,va
real(kind=8),intent(inout)  ::eht
complex(kind=8),intent(in)  ::rho(nz2)
complex(kind=8),intent(out) ::vg(nz2)
!local
integer         ::nz,iz
real(kind=8)    ::pi,pi2,pi4,z0,sgp,z,kn
complex(kind=8) ::ci=(0.d0,1.d0),tmp1,tmp2
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
nz=nz2/2
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
z0=L/2.d0
sgp=sqrt(gp)
!
!!!!!!!!!!! start apply greens function to rho !!!!!!!!!!!
tmp1=0.d0; tmp2=0.d0
!...bare Coulomb
do iz=1,nz2
  if(iz<=nz) kn=int(iz-1    )*2.d0*pi/L
  if(iz> nz) kn=int(iz-1-nz2)*2.d0*pi/L
if (esm%bctype .eq. 1) then
  tmp1=tmp1+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0))/(sgp-ci*kn)
  tmp2=tmp2+rho(iz)*(cos(kn*z0)-ci*sin(kn*z0))/(sgp+ci*kn)
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  tmp1=tmp1+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0)) &
      *((sgp+ci*kn)*exp(sgp*(z1-z0))+(sgp-ci*kn)*exp(-sgp*(z1-z0))) &
      /(sgp**2+kn**2)/2.d0/sgp
  tmp2=tmp2+rho(iz)*(cos(kn*z0)-ci*sin(kn*z0)) &
      *((sgp-ci*kn)*exp(sgp*(z1-z0))+(sgp+ci*kn)*exp(-sgp*(z1-z0))) &
      /(sgp**2+kn**2)/2.d0/sgp
else if (esm%bctype .eq. 3) then
  tmp1=tmp1+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0)) &
      *((sgp+ci*kn)*exp(sgp*(z1-z0))+(sgp-ci*kn)*exp(-sgp*(z1-z0))) &
      /(sgp**2+kn**2)/2.d0/sgp
  tmp2=tmp2+rho(iz)*(cos(kn*z0)-ci*sin(kn*z0))*(sgp-ci*kn) &
      /(sgp**2+kn**2)/sgp
endif
  vg(iz)=pi4*rho(iz)/(sgp**2+kn**2)
enddo
!call wfft1d(nz2,vg,1) ! FFT(g->r)
call fft1_bw(fs,vg,fftw1,fftw2) !FFT(g->r)
do iz=1,nz2
  z=(iz-1)*L/nz2
  if(z>L/2) z=z-L
if (esm%bctype .eq. 1) then
  vg(iz)=vg(iz)-pi2*exp( sgp*(z-z0))*tmp1/sgp &
               -pi2*exp(-sgp*(z+z0))*tmp2/sgp
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  vg(iz)=vg(iz)-pi4*(exp(sgp*(z-z1))-exp(-sgp*(z+3.d0*z1)))*tmp1 &
                /(1.d0-exp(-4.d0*sgp*z1)) &
               +pi4*(exp(sgp*(z-3.d0*z1))-exp(-sgp*(z+z1)))*tmp2 &
                /(1.d0-exp(-4.d0*sgp*z1))
else if (esm%bctype .eq. 3) then
  vg(iz)=vg(iz)-pi4*exp(sgp*(z-z1))*tmp1 &
               +pi2*(exp(sgp*(z-z0-2.d0*z1))-exp(-sgp*(z+z0)))*tmp2
endif
enddo
!!!!!!!!!!! end apply greens function to rho !!!!!!!!!!!
!
!call wfft1d(nz2,vg,-1)   ! FFT(r->g)
call fft1_fw(fs,vg,fftw1,fftw2) !FFT(g->r)
!
!!!!!!!!!!! calculate energy !!!!!!!!!!!!
do iz=1,nz2
  eht=eht+vg(iz)*dconjg(rho(iz))*va*0.5d0
enddo
return
end subroutine m_pot_solve_Poisson_equation

!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------------main routine--gp=0----------------------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_pot_solve_Poisson_equation_gp0(va,z1,z_r,z_l,nz_r,nz_l &
                                           ,nz2,L,eht,vg,rho,fftw1,fftw2,fs)
use fft_1d
implicit none
integer,intent(in)          ::nz2,nz_r,nz_l
real(kind=8),intent(in)     ::L,va,z1,z_r,z_l
real(kind=8),intent(inout)  ::eht
complex(kind=8),intent(in)  ::rho(nz2)
complex(kind=8),intent(out) ::vg(nz2)
!local
integer         ::nz,iz
real(kind=8)    ::pi,pi2,pi4,z0,z,zr,zl,kn
complex(kind=8) ::ci=(0.d0,1.d0),tmp1,tmp2,tmp3,tmp4 &
                 ,c_r,c_l,s_r,s_l,f1,f2,f3,f4,a0,a1,a2,a3
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
nz=nz2/2
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
z0=L/2.d0
!
!!!!!!!!!!! start apply greens function to rho !!!!!!!!!!!
tmp1=0.d0; tmp2=0.d0; tmp3=0.d0; tmp4=0.d0
f1=0.d0; f2=0.d0; f3=0.d0; f4=0.d0
!...bare Coulomb
if (esm%bctype .eq. 1) then
vg(1)=-pi2*z0**2*rho(1)
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
vg(1)=pi2*(2.d0*z1-z0)*z0*rho(1)
else if (esm%bctype .eq. 3) then
vg(1)=pi2*(4.d0*z1-z0)*z0*rho(1)
endif
do iz=2,nz2
  if(iz<=nz) kn=int(iz-1    )*2.d0*pi/L
  if(iz> nz) kn=int(iz-1-nz2)*2.d0*pi/L
if (esm%bctype .eq. 1) then
  tmp1=tmp1+ci*rho(iz)*(cos(kn*z0)+ci*sin(kn*z0))/kn
  tmp2=tmp2+ci*rho(iz)*(cos(kn*z0)-ci*sin(kn*z0))/kn
  tmp3=tmp3+rho(iz)*cos(kn*z0)/kn**2
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  tmp1=tmp1+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0))/kn**2
  tmp2=tmp2+rho(iz)*(cos(kn*z0)-ci*sin(kn*z0))/kn**2
  tmp3=tmp3+rho(iz)*ci*cos(kn*z0)/kn
  tmp4=tmp4+rho(iz)*sin(kn*z0)/kn
else if (esm%bctype .eq. 3) then
  tmp1=tmp1+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0))/kn**2
  tmp2=tmp2+rho(iz)*(cos(kn*z0)-ci*sin(kn*z0))/kn
  tmp3=tmp3+rho(iz)*(cos(kn*z0)+ci*sin(kn*z0))/kn
endif
  vg(iz)=pi4*rho(iz)/kn**2
  !____for smoothing____
  c_r=cos(kn*z_r)
  s_r=sin(kn*z_r)
  c_l=cos(kn*z_l)
  s_l=sin(kn*z_l)
  f1=f1+pi4*   rho(iz)*(c_r+ci*s_r)/kn**2
  f2=f2+pi4*   rho(iz)*(c_l+ci*s_l)/kn**2
  f3=f3+pi4*ci*rho(iz)*(c_r+ci*s_r)/kn
  f4=f4+pi4*ci*rho(iz)*(c_l+ci*s_l)/kn
enddo
!call wfft1d(nz2,vg,1) ! FFT(g->r)
call fft1_bw(fs,vg,fftw1,fftw2) !FFT(g->r)
!...bare Coulomb rest
do iz=1,nz2
  z=(iz-1)*L/nz2
  if(z>L/2) z=z-L
if (esm%bctype .eq. 1) then
  vg(iz)=vg(iz)-pi2*z**2*rho(1) &
                 -pi2*(z-z0)*tmp1 &
                 -pi2*(z+z0)*tmp2 &
                 -pi4*tmp3
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  vg(iz)=vg(iz)-pi2*z**2*rho(1) &
                 -pi2*(z+z1)*tmp1/z1 &
                 +pi2*(z-z1)*tmp2/z1 &
                 -pi4*z*(z1-z0)/z1*tmp3 &
                 +pi4  *(z1-z0)   *tmp4
else if (esm%bctype .eq. 3) then
  vg(iz)=vg(iz)-pi2*(z**2+2.d0*z*z0)*rho(1) &
                 -pi4*tmp1 &
                 -pi4*ci*(z-z1)*tmp2 &
                 -pi4*ci*(z1-z0)*tmp3
endif
enddo
!____for smoothing____
if (esm%bctype .eq. 1) then
f1=f1-pi2*z_r**2*rho(1) &
     -pi2*(z_r-z0)*tmp1 &
     -pi2*(z_r+z0)*tmp2 &
     -pi4*tmp3 &
     -pi2*z0**2*rho(1)
f2=f2-pi2*z_l**2*rho(1) &
     -pi2*(z_l-z0)*tmp1 &
     -pi2*(z_l+z0)*tmp2 &
     -pi4*tmp3 &
     -pi2*z0**2*rho(1)
f3=f3-pi2*tmp1 &
     -pi2*tmp2 &
     -pi4*z_r*rho(1)
f4=f4-pi2*tmp1 &
     -pi2*tmp2 &
     -pi4*z_l*rho(1)
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
f1=f1-pi2*z_r**2*rho(1) &
     -pi2*(z_r+z1)*tmp1/z1 &
     +pi2*(z_r-z1)*tmp2/z1 &
     -pi4*z_r*(z1-z0)/z1*tmp3 &
     +pi4  *(z1-z0)   *tmp4 &
     +pi2*(2.d0*z1-z0)*z0*rho(1)
f2=f2-pi2*z_r**2*rho(1) &
     -pi2*(z_l+z1)*tmp1/z1 &
     +pi2*(z_l-z1)*tmp2/z1 &
     -pi4*z_l*(z1-z0)/z1*tmp3 &
     +pi4  *(z1-z0)   *tmp4 &
     +pi2*(2.d0*z1-z0)*z0*rho(1)
f3=f3-pi4*z_r*rho(1) &
     -pi2*tmp1/z1 &
     +pi2*tmp2/z1 &
     -pi4*(z1-z0)/z1*tmp3
f4=f4-pi4*z_l*rho(1) &
     -pi2*tmp1/z1 &
     +pi2*tmp2/z1 &
     -pi4*(z1-z0)/z1*tmp3
else if (esm%bctype .eq. 3) then
f1=f1-pi2*(z_r**2+2.d0*z_r*z0)*rho(1) &
     -pi4*tmp1 &
     -pi4*ci*(z_r-z1)*tmp2 &
     -pi4*ci*(z1-z0)*tmp3 &
     +pi2*(4.d0*z1-z0)*z0*rho(1)
f2=f2-pi2*(z_l**2+2.d0*z_l*z0)*rho(1) &
     -pi4*tmp1 &
     -pi4*ci*(z_l-z1)*tmp2 &
     -pi4*ci*(z1-z0)*tmp3 &
     +pi2*(4.d0*z1-z0)*z0*rho(1)
f3=f3-pi2*(2.d0*z_r+2.d0*z0)*rho(1) &
     -pi4*ci*tmp2
f4=f4-pi2*(2.d0*z_l+2.d0*z0)*rho(1) &
     -pi4*ci*tmp2
endif
!!!!!!!!!!! end apply greens function to rho !!!!!!!!!!!

!!!!!!!!!!! start smooth_potential only gp=0 !!!!!!!!!!!!
  zr=z_r
  zl=z_l+L
  a0=(f1*zl**2*(zl-3.d0*zr)+z_r*(f3*zl**2*(-zl+zr) &
     +zr*(f2*(3.d0*zl-zr)+f4*zl*(-zl+zr))))/(zl-zr)**3
  a1=(f3*zl**3+zl*(6.d0*f1-6.d0*f2+(f3+2.d0*f4)*zl)*zr &
    -(2*f3+f4)*zl*zr**2-f4*zr**3)/(zl-zr)**3
  a2=(-3*f1*(zl+zr)+3.d0*f2*(zl+zr)-(zl-zr)*(2*f3*zl &
     +f4*zl+f3*zr+2*f4*zr))/(zl-zr)**3
  a3=(2.d0*f1-2.d0*f2+(f3+f4)*(zl-zr))/(zl-zr)**3
  do iz=nz_r,nz_l
    z=(iz-1)*L/nz2
    vg(iz)=a0+a1*z+a2*z**2+a3*z**3
  enddo
!!!!!!!!!!! end smooth_potential only gp=0 !!!!!!!!!!!!
!
!call wfft1d(nz2,vg,-1) ! FFT(r->g)
call fft1_fw(fs,vg,fftw1,fftw2) !FFT(g->r)
!
!!!!!!!!!!! calculate energy !!!!!!!!!!!!
do iz=1,nz2
  eht=eht+vg(iz)*dconjg(rho(iz))*va*0.5d0
enddo
return
end subroutine m_pot_solve_Poisson_equation_gp0

end module m_pot
