!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!

#define FFT3D_PROF

module fft_3d
  use mkl_dfti
  !
  type fft3_struct
     integer:: nx,ny,nz,nxyz,lx,ly,lz,lxyz
     type (dfti_descriptor), pointer:: dh,dhb
     logical:: ifsft
  end type fft3_struct

  logical,parameter:: fft_3d_cap_rh = .true.
contains
  !
  subroutine fft3_init(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    integer:: status,length(3),stride(4)
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz

    length(1) = mx
    length(2) = my
    length(3) = mz
    status = DftiCreateDescriptor(fs%dh, DFTI_DOUBLE, DFTI_COMPLEX, 3, length)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    stride(1) = 0
    stride(2) = 1
    stride(3) = lmx
    stride(4) = lmx*lmy
    status = DftiSetValue(fs%dh, DFTI_INPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    status = DftiSetValue(fs%dh, DFTI_OUTPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiSetValue(fs%dh, DFTI_COMPLEX_STORAGE, DFTI_COMPLEX_COMPLEX)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiSetValue(fs%dh, DFTI_FORWARD_SCALE, 1.0d0/fs%nxyz)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    status = DftiSetValue(fs%dh, DFTI_BACKWARD_SCALE, 1.0d0)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiCommitDescriptor(fs%dh)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiCopyDescriptor(fs%dh,fs%dhb)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    fs%ifsft = .false.
    return
  end subroutine fft3_init

  subroutine fft3_init_rh(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    integer:: status,length(3),stride(4)
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz

    length(1) = mx
    length(2) = my
    length(3) = mz
    status = DftiCreateDescriptor(fs%dh, DFTI_DOUBLE, DFTI_REAL, 3, length)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    stride(1) = 0
    stride(2) = 1
    stride(3) = 2*lmx
    stride(4) = 2*lmx*lmy
    status = DftiSetValue(fs%dh, DFTI_INPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    stride(1) = 0
    stride(2) = 1
    stride(3) = lmx
    stride(4) = lmx*lmy
    status = DftiSetValue(fs%dh, DFTI_OUTPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiSetValue(fs%dh, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiSetValue(fs%dh, DFTI_FORWARD_SCALE, 1.0d0/fs%nxyz)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    status = DftiSetValue(fs%dh, DFTI_BACKWARD_SCALE, 1.0d0)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiCopyDescriptor(fs%dh,fs%dhb)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    stride(1) = 0
    stride(2) = 1
    stride(3) = lmx
    stride(4) = lmx*lmy
    status = DftiSetValue(fs%dhb, DFTI_INPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)
    stride(1) = 0
    stride(2) = 1
    stride(3) = 2*lmx
    stride(4) = 2*lmx*lmy
    status = DftiSetValue(fs%dhb, DFTI_OUTPUT_STRIDES, stride)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiCommitDescriptor(fs%dh)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    status = DftiCommitDescriptor(fs%dhb)
    if (status .ne. 0) write(6,*) DftiErrorMessage(status)

    fs%ifsft = .false.
    return
  end subroutine fft3_init_rh

  subroutine fft3_done(fs)
    implicit none
    type (fft3_struct):: fs
    integer:: status
    !
    status = DftiFreeDescriptor(fs%dh)
    status = DftiFreeDescriptor(fs%dhb)
    return
  end subroutine fft3_done

  subroutine fft3_fw(fs,dat,tmp)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)

#ifdef FFT3D_PROF
    call tpf_begin(tpf_ffts)
    call tpf_begin(tpf_fftsc)
#endif
    call mkl_call_dfti_computeforward(fs,dat)
#ifdef FFT3D_PROF
    call tpf_end(tpf_fftsc)
    call tpf_end(tpf_ffts)
#endif
    return
  end subroutine fft3_fw

  subroutine fft3_bw(fs,dat,tmp)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    !

#ifdef FFT3D_PROF
    call tpf_begin(tpf_ffts)
    call tpf_begin(tpf_fftsc)
#endif
    call mkl_call_dfti_computebackward(fs,dat)
#ifdef FFT3D_PROF
    call tpf_end(tpf_fftsc)
    call tpf_end(tpf_ffts)
#endif
    return
  end subroutine fft3_bw
  !
  subroutine fft3_fwsx(fs,dat,tmp,ix)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix
    !
#ifdef FFT3D_PROF
    call tpf_begin(tpf_fftc)
#endif

    if (ix.eq.1) then
       call mkl_call_dfti_computeforward(fs,dat)
    else if (ix.eq.-1) then
       call mkl_call_dfti_computeforward(fs,tmp)
    else
       write(6,*) 'fft3_fwsx: z, abs(ix) != 1'
       stop
    end if

#ifdef FFT3D_PROF
    call tpf_end(tpf_fftc)
#endif
    return
  end subroutine fft3_fwsx

  subroutine fft3_bwsx(fs,dat,tmp,ix)
#ifdef FFT3D_PROF
    use timeprof
#endif
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix
    !
#ifdef FFT3D_PROF
    call tpf_begin(tpf_fftc)
#endif

    if (ix.eq.1) then
       call mkl_call_dfti_computebackward(fs,dat)
    else if (ix.eq.-1) then
       call mkl_call_dfti_computebackward(fs,tmp)
    else
       write(6,*) 'fft3_bwsx: z, abs(ix) != 1'
       stop
    end if

#ifdef FFT3D_PROF
    call tpf_end(tpf_fftc)
#endif
    return
  end subroutine fft3_bwsx

end module fft_3d

subroutine mkl_call_dfti_computeforward(fs,dat)
  use fft_3d
  implicit none
  type (fft3_struct):: fs
  complex(kind=8):: dat(fs%lxyz)
  integer:: status
  status = DftiComputeForward(fs%dh,dat)
  if (status .ne. 0) write(6,*) DftiErrorMessage(status)
  return
end subroutine mkl_call_dfti_computeforward

subroutine mkl_call_dfti_computebackward(fs,dat)
  use fft_3d
  implicit none
  type (fft3_struct):: fs
  complex(kind=8):: dat(fs%lxyz)
  integer:: status
  status = DftiComputeBackward(fs%dhb,dat)
  if (status .ne. 0) write(6,*) DftiErrorMessage(status)
  return
end subroutine mkl_call_dfti_computebackward
