!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module mult_loc_pot
private calcvpsi
contains
  subroutine calcvpsi(ncomp,nlxyz,ndens,vr,vztmp)
    implicit none
    integer,intent(in):: ncomp,nlxyz,ndens
    real(kind=8),intent(in):: vr(nlxyz,ndens)
    real(kind=8):: FFT_A1M_DEFSZ(vztmp,nlxyz,ncomp)

    real(kind=8):: vdia(2)
    complex(kind=8):: psi(2),voff,vpsi(2)
    integer:: ig

    if (ncomp.eq.1) then
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1M_RL(vztmp,ig,1) = vr(ig,1)*FFT_A1M_RL(vztmp,ig,1)
          FFT_A1M_IM(vztmp,ig,1) = vr(ig,1)*FFT_A1M_IM(vztmp,ig,1)
       end do
    else
       if (ndens.eq.1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1M_RL(vztmp,ig,1) = vr(ig,1)*FFT_A1M_RL(vztmp,ig,1)
             FFT_A1M_IM(vztmp,ig,1) = vr(ig,1)*FFT_A1M_IM(vztmp,ig,1)
             FFT_A1M_RL(vztmp,ig,2) = vr(ig,1)*FFT_A1M_RL(vztmp,ig,2)
             FFT_A1M_IM(vztmp,ig,2) = vr(ig,1)*FFT_A1M_IM(vztmp,ig,2)
          end do
       else
!$omp parallel do private(ig,psi,vdia,voff,vpsi)
          do ig=1,nlxyz
             psi(1) = dcmplx(FFT_A1M_RL(vztmp,ig,1),FFT_A1M_IM(vztmp,ig,1))
             psi(2) = dcmplx(FFT_A1M_RL(vztmp,ig,2),FFT_A1M_IM(vztmp,ig,2))
             vdia(1) = vr(ig,1)+vr(ig,4)
             vdia(2) = vr(ig,1)-vr(ig,4)
             voff = dcmplx(vr(ig,2),vr(ig,3))
             vpsi(1) = vdia(1)*psi(1) + conjg(voff)*psi(2)
             vpsi(2) = voff*psi(1) + vdia(2)*psi(2)
             FFT_A1M_RL(vztmp,ig,1) = dble(vpsi(1))
             FFT_A1M_IM(vztmp,ig,1) = aimag(vpsi(1))
             FFT_A1M_RL(vztmp,ig,2) = dble(vpsi(2))
             FFT_A1M_IM(vztmp,ig,2) = aimag(vpsi(2))
          end do
       end if
    end if
    return
  end subroutine calcvpsi

  subroutine hploc(ncomp,nlxyz,nkm,bkm,ekm,nw,p,hp,ndens,vr,j2gs,fs, &
       nproc,g_count,g_displ,comm)
    !
    use mpi
    use fft_3d
    use timeprof
    implicit none
    integer,intent(in):: ncomp,ndens
    integer,intent(in):: nlxyz,nkm,bkm,ekm,nw
    complex(kind=8),intent(in):: p(ncomp,bkm:ekm,nw)
    real(kind=8),intent(in):: vr(nlxyz,ndens)
    complex(kind=8),intent(out):: hp(ncomp,bkm:ekm,nw)
    integer:: j2gs(nkm,2)
    type(fft3_struct):: fs
    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm
    !
    real(kind=8),dimension(:,:,:),allocatable:: vztmp1,vztmp2
    complex(kind=8),dimension(:,:),allocatable:: tp,thp
    integer,dimension(:),allocatable:: b_count,b_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: err
    !
    integer:: ig,jg,kg,ib,ix,tix,irnk,myrank
    integer:: mpi_typ_wfn,ic
    !
    call tpf_begin(tpf_hpl)

    call MPI_COMM_RANK(comm,myrank,err)
    allocate(FFT_A1M_DEFSZ(vztmp1,nlxyz,ncomp),FFT_A1M_DEFSZ(vztmp2,nlxyz,ncomp), &
         b_count(0:nproc-1),b_displ(0:nproc-1), &
         s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=err)
    call chkalloc('hploc','tmp',err)

    b_count = nw/nproc
    do irnk=0,mod(nw,nproc)-1
       b_count(irnk) = b_count(irnk)+1
    end do
    b_displ(0) = 0
    do irnk=0,nproc-2
       b_displ(irnk+1) = b_displ(irnk) + b_count(irnk)
    end do

    allocate(tp(ncomp,nkm*b_count(myrank)),thp(ncomp,nkm*b_count(myrank)),stat=err)
    call chkalloc('hploc','tmp2',err)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,err)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,err)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,err)
    call MPI_ALLTOALLV(p,s_count,s_displ,mpi_typ_wfn, &
         tp,r_count,r_displ,mpi_typ_wfn,comm,err)
    call tpf_end(tpf_mpi)
    !
    !     MAIN LOOP
    !
!CDIR RESERVE
    do ib=0,b_count(myrank)-1
       !
       call tpf_begin(tpf_hplm)
       do ic=1,ncomp
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1M_RL(vztmp1,ig,ic) = 0.0d0
             FFT_A1M_IM(vztmp1,ig,ic) = 0.0d0
          end do
       end do
       call tpf_end(tpf_hplm)
       call tpf_begin(tpf_scgt)
       if (ncomp.eq.1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
             do kg=1,g_count(irnk)
                FFT_A1M_RL(vztmp1,j2gs(ig+kg,1),1) = dble(tp(1,jg+kg))
                FFT_A1M_IM(vztmp1,j2gs(ig+kg,1),1) = aimag(tp(1,jg+kg))
             end do
          end do
       else
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
             do kg=1,g_count(irnk)
                FFT_A1M_RL(vztmp1,j2gs(ig+kg,1),1) = dble(tp(1,jg+kg))
                FFT_A1M_IM(vztmp1,j2gs(ig+kg,1),1) = aimag(tp(1,jg+kg))
                FFT_A1M_RL(vztmp1,j2gs(ig+kg,1),2) = dble(tp(2,jg+kg))
                FFT_A1M_IM(vztmp1,j2gs(ig+kg,1),2) = aimag(tp(2,jg+kg))
             end do
          end do
       end if
       call tpf_end(tpf_scgt)
       !
       ix = 1
       call fft3_bwsx(fs,FFT_A1M_RL(vztmp1,1,1),FFT_A1M_RL(vztmp2,1,1),ix)
       if (ncomp.eq.2) then
          tix = 1
          call fft3_bwsx(fs,FFT_A1M_RL(vztmp1,1,2),FFT_A1M_RL(vztmp2,1,2),tix)
       end if
       !
       call tpf_begin(tpf_hplm)
       if (ix .eq. 1) then
          call calcvpsi(ncomp,nlxyz,ndens,vr,vztmp1)
       else if (ix .eq. -1) then
          call calcvpsi(ncomp,nlxyz,ndens,vr,vztmp2)
       else
          write(6,*) 'hploc: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_hplm)
       !
       tix = ix
       call fft3_fwsx(fs,FFT_A1M_RL(vztmp1,1,1),FFT_A1M_RL(vztmp2,1,1),ix)
       if (ncomp.eq.2) then
          call fft3_fwsx(fs,FFT_A1M_RL(vztmp1,1,2),FFT_A1M_RL(vztmp2,1,2),tix)
       end if
       !
       call tpf_begin(tpf_scgt)
       if (ix .eq. 1) then
          if (ncomp.eq.1) then
             do irnk=0,nproc-1
                jg = r_displ(irnk)+g_count(irnk)*ib
                ig = g_displ(irnk)
!$omp parallel do
                do kg=1,g_count(irnk)
                   thp(1,jg+kg) = dcmplx(FFT_A1M_RL(vztmp1,j2gs(ig+kg,2),1), &
                        FFT_A1M_IM(vztmp1,j2gs(ig+kg,2),1))
                end do
             end do
          else
             do irnk=0,nproc-1
                jg = r_displ(irnk)+g_count(irnk)*ib
                ig = g_displ(irnk)
!$omp parallel do
                do kg=1,g_count(irnk)
                   thp(1,jg+kg) = dcmplx(FFT_A1M_RL(vztmp1,j2gs(ig+kg,2),1), &
                        FFT_A1M_IM(vztmp1,j2gs(ig+kg,2),1))
                   thp(2,jg+kg) = dcmplx(FFT_A1M_RL(vztmp1,j2gs(ig+kg,2),2), &
                        FFT_A1M_IM(vztmp1,j2gs(ig+kg,2),2))
                end do
             end do
          end if
       else if (ix .eq. -1) then
          if (ncomp.eq.1) then
             do irnk=0,nproc-1
                jg = r_displ(irnk)+g_count(irnk)*ib
                ig = g_displ(irnk)
!$omp parallel do
                do kg=1,g_count(irnk)
                   thp(1,jg+kg) = dcmplx(FFT_A1M_RL(vztmp2,j2gs(ig+kg,2),1), &
                        FFT_A1M_IM(vztmp2,j2gs(ig+kg,2),1))
                end do
             end do
          else
             do irnk=0,nproc-1
                jg = r_displ(irnk)+g_count(irnk)*ib
                ig = g_displ(irnk)
!$omp parallel do
                do kg=1,g_count(irnk)
                   thp(1,jg+kg) = dcmplx(FFT_A1M_RL(vztmp2,j2gs(ig+kg,2),1), &
                        FFT_A1M_IM(vztmp2,j2gs(ig+kg,2),1))
                   thp(2,jg+kg) = dcmplx(FFT_A1M_RL(vztmp2,j2gs(ig+kg,2),2), &
                        FFT_A1M_IM(vztmp2,j2gs(ig+kg,2),2))
                end do
             end do
          end if
       else
          write(6,*) 'hploc: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_scgt)
    end do
!CDIR RELEASE
    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(thp,r_count,r_displ,mpi_typ_wfn, &
         hp,s_count,s_displ,mpi_typ_wfn,comm,err)
    call MPI_TYPE_FREE(mpi_typ_wfn,err)
    call tpf_end(tpf_mpi)
    !
    deallocate(vztmp1,vztmp2,tp,thp, &
         b_count,b_displ,s_count,s_displ,r_count,r_displ)
    !
    call tpf_end(tpf_hpl)
    return
  end subroutine hploc
  
  subroutine hploci(nlxyz,nkm,bkm,ekm,nw,p,hp,vr,j2gs,fs, &
       nproc,g_count,g_displ,comm)
    !
    use fft_3d
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: nlxyz,nkm,bkm,ekm,nw
    real(kind=8),intent(in):: p(bkm:ekm,nw),vr(nlxyz)
    real(kind=8),intent(out):: hp(bkm:ekm,nw)
    integer:: j2gs(nkm,2)
    type(fft3_struct):: fs
    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm
    !
    real(kind=8),dimension(:,:),allocatable:: vztmp1,vztmp2
    real(kind=8),dimension(:),allocatable:: tp,thp
    integer,dimension(:),allocatable:: b_count,b_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: err
    !
    integer:: ig,ib,ibr,ix,jg,kg,irnk,myrank
    !
    call tpf_begin(tpf_hpl)

    call MPI_COMM_RANK(comm,myrank,err)
    allocate(FFT_A1D_DEFSZ(vztmp1,nlxyz),FFT_A1D_DEFSZ(vztmp2,nlxyz), &
         b_count(0:nproc-1),b_displ(0:nproc-1), &
         s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=err)
    call chkalloc('hploci','tmp',err)

    b_count = nw/nproc
    do irnk=0,mod(nw,nproc)-1
       b_count(irnk) = b_count(irnk)+1
    end do
    b_displ(0) = 0
    do irnk=0,nproc-2
       b_displ(irnk+1) = b_displ(irnk) + b_count(irnk)
    end do

    allocate(tp(nkm*b_count(myrank)),thp(nkm*b_count(myrank)),stat=err)
    call chkalloc('hploci','tmp2',err)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,err)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(p,s_count,s_displ,MPI_DOUBLE_PRECISION, &
         tp,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,err)
    call tpf_end(tpf_mpi)
    !
    !     MAIN LOOP
    !
!CDIR RESERVE
    ibr = mod(b_count(myrank),2)
    if (ibr .eq. 1) then
       !
       call tpf_begin(tpf_hplm)
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1D_RL(vztmp1,ig) = 0.0d0
          FFT_A1D_IM(vztmp1,ig) = 0.0d0
       end do
       call tpf_end(tpf_hplm)
       call tpf_begin(tpf_scgt)
       do irnk=0,nproc-1
          jg = r_displ(irnk)
          ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(vztmp1)
!VOCL LOOP,DISJOINT(vztmp1)
!CDIR NODEP(vztmp1)
!DEC$ IVDEP
!VOPTION INDEP(vztmp1)
!POPTION INDEP(vztmp1)
          do kg=1,g_count(irnk)
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,1)) = tp(jg+kg)
          end do
       end do
       call tpf_end(tpf_scgt)
       ix = 1
       !
       call fft3_bwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_hplm)
       if (ix .eq. 1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp1,ig) = vr(ig)*FFT_A1D_RL(vztmp1,ig)
             FFT_A1D_IM(vztmp1,ig) = vr(ig)*FFT_A1D_IM(vztmp1,ig)
          end do
       else if (ix .eq. -1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp2,ig) = vr(ig)*FFT_A1D_RL(vztmp2,ig)
             FFT_A1D_IM(vztmp2,ig) = vr(ig)*FFT_A1D_IM(vztmp2,ig)
          end do
       else
          write(6,*) 'hploci: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_hplm)
       !
       call fft3_fwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_scgt)
       if (ix .eq. 1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
!POPTION INDEP(thp)
             do kg=1,g_count(irnk)
                thp(jg+kg) = FFT_A1D_RL(vztmp1,j2gs(ig+kg,2))
             end do
          end do
       else if (ix .eq. -1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
!POPTION INDEP(thp)
             do kg=1,g_count(irnk)
                thp(jg+kg) = FFT_A1D_RL(vztmp2,j2gs(ig+kg,2))
             end do
          end do
       else
          write(6,*) 'hploci: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_scgt)
    end if
    do ib=ibr,b_count(myrank)-1,2
       !
       call tpf_begin(tpf_hplm)
       do ig=1,nlxyz
          FFT_A1D_RL(vztmp1,ig) = 0.0d0
          FFT_A1D_IM(vztmp1,ig) = 0.0d0
       end do
       call tpf_end(tpf_hplm)
       call tpf_begin(tpf_scgt)
       do irnk=0,nproc-1
          jg = r_displ(irnk)+g_count(irnk)*ib
          ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(vztmp1)
!VOCL LOOP,DISJOINT(vztmp1)
!CDIR NODEP(vztmp1)
!DEC$ IVDEP
!VOPTION INDEP(vztmp1)
!POPTION INDEP(vztmp1)
          do kg=1,g_count(irnk)
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,1)) = tp(jg+kg)
             FFT_A1D_IM(vztmp1,j2gs(ig+kg,1)) = tp(jg+kg+g_count(irnk))
          end do
       end do
       call tpf_end(tpf_scgt)
       ix = 1
       !
       call fft3_bwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_hplm)
       if (ix .eq. 1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp1,ig) = vr(ig)*FFT_A1D_RL(vztmp1,ig)
             FFT_A1D_IM(vztmp1,ig) = vr(ig)*FFT_A1D_IM(vztmp1,ig)
          end do
       else if (ix .eq. -1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp2,ig) = vr(ig)*FFT_A1D_RL(vztmp2,ig)
             FFT_A1D_IM(vztmp2,ig) = vr(ig)*FFT_A1D_IM(vztmp2,ig)
          end do
       else
          write(6,*) 'hploci: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_hplm)
       !
       call fft3_fwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_scgt)
       if (ix .eq. 1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
!POPTION INDEP(thp)
             do kg=1,g_count(irnk)
                thp(jg+kg) = FFT_A1D_RL(vztmp1,j2gs(ig+kg,2))
                thp(jg+kg+g_count(irnk)) = FFT_A1D_IM(vztmp1,j2gs(ig+kg,2))
             end do
          end do
       else if (ix .eq. -1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
!POPTION INDEP(thp)
             do kg=1,g_count(irnk)
                thp(jg+kg) = FFT_A1D_RL(vztmp2,j2gs(ig+kg,2))
                thp(jg+kg+g_count(irnk)) = FFT_A1D_IM(vztmp2,j2gs(ig+kg,2))
             end do
          end do
       else
          write(6,*) 'hploci: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_scgt)
    end do
!CDIR RELEASE
    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(thp,r_count,r_displ,MPI_DOUBLE_PRECISION, &
         hp,s_count,s_displ,MPI_DOUBLE_PRECISION,comm,err)
    call tpf_end(tpf_mpi)
    !
    deallocate(vztmp1,vztmp2,tp,thp, &
         b_count,b_displ,s_count,s_displ,r_count,r_displ)
    !
    call tpf_end(tpf_hpl)
    return
  end subroutine hploci

  subroutine hplocg(nlxyz,nkm,bkm,ekm,nw,p,hp,vr,j2gs,fs, &
       nproc,g_count,g_displ,comm)
    !
    use mpi
    use fft_3d
    use timeprof
    implicit none
    integer,intent(in):: nlxyz,nkm,bkm,ekm,nw
    complex(kind=8),intent(in):: p(bkm:ekm,nw)
    real(kind=8),intent(in):: vr(nlxyz)
    complex(kind=8),intent(out):: hp(bkm:ekm,nw)
    integer:: j2gs(nkm,4)
    type(fft3_struct):: fs
    integer:: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm
    !
    real(kind=8),dimension(:,:),allocatable:: vztmp1,vztmp2
    complex(kind=8),dimension(:),allocatable:: tp,thp
    integer,dimension(:),allocatable:: b_count,b_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: err
    !
    integer:: ig,jg,kg,ib,ibr,ix,irnk,myrank
    !
    call tpf_begin(tpf_hpl)

    call MPI_COMM_RANK(comm,myrank,err)
    allocate(FFT_A1D_DEFSZ(vztmp1,nlxyz),FFT_A1D_DEFSZ(vztmp2,nlxyz), &
         b_count(0:nproc-1),b_displ(0:nproc-1), &
         s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=err)
    call chkalloc('hplocg','tmp',err)

    b_count = nw/nproc
    do irnk=0,mod(nw,nproc)-1
       b_count(irnk) = b_count(irnk)+1
    end do
    b_displ(0) = 0
    do irnk=0,nproc-2
       b_displ(irnk+1) = b_displ(irnk) + b_count(irnk)
    end do

    allocate(tp(nkm*b_count(myrank)),thp(nkm*b_count(myrank)),stat=err)
    call chkalloc('hplocg','tmp2',err)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,err)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(p,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         tp,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,err)
    call tpf_end(tpf_mpi)
    !
    !     MAIN LOOP
    !
!CDIR RESERVE
    ibr = mod(b_count(myrank),2)
    if (ibr .eq. 1) then
       call tpf_begin(tpf_hplm)
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1D_RL(vztmp1,ig) = 0.0d0
          FFT_A1D_IM(vztmp1,ig) = 0.0d0
       end do
       call tpf_end(tpf_hplm)
       call tpf_begin(tpf_scgt)
       do irnk=0,nproc-1
          jg = r_displ(irnk)
          ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(vztmp1)
!VOCL LOOP,DISJOINT(vztmp1)
!CDIR NODEP(vztmp1)
!DEC$ IVDEP
!VOPTION INDEP(vztmp1)
!POPTION INDEP(vztmp1)
          do kg=1,g_count(irnk)
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,1)) = dble(tp(jg+kg))
             FFT_A1D_IM(vztmp1,j2gs(ig+kg,1)) = aimag(tp(jg+kg))
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,2)) = dble(tp(jg+kg))
             FFT_A1D_IM(vztmp1,j2gs(ig+kg,2)) = -aimag(tp(jg+kg))
          end do
       end do
       call tpf_end(tpf_scgt)
       ix = 1
       !
       call fft3_bwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_hplm)
       if (ix .eq. 1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp1,ig) = vr(ig)*FFT_A1D_RL(vztmp1,ig)
             FFT_A1D_IM(vztmp1,ig) = vr(ig)*FFT_A1D_IM(vztmp1,ig)
          end do
       else if (ix .eq. -1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp2,ig) = vr(ig)*FFT_A1D_RL(vztmp2,ig)
             FFT_A1D_IM(vztmp2,ig) = vr(ig)*FFT_A1D_IM(vztmp2,ig)
          end do
       else
          write(6,*) 'hplocg: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_hplm)
       !
       call fft3_fwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_scgt)
       if (ix .eq. 1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!POPTION INDEP(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
             do kg=1,g_count(irnk)
                thp(jg+kg) = dcmplx(FFT_A1D_RL(vztmp1,j2gs(ig+kg,3)), &
                     FFT_A1D_IM(vztmp1,j2gs(ig+kg,3)))
             end do
          end do
       else if (ix .eq. -1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!POPTION INDEP(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
             do kg=1,g_count(irnk)
                thp(jg+kg) = dcmplx(FFT_A1D_RL(vztmp2,j2gs(ig+kg,3)), &
                     FFT_A1D_IM(vztmp2,j2gs(ig+kg,3)))
             end do
          end do
       else
          write(6,*) 'hplocg: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_scgt)
       ! set imag part of G = 0
       thp(1) = dble(thp(1))
    end if

    do ib=ibr,b_count(myrank)-1,2
       !
       call tpf_begin(tpf_hplm)
!$omp parallel do
       do ig=1,nlxyz
          FFT_A1D_RL(vztmp1,ig) = 0.0d0
          FFT_A1D_IM(vztmp1,ig) = 0.0d0
       end do
       call tpf_end(tpf_hplm)
       call tpf_begin(tpf_scgt)
       do irnk=0,nproc-1
          jg = r_displ(irnk)+g_count(irnk)*ib
          ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(vztmp1)
!VOCL LOOP,DISJOINT(vztmp1)
!CDIR NODEP(vztmp1)
!DEC$ IVDEP
!VOPTION INDEP(vztmp1)
!POPTION INDEP(vztmp1)
          do kg=1,g_count(irnk)
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,1)) = dble(tp(jg+kg))-aimag(tp(jg+kg+g_count(irnk)))
             FFT_A1D_IM(vztmp1,j2gs(ig+kg,1)) = aimag(tp(jg+kg))+dble(tp(jg+kg+g_count(irnk)))
             FFT_A1D_RL(vztmp1,j2gs(ig+kg,2)) = dble(tp(jg+kg))+aimag(tp(jg+kg+g_count(irnk)))
             FFT_A1D_IM(vztmp1,j2gs(ig+kg,2)) = -aimag(tp(jg+kg))+dble(tp(jg+kg+g_count(irnk)))
          end do
       end do
       call tpf_end(tpf_scgt)
       ix = 1
       !
       call fft3_bwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_hplm)
       if (ix .eq. 1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp1,ig) = vr(ig)*FFT_A1D_RL(vztmp1,ig)
             FFT_A1D_IM(vztmp1,ig) = vr(ig)*FFT_A1D_IM(vztmp1,ig)
          end do
       else if (ix .eq. -1) then
!$omp parallel do
          do ig=1,nlxyz
             FFT_A1D_RL(vztmp2,ig) = vr(ig)*FFT_A1D_RL(vztmp2,ig)
             FFT_A1D_IM(vztmp2,ig) = vr(ig)*FFT_A1D_IM(vztmp2,ig)
          end do
       else
          write(6,*) 'hplocg: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_hplm)
       !
       call fft3_fwsx(fs,vztmp1,vztmp2,ix)
       !
       call tpf_begin(tpf_scgt)
       if (ix .eq. 1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!POPTION INDEP(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
             do kg=1,g_count(irnk)
                thp(jg+kg) = dcmplx( &
                     0.5d0*( FFT_A1D_RL(vztmp1,j2gs(ig+kg,3)) + FFT_A1D_RL(vztmp1,j2gs(ig+kg,4)) ), &
                     0.5d0*( FFT_A1D_IM(vztmp1,j2gs(ig+kg,3)) - FFT_A1D_IM(vztmp1,j2gs(ig+kg,4)) ) )
                thp(jg+kg+g_count(irnk)) = dcmplx( &
                     0.5d0*( FFT_A1D_IM(vztmp1,j2gs(ig+kg,3)) + FFT_A1D_IM(vztmp1,j2gs(ig+kg,4)) ), &
                     0.5d0*(-FFT_A1D_RL(vztmp1,j2gs(ig+kg,3)) + FFT_A1D_RL(vztmp1,j2gs(ig+kg,4)) ) )
             end do
          end do
       else if (ix .eq. -1) then
          do irnk=0,nproc-1
             jg = r_displ(irnk)+g_count(irnk)*ib
             ig = g_displ(irnk)
!$omp parallel do
!OCL NOVREC(thp)
!POPTION INDEP(thp)
!CDIR NODEP(thp)
!DEC$ IVDEP
             do kg=1,g_count(irnk)
                thp(jg+kg) = dcmplx( &
                     0.5d0*( FFT_A1D_RL(vztmp2,j2gs(ig+kg,3)) + FFT_A1D_RL(vztmp2,j2gs(ig+kg,4)) ), &
                     0.5d0*( FFT_A1D_IM(vztmp2,j2gs(ig+kg,3)) - FFT_A1D_IM(vztmp2,j2gs(ig+kg,4)) ) )
                thp(jg+kg+g_count(irnk)) = dcmplx( &
                     0.5d0*( FFT_A1D_IM(vztmp2,j2gs(ig+kg,3)) + FFT_A1D_IM(vztmp2,j2gs(ig+kg,4)) ), &
                     0.5d0*(-FFT_A1D_RL(vztmp2,j2gs(ig+kg,3)) + FFT_A1D_RL(vztmp2,j2gs(ig+kg,4)) ) )
             end do
          end do
       else
          write(6,*) 'hplocg: abs(ix) != 1'
          stop
       end if
       call tpf_end(tpf_scgt)
       !
       ! set imag part of G = 0
       thp(g_count(0)*ib + 1) = dble(thp(g_count(0)*ib + 1))
       thp(g_count(0)*(ib+1) + 1) = dble(thp(g_count(0)*(ib+1) + 1))
    end do
!CDIR RELEASE
    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(thp,r_count,r_displ,MPI_DOUBLE_COMPLEX, &
         hp,s_count,s_displ,MPI_DOUBLE_COMPLEX,comm,err)
    call tpf_end(tpf_mpi)
    !
    deallocate(vztmp1,vztmp2,tp,thp, &
         b_count,b_displ,s_count,s_displ,r_count,r_displ)
    !
    call tpf_end(tpf_hpl)
    return
  end subroutine hplocg
end module mult_loc_pot
