!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2014,2017 Yoshihide Yoshimoto
!
!
program inipot
  !*************************************************************
  !  CG  DIAGONALIZATION PROGRAM WITH VB POTENTIAL
  !   irf,iif,icf,ilf are introduced 1995-06-09
  !  Screw symmetry revived. 1995-10-20
  !  Introducing uniform background charge,
  !     we can calculate charged state.
  !    Check 'znel' 1996-08-19
  !     znext: extra electron per supercell
  !      Ex. znext= 2: -2 state
  !          znext=-2: +2 state
  !*************************************************************
  use iounit
  use bsplpot
  use util
  use xcf_types
  use mpi
  use replica
  use chkinput
  use subr_rddata
  use subr_kpoint
#ifdef ESM
  use esm_manage
#endif
  implicit none
#include "config90.h"
  !
  integer pg(3,nsymq),rg(3,3,nsymq),srg(nsymq)
  real(kind=8):: aa(3,3),bb(3,3),ax,bx,va,vb
  integer:: ng,nnp,nsv,ismode
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: nsi,nkd
  integer:: nrx,nry,nrz,nhx,nhy,nhz,nrxyz,ntt &
       ,nrcell,nspin,ngx,ngy,ngz,ngxyz &
       ,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz
  integer:: ndens
  real(kind=8):: qf,rf,qm,qc,qwf,qbtp(3)
  real(kind=8):: qbsp
  integer:: nbsf
  integer:: nkr,nkf,nkfxa,nkc,naf,narf,nng,nband,nw,nb1,nb2
  integer:: nac,nam
  !
  integer:: iopt(nopt)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer nsk,nks,nk,nkmesh
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  type (irdkpt_struct),pointer:: irdkpt(:)
  !
  real(kind=8) akk(:)
  integer kkg(:,:,:),iwk(:),kr(:,:)
  allocatable akk,kkg,iwk,kr
  real(kind=8),allocatable:: akm(:,:)
  integer,allocatable:: kg(:,:),kqg(:,:)
  real(kind=8),allocatable:: qgk(:,:)
  integer ikd, i, j, isk
  integer ig
  integer,allocatable:: nkmary(:),nknlary(:)
  integer:: iks
  real(kind=8):: clk
  integer:: mssg,ierr
  real(kind=8):: ekbt
  integer:: nfpot
  real(kind=8):: fwqcom
  integer:: xctp
  real(kind=8),pointer:: zoin(:),znin(:)
  !
  integer:: nproc,myrank,comm
  type (replica_struct):: repls
  !
  real(kind=8),pointer:: udirchr(:,:)
  !
  call MPI_INIT(ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD,nproc,ierr)
  if (nproc.eq.1) then
     call MPI_COMM_DUP(MPI_COMM_WORLD,comm,ierr)
     call MPI_COMM_DUP(MPI_COMM_WORLD,repls%comt,ierr)
     repls%ifreplm = .false.
  else
     call setup_replica(NFINPUT,MPI_COMM_WORLD,comm,repls)
  end if
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_COMM_RANK(comm,myrank,ierr)
  !
  ! inipot should be performed by single MPI proc for each replica.
  !
  if (myrank.ne.0) then
     call MPI_FINALIZE(ierr)
     stop
  end if

  clk=xetime(0)
  write(6,*)'INIPOT'
  mssg=1
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdinidata(ax,bx,aa,bb,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz,ntt,ngx,ngy,ngz,ngxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,ekbt &
       ,NFINPUT,mssg &
       ,qbsp,nbsf,fwqcom,xctp,zoin,znin)

  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc('INIPOT','bspots',ierr)

  call potheader(xctp,nkd,zoin,znin,NFPOT1,bspots)
  !
  call potfitc(4,qbsp,nbsf,bspotfc)
  !
  call fitlpt(bspotfc,nkd,NFPOT1,bspots)
  call prntime('FITLPT',mssg)
  !
  do ikd=1,nkd
     nfpot=NFPOT1-1+ikd
     call fitnlpot(nfpot,bspotfc, &
          bspots(ikd),fwqcom,qbtp(2),qc)
  end do

  call prntime('FITNLPOT',mssg)

  call fitachg(bspotfc,nkd,NFRHI1,bspots)

  call prntime('FITACHG',mssg)

  call fitionchg(bspotfc,nkd,NFIONC1,bspots)

  call prntime('FITIONCHG',mssg)

  allocate(udirchr(3,nsi))
  udirchr = 0.0d0
  call chkall(ng,rg,pg,srg,nnp,aa,nsi,kd,asi,ndens,udirchr,MPI_COMM_SELF)
  deallocate(udirchr)
  !
  !     narf: maximum number of lattice vectors in the real space
  !
  narf=int(8.0d0*(rf**3)/va) + 1
  allocate(kr(3,narf),stat=ierr)
  call chkalloc('INIPOT2','rarr',ierr)
  call rarr(kr,nkr,rf,rg,ng,nng,aa &
       ,narf,idat,nk,mmm,nsk,mssg)
  call prntime('RARR2  ',mssg)
  !
  call meshk(bb,ng,rg,nnp,pg,srg,nk,nsk,mmm,nkr,kr,nks,nkmesh,irdkpt)
  call prntime('MESHK  ',mssg)
  deallocate(kr)
  !
  call ymshr(rg,pg,ng,nnp,nrx,nry,nrz,nrxyz,ntt,nrcell)
  call prntime('YMSHR  ',mssg)
  !
  ! naf: maximum number of plane waves to expand potentials
  !
  naf=(2*ngx+1)*(2*ngy+1)*(2*ngz+1)
  allocate(akk(naf),kkg(3,naf,ng*2),iwk(naf),stat=ierr)
  call chkalloc('INIPOT','ykarr2',ierr)
  call ykarr2(akk,kkg,iwk,nkf,qf,rg,pg &
       ,ng,nnp,bb,naf,ngx,ngy,ngz,nvx,nvy,nvz,nvxyz)
  call prntime('YKARR2 ',mssg)
  !
  call ggenpw(akk,kkg,iwk,nrx,nry,nrz &
       ,nkfxa,naf,nkf,ng)
  call prntime('GGENPW ',mssg)
  !
  ! nac: maximum number of plane waves to expand core charge
  !
  nac=(2*nrx+1)*(2*nry+1)*(2*nrz+1)
  allocate(qgk(4,nac),kqg(3,nac),stat=ierr)
  call chkalloc('INIPOT','qgk,kqg',ierr)
  !
  call qarr3(qgk,kqg,bb,nkc,qc,nac,nrx,nry,nrz)
  call xrewind(NFQG)
  call wrkg(kqg,nkc,NFQG)
  call prntime('QARR   ',mssg)
  !
  ! nam: maximum number of plane waves to expand wave functions
  nam=8*nhx*nhy*nhz
  allocate(nkmary(nks),nknlary(nks),akm(nam,nks), &
       kg(3,nam),stat=ierr)
  call chkalloc('INIPOT','nkmary,nknlary,akm,kg',ierr)

  call xrewind(NFNKMKG)
  write(6,'(1x,a)') 'iks, nkm ='
  do iks=1,nks
     !
     ! ORDERING THE RECIPROCAL LATTICE VECTOR
     !
     !   branch here to make kg agree with the kg of no real space
     !   projection version.
     !
     !
     !   make plane wave basis for wave function (nkm)
     !
     call ykarr3(akm(1,iks),kg,irdkpt(iks)%sk,bb,nkmary(iks), &
          nknlary(iks),qm,qm,nam,nhx,nhy,nhz)
     call wrkg(kg,nkmary(iks),NFNKMKG)
     write(6,'(3x,i3,2(1x,i6))') iks,nkmary(iks)
  end do
  call prntime('YKARR3 ',mssg)

  deallocate(nkmary,nknlary)
  !
  deallocate(akm,kg)
  deallocate(qgk,kqg)
  deallocate(akk,kkg,iwk)
  !
  call xrewind(NFPOTS)
  write(NFPOTS) nkd, nsi, aa, qf, rf, qm, qc, ng, nnp
  write(NFPOTS) (((rg(i,j,ig),i=1,3),j=1,3),ig=1,ng)
  write(NFPOTS) ((pg(i,ig),i=1,3),ig=1,ng),(srg(ig),ig=1,ng)
  write(NFPOTS) nk, nsk
  write(NFPOTS) (((mmm(i,j,isk),i=1,3),j=1,2),isk=1,nsk)
  write(NFPOTS) (idat(i),i=1,3)
  write(NFPOTS) nvx,nvy,nvz,nvxyz
  write(NFPOTS) nng,nrcell,nkr,nkfxa
  write(NFPOTS) nkc
  call store_irdkpt(NFPOTS,nkmesh,nks,irdkpt)
  !
  call storebspot(NFBSPOT,nkd,xctp,bspotfc,bspots)

  deallocate(kd,asi,mmm,zoin,znin)

  clk=xetime(1)
  write(6,*) 'TOTAL TIME=',clk
  call MPI_FINALIZE(ierr)
  stop
contains
  subroutine wrkg(kg,nkm,nfile)
    implicit none
    integer nkm,kg(3,nkm),nfile
    write(nfile) nkm
    write(nfile) kg
    return
  end subroutine wrkg
end program inipot
