!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module iounit
  implicit none

  integer:: NFCHR
  parameter(NFCHR=25)
  ! to store total CHaRge

  integer:: NFINPUT
  parameter(NFINPUT=10)
  ! input data for parallel program

  integer:: NFLPT
  parameter(NFLPT=11)
  ! to store Local PoTentials

  integer:: NFNSYKZ
  parameter(NFNSYKZ=13)
  ! irreducible real lattice point data.
  ! nsy(oribt): # of lattice point in each orbit.
  ! kz([x,y,z],orbit,operation):
  !   real lattice points classifid under the oribits.

  integer:: NFNWK
  parameter(NFNWK=14)
  ! irreducible FFT mesh point data.
  ! nwk(1:nrxyz,1:3): mesh points grouped into the each orbit.
  !   The order is {1, ..., nwk(1)}, {nwk(1)+1, ..., nwk(1)+nwk(2)}, ...
  ! idcwk(1: # of orbits, 1): index to the head of each orbit in nwk.
  ! idcwk(1: # of orbits, 2): # of mesh points belongs to the orbit.
  ! used in ysymch.

  integer:: NFKF
  parameter(NFKF=15)
  ! aff(nkfxa), kf(3,nkfxa) and ind(nkfxa) (plane wave for local potential)
  ! generated in ggenpw are stored.
  ! ind associates aff with akk.

  integer:: NFRHI1,NFRHIF
  parameter(NFRHI1=28,NFRHIF=33)
  ! to store the 1st atomic charge of the pseudo-potential.
  ! remaining atomic charges are assigned to
  ! nfrhi1, nfrhi1+1, nfrhi+2, ...

  integer:: NFPOT1,NFPOTF
  parameter(NFPOT1=34,NFPOTF=39)
  ! to store the 1st pseudo POTential. remaining PPs are
  ! assigned to nfpot1, nfpot1+1, nfpot1+2, ...

  integer:: NFIONC1,NFIONCF
  parameter(NFIONC1=151,NFIONCF=156)
  ! to store the 1st atomic charge of the pseudo-potential.
  ! remaining atomic charges are assigned to
  ! nfionc1, nfionc1+1, nfionc+2, ...

  integer:: NFQG
  parameter(NFQG=40)
  ! to store QG(variable name)

  integer:: NFPE
  parameter(NFPE=51)
  ! to store nb1, nb2, nkr, nng, pe, kz, nsy
  !  : cosine fitting data of band. passed to pefcos.

  integer:: NFWFI
  parameter(NFWFI=95)
  ! for Wave Function to Input

  integer:: NFWFO
  parameter(NFWFO=96)
  ! for Wave Function to Output

  integer:: NFRSL
  parameter(NFRSL=99)
  ! to store ReSuLts of optimization

  integer:: NFSAMPK
  parameter(NFSAMPK=21)
  ! sk(sample k point) and rco(\sum_{k in orbit c(sk)}\cos(k\cdot R))
  ! are stored.

  integer:: NFNKMKG
  parameter(NFNKMKG=22)
  ! nkm,kg(3,nkm) generated in vbnlp for each sample
  ! k points are stored in the same order of the sample
  ! k points.

  integer:: NFACGD
  parameter(NFACGD=70)
  ! atom cg data stored. used for chain job.

  integer:: NFLMD
  parameter(NFLMD=71)
  ! linmin data stored. used for chain job.

  integer:: NFSCFD
  parameter(NFSCFD=72)
  ! scf data stored. used for chain job.

  integer:: NFPOTS
  parameter(NFPOTS=73)
  ! atomic potential data stored. used for recycling of the initialization.

  integer:: NFBSPOT
  parameter(NFBSPOT=74)
  ! atomic potential data fitted with B-spline

  integer:: NFMDCNT
  parameter(NFMDCNT=75)
  ! data to continue molecular dynamics

  integer:: NFMDMBMT
  parameter(NFMDMBMT=76)
  ! data for multi-baric multi-thermal MD

  integer:: NFUMBMT
  parameter(NFUMBMT=77)
  ! updated data for multi-baric multi-thermal MD

  integer:: NFSMBMT
  parameter(NFSMBMT=78)
  ! snap data for multi-baric multi-thermal MD

  integer:: NFMDLOG
  parameter(NFMDLOG=16)
  ! log of molecular dynamics

  integer:: NFMDSMPL
  parameter(NFMDSMPL=17)
  ! sampled configurations

  integer:: NFVBCHRGWAV
  parameter(NFVBCHRGWAV=18)

  integer:: NFIDXPAR
  parameter(NFIDXPAR=19)
  
  integer:: NFRECIPWAN
  parameter(NFRECIPWAN=113)

  integer:: NFAMATWAN
  parameter(NFAMATWAN=150)

  integer:: NFHMATRWAN
  parameter(NFHMATRWAN=122)

  integer:: NFREALWAN
  parameter(NFREALWAN=116)

  integer:: NFPOSWAN
  parameter(NFPOSWAN=134)

  integer:: NFEXCOR
  parameter(NFEXCOR=141)

  integer:: NFHRTOR
  parameter(NFHRTOR=142)

  integer:: NFATMCHRD
  parameter(NFATMCHRD=143)

  integer:: NFEKINOR
  parameter(NFEKINOR=144)

  integer:: NFSEKINOR
  parameter(NFSEKINOR=145)
  
  integer:: NFSHRTOR
  parameter(NFSHRTOR=146)

  integer:: NFEVXCOR
  parameter(NFEVXCOR=147)

  integer:: NFLOCENG
  parameter(NFLOCENG=148)

  integer:: NFCAR2XTP
  parameter(NFCAR2XTP=149)

  integer,parameter:: nfstm = 80
  integer,parameter:: nfbnd = 50
  integer,parameter:: nfout = 55
  integer,parameter:: nfrok = 57
  integer,parameter:: nfwfk = 58
  integer,parameter:: nfoutkmsh=59

  character(len=256):: iou_bnam
  character(len=256):: iou_rhinam(1:NFRHIF-NFRHI1+1)
  character(len=256):: iou_potnam(1:NFPOTF-NFPOT1+1)
  character(len=256):: iou_ioncnam(1:NFIONCF-NFIONC1+1)
  logical:: iou_fmap

contains
  subroutine xrewind(nf)
    implicit none
    integer,intent(in):: nf

    logical:: lst
    integer:: ipot, ios
    character(len=256):: fname

    inquire(unit=nf, opened=lst)
    if (lst) then
       rewind(nf)
       return
    end if

    if (iou_fmap) then
       select case (nf)
       case(NFCHR)
          fname = trim(iou_bnam)//'.rho'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFLPT)
          fname = trim(iou_bnam)//'.lpt'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFNSYKZ)
          fname = trim(iou_bnam)//'.nsykz'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFNWK)
          fname = trim(iou_bnam)//'.nwk'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFKF)
          fname = trim(iou_bnam)//'.kf'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFRHI1:NFRHIF)
          ipot = nf-NFRHI1+1
          fname = iou_rhinam(ipot)
          if (trim(fname) .ne. 'NONE') then
             open(nf, file=fname, iostat=ios)
          end if
       case(NFPOT1:NFPOTF)
          ipot = nf-NFPOT1+1
          fname = iou_potnam(ipot)
          open(nf, file=fname, iostat=ios)
       case(NFIONC1:NFIONCF)
          ipot = nf-NFIONC1+1
          fname = iou_ioncnam(ipot)
          if (trim(fname) .ne. 'NONE') then
             open(nf, file=fname, iostat=ios)
          end if
       case(NFQG)
          fname = trim(iou_bnam)//'.qg'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFPE)
          fname = trim(iou_bnam)//'.pe'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFWFI)
          fname = trim(iou_bnam)//'.wfn'
          inquire(NFWFO,opened=lst)
          if (lst) close(NFWFO)
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFWFO)
          fname = trim(iou_bnam)//'.wfn'
          inquire(NFWFI,opened=lst)
          if (lst) close(NFWFI)
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFRSL)
          fname = trim(iou_bnam)//'.str'
          open(nf, file=fname, iostat=ios)
       case(NFSAMPK)
          fname = trim(iou_bnam)//'.sampk'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFNKMKG)
          fname = trim(iou_bnam)//'.nkmkg'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFACGD)
          fname = trim(iou_bnam)//'.acgd'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFLMD)
          fname = trim(iou_bnam)//'.lmd'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFSCFD)
          fname = trim(iou_bnam)//'.scfd'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFPOTS)
          fname = trim(iou_bnam)//'.pots'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFBSPOT)
          fname = trim(iou_bnam)//'.bspot'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFMDCNT)
          fname = trim(iou_bnam)//'.mdcnt'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFMDLOG)
          fname = trim(iou_bnam)//'.mdlog'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFVBCHRGWAV)
          fname = trim(iou_bnam)//'.pwav'
          open(nf, file=fname, iostat=ios)
       case(NFIDXPAR)
          fname = trim(iou_bnam)//'.idxpar'
          open(nf, file=fname, iostat=ios)
       case(nfstm)
          fname = trim(iou_bnam)//'.stm'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(nfbnd)
          fname = trim(iou_bnam)//'.band'
          open(nf, file=fname, iostat=ios)
       case(nfout)
          fname = trim(iou_bnam)//'.bunpu'
          open(nf, file=fname, iostat=ios)
       case(nfrok)
          fname = trim(iou_bnam)//'.rok'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(nfwfk)
          fname = trim(iou_bnam)//'.wfk'
          open(nf, file=fname, iostat=ios, form='unformatted')

       case(NFRECIPWAN)
          fname = trim(iou_bnam)//'.rcpw'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFAMATWAN)
          fname = trim(iou_bnam)//'.amat'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFHMATRWAN)
          fname = trim(iou_bnam)//'.hmatr'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFREALWAN)
          fname = trim(iou_bnam)//'.rwan'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFPOSWAN)
          fname = trim(iou_bnam)//'.posw'
          open(nf, file=fname, iostat=ios)
       case(NFEXCOR)
          fname = trim(iou_bnam)//'.excor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFHRTOR)
          fname = trim(iou_bnam)//'.hrtor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFATMCHRD)
          fname = trim(iou_bnam)//'.atmcd'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFEKINOR)
          fname = trim(iou_bnam)//'.ekinor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFSEKINOR)
          fname = trim(iou_bnam)//'.sekinor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFSHRTOR)
          fname = trim(iou_bnam)//'.shrtor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFEVXCOR)
          fname = trim(iou_bnam)//'.evxcor'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFLOCENG)
          fname = trim(iou_bnam)//'.loceng'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case(NFCAR2XTP)
          fname = trim(iou_bnam)//'.car2xtp'
          open(nf, file=fname, iostat=ios)
       case(nfoutkmsh)
          fname = trim(iou_bnam)//'.dosms'
          open(nf, file=fname, iostat=ios, form='unformatted')
       case default
          rewind(nf)
       end select
    else
       rewind(nf)
    end if

    return
  end subroutine xrewind

end module iounit
