!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
!
module localdos
contains
  subroutine calefldos(ndx, ndy, ndz, nb1, nb2, nkr, &
       nng, nks, nkmesh, ncomp, nspin, nrx, nry, nrz, nrxyz, &
       ng, nnp, rg, pg, srg, nrcell, ldkg, laprj, nkd, nsi, &
       nband, nw, bkc, ekc, mssg, &
       nkm, kg, nwk, idcwk, nsy, kz, kqg, dqgs, &
       ekbt, rco, bzwk, va, aa, asi, pby, &
       kd, ee, ef, delef, efldos, dp, bspots, irdkpt, &
       ifinvsym, myrank, nproc, g_count, g_displ, comm)
    use bsplpot
    use uspp_cr
    use bandchrg
    use types
    use soccup_num
    use util
    use subr_kpoint
    implicit none
    integer,intent(in):: ndx, ndy, ndz, nb1, nb2, nkr
    integer,intent(in):: ncomp
    integer,intent(in):: nng, nks, nkmesh, nspin, nrx, nry, nrz, nrxyz
    integer,intent(in):: ng, nrcell, ldkg, laprj, nkd, nsi
    integer,intent(in):: nnp,rg(3,3,ng),pg(3,ng),srg(ng)
    integer,intent(in):: nband, nw, bkc, ekc, mssg
    integer,intent(in):: nkm(nks), kg(3,ldkg,nks)
    integer,intent(in):: nwk(nrxyz,3), idcwk(nrcell,2)
    integer,intent(in):: nsy(nkr), kz(3,nkr,nng), kqg(3,bkc:ekc)
    type(qg_for_atm):: dqgs(nkd)
    real(kind=8),intent(in):: ekbt
    real(kind=8),intent(in):: rco(nkr,nks), bzwk(nks), va, aa(3,3)
    real(kind=8),intent(in):: asi(3,nsi)
    complex(kind=8),intent(in):: pby(laprj,nsi,nw,nks,nspin)
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: ee(nband,nks,nspin)
    real(kind=8),intent(in):: ef, delef
    type (bspot_for_atm):: bspots(nkd)
    type (irdkpt_struct),pointer:: irdkpt(:)
    logical,intent(in):: ifinvsym

    real(kind=8),intent(out):: efldos(8*nrx*nry*nrz,nspin)
    type (wfn_for_k):: dp(nks,nspin)

    integer:: myrank,nproc,comm
    integer:: g_count(0:nproc-1,nks),g_displ(0:nproc-1,nks)

    real(kind=8),parameter:: eps = 1.0d-14

    real(kind=8),allocatable:: pedef(:,:,:),wgtkdef(:,:,:),wgtkdefnz(:,:,:)
    real(kind=8),allocatable:: crho(:,:)

    real(kind=8):: efpdel, efmdel, pwchrg, crchrg, tchrg
    integer:: ierr
    integer:: nb1d, nb2d, iks, is

    allocate(pedef(nkr,nb1:nb2,nspin),wgtkdef(nb1:nb2,nks,nspin), &
         crho(8*nrx*nry*nrz,nspin),stat=ierr)
    call chkalloc('efldos','pedef,wgtkdef',ierr)

    efpdel = ef + delef
    efmdel = ef - delef
    call sdoss(nband,ee,rco,nkmesh,pedef,efpdel,efmdel,ndx,ndy,ndz &
         ,nb1,nb2,nkr,nng,nsy,kz,nks,ncomp,nspin,ekbt,wgtkdef,mssg)

    nb1dloop: do nb1d=nb1,nb2
       do is=1,nspin
          do iks = 1, nks
             if (wgtkdef(nb1d,iks,is) .gt. eps) then
                exit nb1dloop
             end if
          end do
       end do
    end do nb1dloop
    nb2dloop: do nb2d=nb2,nb1,-1
       do is=1,nspin
          do iks = 1, nks
             if (wgtkdef(nb2d,iks,is) .gt. eps) then
                exit nb2dloop
             end if
          end do
       end do
    end do nb2dloop
    if (mssg.gt.0) then
       write(6,'(1x,a,1x,i3,1x,i3)') 'nb1d, nb2d =',nb1d,nb2d
    end if
    allocate(wgtkdefnz(nb1d:nb2d,nks,nspin),stat=ierr)
    call chkalloc('efldos','wgtkdefnz',ierr)
    wgtkdefnz = wgtkdef(nb1d:nb2d,:,:)

    !call yps3(dp(1,is),efldos(1,is),bzwk,wgtkdefnz,va,nkmesh &
    !     ,nrx,nry,nrz,nrxyz &
    !     ,ng,nrcell,nks,nkm,kg,ldkg,nb1d,nb2d,nwk,idcwk,irdkpt &
    !     ,(ifinvsym .or. xcf_exch_factor(xctp).ne.0.0d0) &
    !     ,0,myrank,nproc,g_count,g_displ,comm)
    call yps3(nspin,dp,efldos,bzwk,wgtkdefnz,va,aa,nkmesh &
         ,ncomp,1,nrx,nry,nrz,nrxyz &
         ,ng,nnp,rg,pg,srg,nrcell,nks,nkm,kg,ldkg,nb1d,nb2d,nwk,idcwk,irdkpt &
         ,.true.,0,myrank,nproc,g_count,g_displ,comm)
    call crgen(nspin,bkc,ekc,kqg,pby,1,crho &
         ,bzwk,wgtkdefnz,nsi,nkd,kd &
         ,laprj,ng,nnp,rg,pg,srg,aa,nrcell,nkmesh,nks,asi &
         ,nrx,nry,nrz,nrxyz,nb1d,nb2d,nw,nwk,idcwk &
         ,dqgs,bspots,0,mssg,comm)
    do is=1,nspin
       call calchr(efldos(1,is),nrxyz,va,pwchrg)
       call calchr(crho(1,is),nrxyz,va,crchrg)
       tchrg = pwchrg + crchrg
       if (mssg.gt.0) write(6,500)tchrg,crchrg,pwchrg
500    format(' TOTAL STATE=',f25.20,' CORE STATE=',f25.20/, &
            39x,' PW   STATE=',f25.20)
       efldos(:,is) = efldos(:,is) + crho(:,is)
    end do

    deallocate(wgtkdefnz)
    deallocate(pedef,wgtkdef,crho)
    efldos = efldos/(efpdel - efmdel)
  end subroutine calefldos
end module localdos
