!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
!**********************************************************************
!   1993-09-02 AA-VERSION
!   1993-09-03 CUBIC INTERPOLATION ADDED
!   1993-09-03 CHECK STEP RENEWAL
!   1993-09-17 MINOR CHANGE IN DDMAX
!   1993-09-24 CHSTEP & CASE(ALPH1 < ALPH2) CHANGED
!
! LINEAR MINIMIZATION ALONG TO H-VECTOR
!                      BY SECANT & BISECTION METHOD
!
!     MOST : THE MAXIMUM # OF LINMIN-LOOP
!     IEXIT:   0     CONVERGED
!             -1 NOT CONVERGENCE WITHIN MOST-TIMES
!     ISCF :  # OF CALLING SCF-LOOP
!     NCSCF:  # OF CALLING VBSCF
!     IFAR : FLAG OF EXISTENCE OF UPPER BOUNDARY
!       IFAR=1 => SOLUTION EXISTS IN (AL(1),AR(1))
!
!     A?(1):POSITION PARAMETER IN ONE DIMENSION
!     A?(2):TOTAL ENERGY AT A?(1)
!     A?(2):FORM*H       AT A?(1)
!
!
!     ALL(1)<AL(1)<AR(1) & ALL(2)<AL(2) & ALL(3),AL(3)>0
!     & (AR(3)<0 OR AR(2)>AL(2))
!
!**********************************************************************
module linear_minim
  private parmin, cbmin, cnvch1
contains
  subroutine linmin(asi,cdm,h,hcdm,most,iexit,scfexit,iscf,ncscf &
       ,alpha,okstep,decr,feps,eeps,stscl,syspext,optcons &
       ,lpt,rho,vxcr,excor,evxcor,hrt,crho,pby,dp &
       ,qbtp &
       ,nopt,iopt,ismode,ekbt,etotal,eleng,ewald,sewald &
       ,forel,dforel,for,dfor,force,dforce,dform,setotm &
       ,ef,eetrp,setotal,elseng,engatm,sengatm &
       ,nband,laprj,nsi,nkd,rg,pg,ng,nnp &
       ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
       ,aa0,kd,ee,pe,wgtk,dd &
       ,bspotfc,bspots &
       ,scfcs,mgs,ifinvsym,mssg,ifchlin,axetp &
       ,myrank,nproc,comm)
    use chain_linmin
    use types
    use mpi
    use util
    use scfloop
    use engdffr
    use bsplpot
    use meshgrid
    use ewald_sum
    use stropt_cnstr
    implicit none
    integer:: most
    integer:: iexit,iscf,ncscf
    integer:: nopt,ismode,nband,laprj,nsi,nkd,ng,nnp
    integer:: ncomp,nspin,ndens
    integer:: nw,nb1,nb2
    integer:: mssg,ifchlin
    !      
    real(kind=8):: etotal
    real(kind=8):: ewald,sewald(6)
    type(elec_energy):: eleng
    real(kind=8):: ef(nspin),eetrp
    real(kind=8):: setotal(6)
    type(elec_senergy):: elseng
    type(energy_per_atom):: engatm
    type(senergy_per_atom):: sengatm
    real(kind=8):: znext, dspin
    real(kind=8):: okstep(2), decr, alpha, feps(2), eeps
    real(kind=8):: stscl(6), syspext
    type (optcons_struct):: optcons
    real(kind=8):: ekbt
    !
    type (scfctrl_struct):: scfcs
    type (meshgrid_struct):: mgs
    !
    real(kind=8):: rho(mgs%nrxyz,nspin*ndens),vxcr(mgs%nrxyz,nspin*ndens)
    real(kind=8):: excor(mgs%nrxyz),evxcor(mgs%nrxyz,0:6)
    complex(kind=8):: hrt(mgs%nkfxa)
    real(kind=8):: crho(mgs%nrxyz,nspin*ndens)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    complex(kind=8):: lpt(mgs%nvxyz,nspin*ndens)
    !
    real(kind=8):: aa0(3,3),asi(3,nsi)
    real(kind=8):: cdm(6)
    integer:: kd(nsi)
    real(kind=8):: ee(nband,mgs%nks,nspin),pe(mgs%nkr,nb1:nb2,nspin)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: force(3,nsi),dforce(3,nsi),dform(3,nsi),h(3,nsi)
    real(kind=8):: setotm(6),hcdm(6)
    real(kind=8):: for(3,nsi),dfor(3,nsi),forel(3,nsi),dforel(3,nsi)
    real(kind=8):: qbtp(3)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer:: iopt(nopt)
    real(kind=8):: scfexit(nscfexit)
    integer:: rg(3,3,ng),pg(3,ng)
    type (wfn_for_k):: dp(mgs%nks,nspin)
    logical:: ifinvsym
    integer:: axetp
    integer:: myrank,nproc,comm
    !
    real(kind=8):: aa(3,3),cdm1(6)
    real(kind=8),allocatable:: asi1(:,:)
    real(kind=8):: all(3),al(3),ar(3),arr(3)
    real(kind=8):: safety,safe
    data      safety,safe/0.01d0,0.01d0/
    real(kind=8):: eokstp
    !
    real(kind=8):: grad0,signn,entpy00,alph1,alph2,almax
    real(kind=8):: wd,alsave,f,dif,fmax(2)
    real(kind=8):: entpysave, eminp, eminc, ddmax, temp
    real(kind=8):: xminp, xminc
    integer:: it,i,ib,is,ifar,ini_itlin,itlin,itmax,isafe,icflg
    integer:: iex
    integer:: iexp, iexc
    logical:: tflg
    integer:: ierr

    call cdm2aa(aa0,cdm,aa)
    
    allocate(asi1(3,nsi),stat=ierr)
    call chkalloc('linmin','asi1',ierr)

    ! IF    ABS(AR-ALPH1)/ABS(AR-AL)<SAFETY
    !    OR.ABS(ALPH1-AL)/ABS(AR-AL)<SAFETY THEN
    !        SET   ALPH1=AL+SAFE*(AR-AL)
    !
    if (mssg.gt.-1) then
       write(6,*)'########## LINEAR MINIMIZATION START ##########'
    end if
    iexit=0
    iscf=0
    ncscf=0
    ! DIRECT H(N) ALONG TO DFORM(N)
    grad0=sekiaa(aa,h,dform,nsi,hcdm,setotm)
    !
    signn=1.0d0
    if (grad0.lt.0.0d0) then
       signn=-1.0d0
       if (mssg.gt.-1) write(6,*)' SIGNN=-1 !!!!!!!!!!'
    end if

    h = signn*h
    hcdm = signn*hcdm
    grad0=grad0*signn

    entpy00=etotal+syspext*cellvol(aa)
    if (ifchlin .ne. 1) then
       ifar=0
       alph1=0.0d0
       alph2=0.0d0
       do i=1,3
          all(i)=-1.d15
          ar (i)= 1.d15
          arr(i)= 1.d15
       end do
       al(1)=0.0d0
       al(2)=etotal+syspext*cellvol(aa)
       al(3)=grad0
       ini_itlin=1
       scfcs%ifchscf=0
    else
       if (mssg.gt.-1) write(6,*) 'loading linmin data ...'
       call load_linmin_data(ifar,ini_itlin,alph1,alph2,etotal, &
            okstep,all,al,ar,arr,myrank,comm)
       call MPI_BARRIER(comm,ierr)
       ifchlin=0
       if ((iopt(2) .eq. 1) .and. (iopt(3) .eq. 1)) then
          scfcs%ifchscf=1
       else
          if (mssg.gt.-1) then
             write(6,*)'local potential or wave function ', &
                  'is not available.'
             write(6,*)'Do not chain scf.'
          end if
          scfcs%ifchscf=0
       end if
    end if
    !
    do itlin=ini_itlin,most
       if (myrank.eq.0) then
          call store_linmin_data(ifar,itlin,alph1,alph2,etotal, &
               okstep,all,al,ar,arr)
       end if
       ncscf=ncscf+1
       entpysave=etotal+syspext*cellvol(aa)
       if (mssg.gt.-1) then
          write(6,*)'****** # OF SCF & CALLING VBSCF IN LINMIN=' &
               ,iscf,ncscf
       end if
       !
       ! DETERMINE NEXT TRIAL POSITION
       xminp=0.0d0
       eminp=0.0d0
       iexp=0
       xminc=0.0d0
       eminc=0.0d0
       iexc=0
       ddmax=1000.d0
       if (itlin.eq.1) then
          alph1=alpha
       else if (ifar.eq.0) then
          ! FACTOR
          okstep=okstep*2.0d0
          alph2=alph1
          ! SECANT BY ALL & AL
          call parmin(all,al,xminp,eminp,iexp)
          !            ALPH1=-(ALL(3)*AL(1)-AL(3)*ALL(1))/(AL(3)-ALL(3))
          ! CUBIC  BY ALL & AL
          call cbmin(all,al,xminc,eminc,iexc)
          !
          ! XXX: cbmin() is not appropriate for the force inversion method
          !      because it depends on not only forces but also energies.
          !
          ! ADOPT PARABOLA FIT
          alph1=xminp
          ! ADOPT CUBIC    FIT
          ! ALPH1=XMINC
          !
          temp=alph1-alph2
          call chstep(aa,h,hcdm,nsi,okstep,temp,ddmax,almax,itmax,eokstp)
          if (mssg.gt.-1) then
             write(6,90)ddmax,itmax,eokstp,temp,almax
          end if
90        format(' ','MAXIMUM DISPLACEMENT(A.U.)  & ATOM=',e15.7,i5/ &
               ' ','OKSTEP, ALPH1-ALPH2,ALMAX:',3e15.7)
          ! BRANCH
          if (ddmax.gt.eokstp) then
             if (mssg.gt.-1) write(6,100)alph1,alph2,almax
100          format(' ','ALPH1 FROM AL & ALL BY SECANT:',e15.7/ &
                  , ' ','ALPH2(LAST STEP)             :',e15.7/ &
                  ,' ','ALPH1 IS TOO LARGE & REPLACED BY ALMAX',e15.7)
             alph1=alph2+almax
             ddmax=eokstp
          else if (alph1.lt.alph2) then
             temp=abs(al(1)-all(1))*10
             if (mssg.gt.-1) write(6,110)alph1,alph2,temp
110          format(' ','ALPH1 FROM AL & ALL BY SECANT:',e15.7/ &
                  , ' ','ALPH2(LAST STEP)             :',e15.7/ &
                  ,' ','ALPH1 IS IRREGULAR & REPLACED BY ALMAX',e15.7)
             alph1=alph2+temp
             ddmax=temp/almax*eokstp
          end if
       else
          alph2=alph1
          ! SECANT    METHOD USING AL & AR
          call parmin(al,ar,xminp,eminp,iexp)
          !            ALPH1=-(AL(3)*AR(1)-AR(3)*AL(1))/(AR(3)-AL(3))
          ! CUBIC     METHOD  BY AL & AR
          call cbmin(al,ar,xminc,eminc,iexc)
          !
          ! XXX: cbmin() is not appropriate for the force inversion method
          !      because it depends on not only forces but also energies.
          !
          ! ADOPT PARABOLA FIT
          alph1=xminp
          ! ADOPT CUBIC    FIT
          ! ALPH1=XMINC
          if (.not.((alph1.gt.al(1)).and.(alph1.lt.ar(1)))) then
             ! BISECTION METHOD USING AL& AR
             alph1=0.5d0*(al(1)+ar(1))
          end if
          temp=alph1-alph2
          call chstep(aa,h,hcdm,nsi,okstep,temp,ddmax,almax,itmax,eokstp)
          if (mssg.gt.-1) write(6,90)ddmax,itmax,eokstp,temp,almax
          ! SAFETY FOR STABLE CONVERGENCE
          wd=abs(ar(1)-al(1))
          isafe=0
          alsave=alph1
          if (abs(alph1-al(1)).lt.safety*wd) then
             isafe=1
             alph1=al(1)+safe*(ar(1)-al(1))
          else if (abs(ar(1)-alph1).lt.safety*wd) then
             isafe=1
             alph1=ar(1)-safe*(ar(1)-al(1))
          end if
          if (isafe.eq.1) then
             ddmax=ddmax*abs((alph1-alph2)/temp)
             if (mssg.gt.-1) write(6,140)safety,safe,alsave,alph1
140          format(' ',' SAFETY MECHANISM IS APPLIED. SAFETY & SAFE=' &
                  ,2e15.7/ &
                  ,' ',' ALPH1:',e15.7,'   IS REPLACED BY',e15.7)
          end if
       end if
       if (mssg.gt.-1) write(6,150)
150    format(14x,'POSISION',9x,'ENTHALPY',9x,'FORCE PROJECTION')
       if (mssg.gt.-1) write(6,160)(all(i),i=1,3),(al (i),i=1,3) &
            ,(ar (i),i=1,3),(arr(i),i=1,3) &
            ,alph1,xminp,eminp,iexp,xminc,eminc,iexc
160    format(' ','ALL(1-3):',3e17.9/,' ','AL (1-3):',3e17.9/ &
            ,' ','AR (1-3):',3e17.9/,' ','ARR(1-3):',3e17.9/ &
            ,' ','ALPH1   :', e17.9/ &
            ,' ','XMIN PAR:', e17.9,' ','EXPECTED ENTHALPY:',e17.9 &
            ,' IEX:',i3/ &
            ,' ','XMIN CUB:', e17.9,' ','EXPECTED ENTHALPY:',e17.9 &
            ,' IEX:',i3/)
       !
       ! CALCULATE TOTAL ENERGY & FORCE AT NEW TRIAL POINT ALPH1
       do it=1,nsi
          do i=1,3
             asi1(i,it)=modulo(asi(i,it)+alph1*h(i,it),1.0d0)
          end do
       end do
       cdm1 = cdm + alph1*hcdm
       call symstress(ng,rg,nnp,pg,aa0,cdm1)
       call cdm2aa(aa0,cdm1,aa)
       !
       call ewalds(asi1,aa,nsi,nkd,kd &
            ,for,dfor,ewald,sewald,engatm%ewald,sengatm%sewald &
            ,bspots,myrank,nproc,comm,mssg)
       call prntime('EWALDS ',mssg)
       !
       call vbscf(dp,rho,pby,lpt,vxcr,excor,evxcor,hrt,crho &
            ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
            ,ef,eetrp,nband,laprj,nsi,nkd,ng &
            ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
            ,aa,asi1,kd,ee,pe,wgtk,dd &
            ,scfcs,mgs,bspotfc,bspots,ifinvsym &
            ,mssg,myrank,nproc,comm)
       !
       if (iex .eq. -1) then
          iscf=iscf+scfcs%niter
       else if (iex .eq. -2) then
          if (mssg.gt.-1) then
             write(6,*) 'vbscf has exited because of uptime.'
          end if
          iexit=-2
          deallocate(asi1)
          return
       else if (iex .ge. 0) then
          iscf=iscf+iex
       else
          write(6,*) 'unknown result code from vbscf'
          write(6,*) 'iex = ',iex
          stop
       end if
       if (mssg.gt.-1) write(6,250)iex,scfexit(1),scfexit(2)
250    format(' ',' IEX=',i5,3x,'DIFFERENCE OF TOTAL ENERGY=',e15.7/ &
            ,' ',13x,'SQUARE ERROR OF LOCAL POT.=',e15.7)
       if (mssg.gt.-1) then
          write(6,'(1x,a)') 'IB, PE(IB,1) = '
          do ib=nb1,nb2
             write(6,'(1x,i4,1x,e15.7,1x,e15.7)') &
                  ib, (pe(1,ib,is),is=1,nspin)
          end do
       end if
       !
       call vbfor(dp,lpt,rho,vxcr,crho &
            ,wgtk,eleng,forel,dforel,for,dfor,force,dforce &
            ,qbtp,elseng,engatm,sengatm &
            ,aa,asi1,kd,dd,ee,nband,laprj &
            ,rg,pg,ng,nnp,ncomp,nspin,ndens,nsi,nkd &
            ,nb1,nb2,mgs,bspotfc,bspots,scfcs%xctp,axetp &
            ,myrank,nproc,comm)
       !
       call dforce_cnstr(aa,nsi,dforce,dform,optcons)
       !
       call vtran(dforce,aa,nsi)
       etotal = eleng%total + ewald
       setotal = elseng%total + sewald
       setotm = setotal
       setotm(1:3) = setotm(1:3) - syspext*cellvol(aa)
       setotm = stscl*setotm/nsi
       !
       call prntime('FORCE  ',mssg)
       if (mssg.gt.-1) then
          call prnatmst(aa,nsi,asi1,dforce,dform,etotal,setotal, &
               eleng,nspin,ef,ewald)
       end if
       !
       f=sekiaa(aa,h,dform,nsi,hcdm,setotm)
       if (mssg.gt.-1) write(6,450)f,etotal+syspext*cellvol(aa)
450    format(' ','DFORM*H, ENTHALPY=',e15.7,1x,e18.10)
       !
       if (optcons%mode.eq.2) then
          tflg = f.le.0.d0
       else
          tflg = (etotal+syspext*cellvol(aa).gt.al(2)).or.(f.le.0.d0)
       end if

       if (tflg) then
          ifar=1
          do i=1,3
             arr(i)=ar(i)
          end do
          ar(1)=alph1
          ar(2)=etotal+syspext*cellvol(aa)
          ar(3)=f
       else
          do i=1,3
             all(i)=al(i)
          end do
          al(1)=alph1
          al(2)=etotal+syspext*cellvol(aa)
          al(3)=f
       end if
       !
       ! CONVERGENCE CHECK & LOOP EXIT
       dif = etotal + syspext*cellvol(aa) - entpysave
       call cnvch1(icflg,fmax,h,hcdm,aa,dform,setotm,grad0,dif,decr &
            ,ddmax,feps,eeps,nsi,mssg)
       if (mssg.gt.-1) write(6,600)itlin,fmax(1),fmax(2)
600    format(' ','######## ITLIN=',i4,'   MAX(FORCE; STRESS) =', 2e13.5)
       if (icflg.ge.0) then
          if (optcons%mode.eq.2) then
             tflg = .true.
          else
             tflg = etotal+syspext*cellvol(aa) .le. entpy00
          end if
          if (tflg) then
             iexit=icflg
             goto 9900
          else
             if (mssg.gt.-1) write(6,650)etotal+syspext*cellvol(aa),entpy00
650          format(' ','!!! CAN NOT EXIT.  ENTHALPY & ENTPY00=',2e20.12)
          end if
       end if
    end do
    iexit=-1

9900 continue

    h = signn*h
    hcdm = signn*hcdm
    asi = asi1
    cdm = cdm1

    deallocate(asi1)
    return
  end subroutine linmin
  !********************************************************************
  !
  !         ICFLG: 0: CONVERGENCE DONE
  !                1: RETURN BECAUSE OF TOO SMALL DISPLACEMENT
  !               -1: NOT CONVERGENT
  !********************************************************************
  subroutine cnvch1(icflg,fmax,h,hcdm,aa,dform,setotm,grad0,dif,decr &
       ,ddmax,feps,eeps,nsi,mssg)
    implicit none
    integer:: icflg,nsi
    real(kind=8),parameter:: ddmin=1.0d-8
    real(kind=8):: h(3,nsi),dform(3,nsi),aa(3,3),fmax(2),grad0,dif,decr
    real(kind=8):: hcdm(6),setotm(6)
    real(kind=8):: f(3),ddmax,feps(2),eeps
    integer mssg
    real(kind=8):: temp1,ff,saru
    integer:: it,i,j
    !
    temp1=sekiaa(aa,h,dform,nsi,hcdm,setotm)
    !
    fmax(1)=0.0d0
    do it=1,nsi
       ff=0.0d0
       do i=1,3
          f(i)=0.0d0
          do j=1,3
             f(i)=f(i)+aa(i,j)*dform(j,it)
          end do
          ff=ff+f(i)**2
       end do
       ff=sqrt(ff)
       fmax(1)=max(fmax(1),ff)
    end do
    fmax(2)=0.0d0
    do it=1,6
       if (abs(setotm(it)).gt.fmax(2)) then
          fmax(2) = abs(setotm(it))
       end if
    end do
    saru=grad0*decr
    if (abs(temp1).le.abs(saru)) then
       icflg=0
       if (mssg.gt.-1) then
          write(6,*)' ------- CONVERGENCE ACHIEVED IN LINMIN -------'
          write(6,200)temp1,saru,decr,grad0
       end if
200    format(' ',' H*DFORM & GRAD0*SEARCH_1D_FRATIO=',e25.17,e15.7/ &
            ' ',' SEARCH_1D_FRATIO & GRAD0=',e12.4,e25.17)
    else if (fmax(1).le.feps(1) .and. fmax(2).le.feps(2)) then
       icflg=0
       if (mssg.gt.-1) then
          write(6,*)' ------- CONVERGENCE ACHIEVED IN LINMIN -------'
          write(6,300)fmax(1),feps(1)
          write(6,310)fmax(2),feps(2)
       end if
300    format(1x,' FORCE: MAX, CONVERGE =',e15.7,e12.4,' a.u.')
310    format(1x,' STRESS: MAX, CONVERGE =',e15.7,e12.4,' a.u./atom')
    else if (abs(dif).le.eeps) then
       icflg=0
       if (mssg.gt.-1) then
          write(6,*)' ------- CONVERGENCE ACHIEVED IN LINMIN -------'
          write(6,400)dif,eeps
       end if
400    format(' ',' DIFF OF TOTAL ENERGY & CONVERGE_ENERGY =',2e15.7,'Ht')
    else if ( ddmax.le.ddmin) then
       icflg=1
       if (mssg.gt.-1) then
          write(6,*)' ------- DDMAX IS TOO SMALL   IN LINMIN -------'
          write(6,500)ddmin,ddmax
       end if
500    format(' ',' DDMIN & DDMAX=',2e15.7)
    else
       icflg=-1
    end if
    return
  end subroutine cnvch1

  function sekiaa(aa,u,v,n,p,q)
    implicit none
    double precision sekiaa
    integer n
    double precision aa(3,3),u(3,n),v(3,n),p(6),q(6)
    !
    integer it
    double precision x1,y1,z1,x2,y2,z2,a,b,c,temp
    temp=0.0d0
    do it=1,n
       a=u(1,it)
       b=u(2,it)
       c=u(3,it)
       x1=a*aa(1,1)+b*aa(1,2)+c*aa(1,3)
       y1=a*aa(2,1)+b*aa(2,2)+c*aa(2,3)
       z1=a*aa(3,1)+b*aa(3,2)+c*aa(3,3)
       a=v(1,it)
       b=v(2,it)
       c=v(3,it)
       x2=a*aa(1,1)+b*aa(1,2)+c*aa(1,3)
       y2=a*aa(2,1)+b*aa(2,2)+c*aa(2,3)
       z2=a*aa(3,1)+b*aa(3,2)+c*aa(3,3)
       temp=temp+x1*x2+y1*y2+z1*z2
    end do
    do it=1,6
       temp=temp+p(it)*( n*q(it) )
    end do
    sekiaa=temp
    return
  end function sekiaa
  !*************************************************************
  !   RETURN MINIMUM VALUE BY CUBIC INTERPOLATION
  !          F(X)=A*X^3+B*X^2+C*X+D
  !     IF ABS(B/A) > BIG THEN QUADRATIC FORM ASSUMED
  !
  ! INPUT                     OUTPUT
  !   F1,2(1): X                 XMIN:EXPECTED X TO GIVE EMIN
  !       (2): F                 EMIN:EXPECTED MINIMUM ENERGY
  !       (3): FD
  !                           IEX=0 : NORMAL END
  !************************************************************
  subroutine cbmin(g1,g2,xmin,emin,iex)
    implicit real*8(a-h,o-z)
    parameter ( big=1.d3)
    dimension f1(3),f2(3),g1(3),g2(3)
    ! FOR FORECE
    do i=1,3
       f1(i)=g1(i)
       f2(i)=g2(i)
    end do
    f1(3)=-g1(3)
    f2(3)=-g2(3)
    !
    iex=0
    dl=f2(1)-f1(1)
    z=3.d0*(f2(2)-f1(2))/dl-f2(3)-f1(3)
    a=(f1(3)+f2(3)-2.d0*z)/3.0d0/dl/dl
    b=(z-f1(3))/dl
    c=f1(3)
    d=f1(2)
    !
    dsc=z*z-f1(3)*f2(3)
    zz=z-f1(3)
    den=f1(3)+f2(3)-2.d0*z
    if (abs(b/a).gt.big) then
       ! QUADRATIC CASE
       if (b.le.0.d0) then
          !    NO MINIMUM
          iex=1
          x=f1(1)+1.d6*abs(dl)
          emin=0.0d0
       else
          xmin=f1(1)-0.5d0*f1(3)*dl/zz
          emin=d-c*c/4.d0/b
       end if
    else
       ! CUBIC CASE
       if (dsc.le.0.0d0) then
          !     MONOTOMICALLY BEHAVIOR
          if (a.gt.0.0d0) then
             iex=2
             xmin=f1(1)-1.d6*abs(dl)
             emin=0.0d0
          else
             iex=3
             xmin=f1(1)+1.d6*abs(dl)
             emin=0.0d0
          end if
       else
          !     NORMAL CASE
          if (zz.gt.0.0d0) then
             alph1=-dl*(zz+sqrt(dsc))/den
          else
             alph1=-dl*(zz-sqrt(dsc))/den
          end if
          alph2=f1(3)*dl*dl/den/alph1
          !          ALMIN=DMIN(ALPH1,ALPH2)
          !          ALMAX=DMAX(ALPH1,ALPH2)
          almin=min(alph1,alph2)
          almax=max(alph1,alph2)
          if (a.ge.0.0d0) then
             xmin=f1(1)+almax
             x=almax
             emin=a*x**3+b*x**2+c*x+d
          else
             xmin=f1(1)+almin
             x=almin
             emin=a*x**3+b*x**2+c*x+d
          end if
       end if
    end if
    return
  end subroutine cbmin
  !*************************************************************
  !   RETURN MINIMUM VALUE BY PARABOLA
  !          F(X)=0.5*A*X^2+B*X+C
  !
  ! INPUT                     OUTPUT
  !   F1,2(1):  X                XMIN:EXPECTED X TO GIVE EMIN
  !       (2):  F                EMIN:EXPECTED MINIMUM ENERGY
  !       (3):  DF/DX
  !                           IEX=0 : NORMAL END
  !************************************************************
  subroutine parmin(g1,g2,xmin,emin,iex)
    implicit real*8(a-h,o-z)
    parameter ( delta=1.d6)
    dimension f1(3),f2(3),g1(3),g2(3)
    ! FOR FORCE
    do i=1,3
       f1(i)=g1(i)
       f2(i)=g2(i)
    end do
    f1(3)=-g1(3)
    f2(3)=-g2(3)
    !
    a=(f2(3)-f1(3))/(f2(1)-f1(1))
    dl=abs(f2(1)-f1(1))
    if (a .le. 0.0d0) then
       ! LINEAR BEHAVIOR B IS SET TO THE AVERAGE OF F1 & F2(3)
       b=(f1(3)+f2(3))*0.5d0
       iex=1
       if (b.ge.0.0d0) then
          xmin=f2(1)-delta*dl
       else
          xmin=f1(1)+delta*dl
       end if
       emin=0.0d0
    else
       iex=0
       b=f1(3)-a*f1(1)
       xmin=-b/a
       d1=abs(xmin-f1(1))
       d2=abs(xmin-f2(1))
       ! USE VALUE DATA NEAR XMIN TO GET CONSTANT TERM
       if (d1.lt.d2) then
          x=f1(1)
          c=f1(2)-0.5d0*a*x*x-b*x
       else
          x=f2(1)
          c=f2(2)-0.5d0*a*x*x-b*x
       end if
       x=xmin
       emin=0.5d0*a*x*x+b*x+c
    end if
    return
  end subroutine parmin
  !****************************************************************
  !  CHECK STEP SIZE
  !    OUTPUT DDMAX:THE LARGEST DISPLACEMENT IN THIS TRIAL
  !                     IN A.U.
  !           ALMAX:THE PERMITTED MAXIMUM ALPHA
  !           ITMAX:THE NUMBER OF ATOM WITH MAXIMUM DISPLACEMENT
  !****************************************************************
  subroutine chstep(aa,h,hcdm,nsi,okstep,alpha,ddmax,almax,itmax,eokstp)
    implicit none
    integer:: nsi,itmax
    real(kind=8):: aa(3,3),h(3,nsi),hcdm(6),okstep(2),alpha,ddmax,almax
    real(kind=8):: eokstp
    !
    integer:: it,imx(2)
    real(kind=8):: temp,a,b,c,x,y,z,ss
    real(kind=8):: dmx(2),amx(2)
    real(kind=8),parameter:: eps = epsilon(1.0d0)

    imx(1)=0
    temp=0.0d0
    do it=1,nsi
       a=h(1,it)
       b=h(2,it)
       c=h(3,it)
       x=a*aa(1,1)+b*aa(1,2)+c*aa(1,3)
       y=a*aa(2,1)+b*aa(2,2)+c*aa(2,3)
       z=a*aa(3,1)+b*aa(3,2)+c*aa(3,3)
       ss=x*x+y*y+z*z
       ss=ss*alpha*alpha
       if (ss.gt.temp) then
          temp=ss
          imx(1)=it
       end if
    end do
    dmx(1)=sqrt(temp)
    imx(2)=0
    temp=0.0d0
    do it=1,6
       ss = abs(hcdm(it)*alpha)
       if (ss.gt.temp) then
          temp = ss
          imx(2) = it
       end if
    end do
    dmx(2)=temp

    if (dmx(1).gt.eps .and. dmx(2).gt.eps) then
       amx = okstep(:)/dmx(:)*abs(alpha)
    else if (dmx(1).gt.eps .and. dmx(2).lt.eps) then
       amx(1) = okstep(1)/dmx(1)*abs(alpha)
       amx(2) = huge(1.0d0)
    else if (dmx(2).gt.eps .and. dmx(1).lt.eps) then
       amx(2) = okstep(2)/dmx(2)*abs(alpha)
       amx(1) = huge(1.0d0)
    else
       write(6,*)'DDMAX IS TOO SMALL. DDMAX=',dmx
       write(6,*)'ALPHA,ITMAX=',alpha,imx
       dmx = 1.0d2*eps
    end if

    if (amx(1).lt.amx(2)) then
       almax = amx(1)
       ddmax = dmx(1)
       itmax = imx(1)
       eokstp = okstep(1)
    else
       almax = amx(2)
       ddmax = dmx(2)
       itmax = nsi+imx(2)
       eokstp = okstep(2)
    end if
    return
  end subroutine chstep

  !***************************************************************
  !
  !    ICFLG: 0: CONVERGENCE DONE
  !          -1: NOT CONVERGENT
  !***************************************************************
  subroutine cnvch2(icflg,fmax,itmax,aa,dform,setotm, &
       dif,feps,eeps,nsi,mssg)
    implicit none
    integer:: icflg,itmax(2),nsi
    real(kind=8):: fmax(2),dif,feps(2),eeps
    real(kind=8):: dform(3,nsi),setotm(6),aa(3,3)
    real(kind=8):: f(3)
    integer:: mssg
    integer:: it,i,j
    real(kind=8):: ff

    fmax(1)=0.0d0
    itmax(1)=0
    do it=1,nsi
       ff=0.0d0
       do i=1,3
          f(i)=0.0d0
          do j=1,3
             f(i)=f(i)+aa(i,j)*dform(j,it)
          end do
          ff=ff+f(i)**2
       end do
       ff=sqrt(ff)
       if (ff.gt.fmax(1)) then
          fmax(1)=ff
          itmax(1)=it
       end if
    end do
    fmax(2)=0.0d0
    itmax(2)=0
    do i=1,6
       if (abs(setotm(i)).gt.fmax(2)) then
          fmax(2) = abs(setotm(i))
          itmax(2) = i
       end if
    end do
    if (fmax(1).le.feps(1) .and. fmax(2).le.feps(2)) then
       icflg=0
       if (mssg.gt.-1) then
          write(6,*)' ------- CONVERGENCE ACHIEVED IN CGMAIN -------'
          write(6,300)fmax(1),feps(1)
          write(6,310)fmax(2),feps(2)
       end if
300    format(1x,' FORCE: MAX, CONVERGE =',e15.7,e12.4,' a.u.')
310    format(1x,' STRESS: MAX, CONVERGE =',e15.7,e12.4,' a.u./atom')
    else if (abs(dif).le.eeps) then
       icflg=0
       if (mssg.gt.-1) then
          write(6,*)' ------- CONVERGENCE ACHIEVED IN CGMAIN -------'
          write(6,400)dif,eeps
       end if
400    format(' ',' DIF OF TOTAL ENERGY & CONVERGE_ENERGY =',2e15.7,'Ht')
    else
       icflg=-1
    end if
    return
  end subroutine cnvch2

  subroutine prnatmst(aa,nsi,asi,dforce,dform,etotal,setotal, &
       eleng,nspin,ef,ewald)
    use util
    use types
    implicit none
    integer,intent(in):: nsi,nspin
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),dimension(3,nsi),intent(in):: asi,dforce,dform
    real(kind=8),intent(in):: etotal,setotal(6)
    type(elec_energy),intent(in):: eleng
    real(kind=8),intent(in):: ef(nspin),ewald

    integer:: it,i,is
    
    write(6,'(/,'' UNIT CELL'')')
    write(6,'(3(1x,e18.10))') aa(1,1),aa(2,1),aa(3,1)
    write(6,'(3(1x,e18.10))') aa(1,2),aa(2,2),aa(3,2)
    write(6,'(3(1x,e18.10))') aa(1,3),aa(2,3),aa(3,3)
    write(6,'(/,'' CELL VOLUME ='',e19.10)') cellvol(aa)
    write(6,120)
    do it=1,nsi
       write(6,310) it,(asi(i,it),i=1,3),(dform(i,it),i=1,3)
    end do
    write(6,220)
    do it=1,nsi
       write(6,310) it,(asi(i,it),i=1,3),(dforce(i,it),i=1,3)
    end do
    write(6,'(/,'' TOTAL     STRESS ='',e19.10)') &
         -( setotal(1)+setotal(2)+setotal(3) )/(3.0d0*cellvol(aa))
    write(6,'('' STRESS TENSOR = '',6(1x,e14.5),/)') &
         (setotal(i),i=1,6)
    write(6,320) (ef(is),is=1,nspin)
    write(6,330) ewald
    write(6,340) eleng%eht
    write(6,350) eleng%exc+eleng%exx,eleng%exa
    write(6,360) eleng%epot,eleng%epotl,eleng%epotnl
    write(6,370) eleng%ekin
    write(6,380) etotal
       !     
120 format(/,' IN',6x,'POSITION OF ATOM',12x,'DFORM IN AA-REP  ')
310 format(' ',i3,',(',3f16.12,'),(',3e15.6,')')
220 format(/,' IN',6x,'POSITION OF ATOM',12x,'TOTAL FORCE IN HT')
320 format(' FERMI     ENERGY =',e19.10,1x,e19.10)
330 format(' EWALD            =',e19.10)
340 format(' HARTREE   ENERGY =',e19.10)
350 format(' EXCHANGE  ENERGY =',e19.10,' EXA =',e19.10)
360 format(' POTENTIAL ENERGY =',e19.10,' LOCAL PART =',e19.10 &
         ,' NONLOCAL PART =',e19.10)
370 format(' KINETIC   ENERGY =',e19.10)
380 format(' TOTAL     ENERGY =',e19.10)
  end subroutine prnatmst

end module linear_minim
