!
! Copyright (c) 2000-2013,2017 Yoshihide Yoshimoto
!
module refewald
  private
  public:: fewald, eatmewald, seatmewald, setfewald

  real(kind=8),parameter:: pi = 3.14159265358979323846d0
  real(kind=8),parameter:: ewaldmax = 6.0d0
  integer:: nkd
  real(kind=8),allocatable:: zo(:)
  real(kind=8),parameter:: alpha = 8.0d0

!
! parameters for erfc
!
  real(kind=8),parameter:: &
     pa = 3.97886080735226000d+00, &
     p0 = 2.75374741597376782d-01, &
     p1 = 4.90165080585318424d-01, &
     p2 = 7.74368199119538609d-01, &
     p3 = 1.07925515155856677d+00, &
     p4 = 1.31314653831023098d+00, &
     p5 = 1.37040217682338167d+00, &
     p6 = 1.18902982909273333d+00, &
     p7 = 8.05276408752910567d-01, &
     p8 = 3.57524274449531043d-01, &
     p9 = 1.66207924969367356d-02, &
     p10 = -1.19463959964325415d-01, &
     p11 = -8.38864557023001992d-02
  real(kind=8),parameter:: &
     p12 = 2.49367200053503304d-03, &
     p13 = 3.90976845588484035d-02, &
     p14 = 1.61315329733252248d-02, &
     p15 = -1.33823644533460069d-02, &
     p16 = -1.27223813782122755d-02, &
     p17 = 3.83335126264887303d-03, &
     p18 = 7.73672528313526668d-03, &
     p19 = -8.70779635317295828d-04, &
     p20 = -3.96385097360513500d-03, &
     p21 = 1.19314022838340944d-04, &
     p22 = 1.27109764952614092d-03

contains
  !
  subroutine fewald(aa,nsi,kd,asi,eng,seng,cforce)
    use util
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(out):: eng,seng(6),cforce(3,nsi)

    integer:: i,j,k,ig,ir,m1,m2,m3
    real(kind=8):: tmp,tmp2,ph,h(3),va,gamma,const,fac,qrd,qrdph,tzz
    real(kind=8):: bb(3,3),ucg(4),tucr(4),dst(3),cdst(3),tcd(4),tcdsq
    integer:: nrm,ngm,nx,ny,nz
    integer,allocatable:: mg(:,:)
    real(kind=8),allocatable:: snph(:),csph(:),ucr(:,:)
    real(kind=8):: cforcer(3,nsi)
    real(kind=8),allocatable,dimension(:,:):: snpx,cspx,snpy,cspy,snpz,cspz
    real(kind=8):: ct,st,c1,s1
    complex(kind=8):: sf
    real(kind=8):: et,eu,eu2,eu4,ey0,ey1,ey2,ey3,ey,esq
    real(kind=8):: tcf(3),tcfi(3)

    if (maxval(zo).eq.0.0d0 .and. minval(zo).eq.0.0d0) then
       eng = 0.0d0
       seng = 0.0d0
       cforce = 0.0d0
       return
    end if

    va = cellvol(aa)
    gamma = (va**(1.0d0/3.0d0))*0.5d0/sqrt(pi)
    gamma = gamma/( (alpha*nsi)**(1.0d0/6.0d0) )
    call recp_uvec(aa,bb)

    qrd = ewaldmax/gamma
    call LatticeInterval(bb,h)
    nx = ceiling(qrd/h(1))
    ny = ceiling(qrd/h(2))
    nz = ceiling(qrd/h(3))

    ngm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(mg(3,0:ngm), &
         snpx(nsi,-nx:nx),cspx(nsi,-nx:nx), &
         snpy(nsi,-ny:ny),cspy(nsi,-ny:ny), &
         snpz(nsi,-nz:nz),cspz(nsi,-nz:nz))
    ig = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             ucg(1) = bb(1,1)*m1 + bb(1,2)*m2 + bb(1,3)*m3
             ucg(2) = bb(2,1)*m1 + bb(2,2)*m2 + bb(2,3)*m3
             ucg(3) = bb(3,1)*m1 + bb(3,2)*m2 + bb(3,3)*m3
             ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
             if (ucg(4).lt.qrd*qrd) then
                mg(1,ig) = m1
                mg(2,ig) = m2
                mg(3,ig) = m3
                ig=ig+1
             end if
          end do
       end do
    end do
    ngm = ig - 1

!$omp parallel do private(k,ph,c1,s1)
    do i=1,nsi
       ph = 2.0d0*pi*asi(1,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(1,i)*nx
       cspx(i,-nx) = cos(ph)
       snpx(i,-nx) = sin(ph)
       do k=-nx+1,nx
          cspx(i,k) = cspx(i,k-1)*c1 - snpx(i,k-1)*s1
          snpx(i,k) = snpx(i,k-1)*c1 + cspx(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(2,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(2,i)*ny
       cspy(i,-ny) = cos(ph)
       snpy(i,-ny) = sin(ph)
       do k=-ny+1,ny
          cspy(i,k) = cspy(i,k-1)*c1 - snpy(i,k-1)*s1
          snpy(i,k) = snpy(i,k-1)*c1 + cspy(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(3,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(3,i)*nz
       cspz(i,-nz) = cos(ph)
       snpz(i,-nz) = sin(ph)
       do k=-nz+1,nz
          cspz(i,k) = cspz(i,k-1)*c1 - snpz(i,k-1)*s1
          snpz(i,k) = snpz(i,k-1)*c1 + cspz(i,k-1)*s1
       end do
    end do

    eng = 0.0d0
    seng = 0.0d0
    cforce = 0.0d0

    const = 4.0d0*pi/va
!$omp parallel private(ucg,dst,snph,csph,ig,i,tmp,fac,ct,st,sf)
    allocate(snph(nsi),csph(nsi))
!$omp do reduction(+:cforce,eng,seng)
!poption psum(eng,cforce(3,nsi),seng(6)),tlocal(ucg,dst,snph,csph)
    do ig=1,ngm
       ucg(1) = bb(1,1)*mg(1,ig) + bb(1,2)*mg(2,ig) + bb(1,3)*mg(3,ig)
       ucg(2) = bb(2,1)*mg(1,ig) + bb(2,2)*mg(2,ig) + bb(2,3)*mg(3,ig)
       ucg(3) = bb(3,1)*mg(1,ig) + bb(3,2)*mg(2,ig) + bb(3,3)*mg(3,ig)
       ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
       tmp = const/ucg(4)*exp(-gamma*gamma*ucg(4))
       sf = (0.0d0, 0.0d0)
       do i=1,nsi
          ct = cspx(i,mg(1,ig))*cspy(i,mg(2,ig)) &
               - snpx(i,mg(1,ig))*snpy(i,mg(2,ig))
          st = cspx(i,mg(1,ig))*snpy(i,mg(2,ig)) &
               + snpx(i,mg(1,ig))*cspy(i,mg(2,ig))
          csph(i) = ct*cspz(i,mg(3,ig)) - st*snpz(i,mg(3,ig))
          snph(i) = ct*snpz(i,mg(3,ig)) + st*cspz(i,mg(3,ig))

          sf = sf + zo(kd(i))*dcmplx(csph(i),-snph(i))
       end do
       fac = ( dble(sf)*dble(sf) + aimag(sf)*aimag(sf) )*tmp
       eng = eng + 0.5d0*fac
       seng(1) = seng(1) + 0.5d0*fac
       seng(2) = seng(2) + 0.5d0*fac
       seng(3) = seng(3) + 0.5d0*fac
       fac = fac*(1.0d0/ucg(4) + gamma*gamma)
       seng(1) = seng(1) - fac*ucg(1)*ucg(1)
       seng(2) = seng(2) - fac*ucg(2)*ucg(2)
       seng(3) = seng(3) - fac*ucg(3)*ucg(3)
       seng(4) = seng(4) - fac*2.0d0*ucg(2)*ucg(3)
       seng(5) = seng(5) - fac*2.0d0*ucg(3)*ucg(1)
       seng(6) = seng(6) - fac*2.0d0*ucg(1)*ucg(2)
       do i=1,nsi
          fac = zo(kd(i))*( dble(sf)*snph(i) + aimag(sf)*csph(i) )*tmp
          cforce(1,i) = cforce(1,i) + ucg(1)*fac
          cforce(2,i) = cforce(2,i) + ucg(2)*fac
          cforce(3,i) = cforce(3,i) + ucg(3)*fac
       end do
    end do
!$omp end do
    deallocate(snph,csph)
!$omp end parallel

    tmp = 0.0d0
!$omp parallel do reduction(+:tmp)
    do i=1,nsi
       tmp = tmp + zo(kd(i))
    end do
    eng = eng - 0.5d0*const*tmp*tmp*gamma*gamma
    seng(1:3) = seng(1:3) - 0.5d0*const*tmp*tmp*gamma*gamma

    deallocate(mg,snpx,cspx,snpy,cspy,snpz,cspz)

    qrd = ewaldmax*2.0d0*gamma
    call LatticeInterval(aa,h)
!
! because dst vector is in
! aa*[-0.5:0.5; -0.5:0.5; -0.5:0.5] range
! add dmaxaa(aa) to qrd
!
    qrdph = qrd + dmaxaa(aa)
    nx = ceiling(qrdph/h(1))
    ny = ceiling(qrdph/h(2))
    nz = ceiling(qrdph/h(3))

    nrm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(ucr(4,0:nrm))
    ir = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             tucr(1) = aa(1,1)*m1 + aa(1,2)*m2 + aa(1,3)*m3
             tucr(2) = aa(2,1)*m1 + aa(2,2)*m2 + aa(2,3)*m3
             tucr(3) = aa(3,1)*m1 + aa(3,2)*m2 + aa(3,3)*m3
             tucr(4) = tucr(1)*tucr(1) + tucr(2)*tucr(2) + tucr(3)*tucr(3)
             tucr(4) = sqrt(tucr(4))
             if (tucr(4).lt.qrdph) then
                ucr(:,ir) = tucr
                ir=ir+1
             end if
          end do
       end do
    end do
    nrm = ir - 1

    const = 1.0d0/sqrt(pi)/gamma

    tmp = 0.0d0
!$omp parallel do reduction(+:tmp)
!poption psum(tmp)
    do i=1,nsi
       tmp = tmp + zo(kd(i))*zo(kd(i))
    end do

    eng = eng - 0.5d0*const*tmp

!$omp parallel do private(tmp2,esq,ey,fac),reduction(+:eng,seng)
!poption psum(eng,seng(6))
    do ir=1,nrm
       esq = exp(-ucr(4,ir)*ucr(4,ir)*0.25d0/(gamma*gamma))
       ey = erfc(ucr(4,ir)/2.0d0/gamma)
       tmp2 = 0.5d0/ucr(4,ir)*ey
       eng = eng + tmp2*tmp
       fac = 0.5d0*tmp*(const*esq + 1.0d0/ucr(4,ir)*ey)/(ucr(4,ir)*ucr(4,ir))
       seng(1) = seng(1) + fac*ucr(1,ir)*ucr(1,ir)
       seng(2) = seng(2) + fac*ucr(2,ir)*ucr(2,ir)
       seng(3) = seng(3) + fac*ucr(3,ir)*ucr(3,ir)
       seng(4) = seng(4) + fac*2.0d0*ucr(2,ir)*ucr(3,ir)
       seng(5) = seng(5) + fac*2.0d0*ucr(3,ir)*ucr(1,ir)
       seng(6) = seng(6) + fac*2.0d0*ucr(1,ir)*ucr(2,ir)
    end do

    cforcer = 0.0d0
!$omp parallel do private(dst,cdst,tcd,tcf,tcfi,j,ir,fac,tzz,tcdsq), &
!$omp private(et,eu,eu2,eu4,ey,esq,ey0,ey1,ey2,ey3), &
!$omp reduction(+:eng,cforcer,seng),schedule(static,1)
!poption psum(eng,cforcer(3,nsi),seng(6)),tlocal(dst,cdst,tcd,tcf,tcfi),cyclic
    do i=1,nsi
       tcfi(1) = 0.0d0
       tcfi(2) = 0.0d0
       tcfi(3) = 0.0d0
       do j=i+1,nsi
          dst(1) = asi(1,i) - asi(1,j)
          dst(2) = asi(2,i) - asi(2,j)
          dst(3) = asi(3,i) - asi(3,j)
          dst(1) = dst(1) - nint(dst(1))
          dst(2) = dst(2) - nint(dst(2))
          dst(3) = dst(3) - nint(dst(3))
          cdst(1) = aa(1,1)*dst(1) + aa(1,2)*dst(2) + aa(1,3)*dst(3)
          cdst(2) = aa(2,1)*dst(1) + aa(2,2)*dst(2) + aa(2,3)*dst(3)
          cdst(3) = aa(3,1)*dst(1) + aa(3,2)*dst(2) + aa(3,3)*dst(3)
          tzz = zo(kd(i))*zo(kd(j))
          tcf(1) = 0.0d0
          tcf(2) = 0.0d0
          tcf(3) = 0.0d0
          do ir=0,nrm
             tcd(1) = cdst(1) - ucr(1,ir)
             tcd(2) = cdst(2) - ucr(2,ir)
             tcd(3) = cdst(3) - ucr(3,ir)
             tcdsq = tcd(1)*tcd(1) + tcd(2)*tcd(2) + tcd(3)*tcd(3)

             if (tcdsq .ge. qrd*qrd) cycle

             tcd(4) = sqrt(tcdsq)
!
! erfc from a package, gamerf obtained from
!   http://www.kurims.kyoto-u.ac.jp/~ooura/index-j.html
! at 2007/4/20.
!
!  Copyright(C) 1996 Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp).
!  You may use, copy, modify this code for any purpose and 
!  without fee. You may distribute this ORIGINAL package.
!
             esq = exp(-tcdsq*0.25d0/(gamma*gamma))

             et = pa / (pa + tcd(4)*0.5d0/gamma)
             eu = et - 0.5d0
             eu2 = eu*eu
             eu4 = eu2*eu2

             ey0 = p20*eu4 + p16
             ey1 = p21*eu4 + p17
             ey2 = p22*eu4 + p18

             ey0 = ey0*eu4 + p12
             ey1 = ey1*eu4 + p13
             ey2 = ey2*eu4 + p14
             ey3 = p19*eu4 + p15

             ey0 = ey0*eu4 + p8
             ey1 = ey1*eu4 + p9
             ey2 = ey2*eu4 + p10
             ey3 = ey3*eu4 + p11

             ey0 = ey0*eu4 + p4
             ey1 = ey1*eu4 + p5
             ey2 = ey2*eu4 + p6
             ey3 = ey3*eu4 + p7

             ey0 = ey0*eu4 + p0
             ey1 = ey1*eu4 + p1
             ey2 = ey2*eu4 + p2
             ey3 = ey3*eu4 + p3

             ey = ((ey3*eu + ey2)*eu2 + ey1*eu + ey0)*et*esq
!
! end erfc
!
             eng = eng + tzz/tcd(4)*ey

             fac = const*esq + 1.0d0/tcd(4)*ey
             fac = fac*tzz/tcdsq

             tcf(1) = tcf(1) + tcd(1)*fac 
             tcf(2) = tcf(2) + tcd(2)*fac 
             tcf(3) = tcf(3) + tcd(3)*fac 

             fac = tzz*(const*esq + 1.0d0/tcd(4)*ey)
             seng(1) = seng(1) + fac*tcd(1)*tcd(1)/tcdsq
             seng(2) = seng(2) + fac*tcd(2)*tcd(2)/tcdsq
             seng(3) = seng(3) + fac*tcd(3)*tcd(3)/tcdsq
             seng(4) = seng(4) + fac*2.0d0*tcd(2)*tcd(3)/tcdsq
             seng(5) = seng(5) + fac*2.0d0*tcd(3)*tcd(1)/tcdsq
             seng(6) = seng(6) + fac*2.0d0*tcd(1)*tcd(2)/tcdsq
          end do
          tcfi(1) = tcfi(1) + tcf(1)
          tcfi(2) = tcfi(2) + tcf(2)
          tcfi(3) = tcfi(3) + tcf(3)
          cforcer(1,j) = cforcer(1,j) - tcf(1)
          cforcer(2,j) = cforcer(2,j) - tcf(2)
          cforcer(3,j) = cforcer(3,j) - tcf(3)
       end do
       cforcer(1,i) = cforcer(1,i) + tcfi(1)
       cforcer(2,i) = cforcer(2,i) + tcfi(2)
       cforcer(3,i) = cforcer(3,i) + tcfi(3)
    end do

    cforce = cforce + cforcer

    deallocate(ucr)

    return
  end subroutine fewald

  subroutine eatmewald(aa,nsi,kd,asi,engatm)
    use util
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(out):: engatm(nsi)

    integer:: i,j,k,ig,ir,m1,m2,m3
    real(kind=8):: tmp,tmp2,ph,h(3),va,gamma,const,fac,qrd,qrdph,tzz
    real(kind=8):: bb(3,3),ucg(4),tucr(4),dst(3),cdst(3),tcd(4),tcdsq
    integer:: nrm,ngm,nx,ny,nz
    integer,allocatable:: mg(:,:)
    real(kind=8),allocatable:: snph(:),csph(:),ucr(:,:)
    real(kind=8):: engatmr(nsi)
    real(kind=8),allocatable,dimension(:,:):: snpx,cspx,snpy,cspy,snpz,cspz
    real(kind=8):: ct,st,c1,s1
    complex(kind=8):: sf
    real(kind=8):: et,eu,eu2,eu4,ey0,ey1,ey2,ey3,ey,esq

    if (maxval(zo).eq.0.0d0 .and. minval(zo).eq.0.0d0) then
       engatm = 0.0d0
       return
    end if

    va = cellvol(aa)
    gamma = (va**(1.0d0/3.0d0))*0.5d0/sqrt(pi)
    gamma = gamma/( (alpha*nsi)**(1.0d0/6.0d0) )
    call recp_uvec(aa,bb)

    qrd = ewaldmax/gamma
    call LatticeInterval(bb,h)
    nx = ceiling(qrd/h(1))
    ny = ceiling(qrd/h(2))
    nz = ceiling(qrd/h(3))

    ngm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(mg(3,0:ngm), &
         snpx(nsi,-nx:nx),cspx(nsi,-nx:nx), &
         snpy(nsi,-ny:ny),cspy(nsi,-ny:ny), &
         snpz(nsi,-nz:nz),cspz(nsi,-nz:nz))
    ig = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             ucg(1) = bb(1,1)*m1 + bb(1,2)*m2 + bb(1,3)*m3
             ucg(2) = bb(2,1)*m1 + bb(2,2)*m2 + bb(2,3)*m3
             ucg(3) = bb(3,1)*m1 + bb(3,2)*m2 + bb(3,3)*m3
             ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
             if (ucg(4).lt.qrd*qrd) then
                mg(1,ig) = m1
                mg(2,ig) = m2
                mg(3,ig) = m3
                ig=ig+1
             end if
          end do
       end do
    end do
    ngm = ig - 1

!$omp parallel do private(k,ph,c1,s1)
    do i=1,nsi
       ph = 2.0d0*pi*asi(1,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(1,i)*nx
       cspx(i,-nx) = cos(ph)
       snpx(i,-nx) = sin(ph)
       do k=-nx+1,nx
          cspx(i,k) = cspx(i,k-1)*c1 - snpx(i,k-1)*s1
          snpx(i,k) = snpx(i,k-1)*c1 + cspx(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(2,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(2,i)*ny
       cspy(i,-ny) = cos(ph)
       snpy(i,-ny) = sin(ph)
       do k=-ny+1,ny
          cspy(i,k) = cspy(i,k-1)*c1 - snpy(i,k-1)*s1
          snpy(i,k) = snpy(i,k-1)*c1 + cspy(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(3,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(3,i)*nz
       cspz(i,-nz) = cos(ph)
       snpz(i,-nz) = sin(ph)
       do k=-nz+1,nz
          cspz(i,k) = cspz(i,k-1)*c1 - snpz(i,k-1)*s1
          snpz(i,k) = snpz(i,k-1)*c1 + cspz(i,k-1)*s1
       end do
    end do

    engatm = 0.0d0

    const = 4.0d0*pi/va
!$omp parallel private(ucg,dst,snph,csph,ig,i,tmp,ct,st,sf)
    allocate(snph(nsi),csph(nsi))
!$omp do reduction(+:engatm)
!poption psum(engatm(nsi)),tlocal(ucg,dst,snph,csph)
    do ig=1,ngm
       ucg(1) = bb(1,1)*mg(1,ig) + bb(1,2)*mg(2,ig) + bb(1,3)*mg(3,ig)
       ucg(2) = bb(2,1)*mg(1,ig) + bb(2,2)*mg(2,ig) + bb(2,3)*mg(3,ig)
       ucg(3) = bb(3,1)*mg(1,ig) + bb(3,2)*mg(2,ig) + bb(3,3)*mg(3,ig)
       ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
       tmp = const/ucg(4)*exp(-gamma*gamma*ucg(4))
       sf = (0.0d0, 0.0d0)
       do i=1,nsi
          ct = cspx(i,mg(1,ig))*cspy(i,mg(2,ig)) &
               - snpx(i,mg(1,ig))*snpy(i,mg(2,ig))
          st = cspx(i,mg(1,ig))*snpy(i,mg(2,ig)) &
               + snpx(i,mg(1,ig))*cspy(i,mg(2,ig))
          csph(i) = ct*cspz(i,mg(3,ig)) - st*snpz(i,mg(3,ig))
          snph(i) = ct*snpz(i,mg(3,ig)) + st*cspz(i,mg(3,ig))

          sf = sf + zo(kd(i))*dcmplx(csph(i),-snph(i))
       end do
       do i=1,nsi
          engatm(i) = engatm(i) + 0.5d0*tmp*zo(kd(i)) &
               *( csph(i)*dble(sf) - snph(i)*aimag(sf) )
       end do
    end do
!$omp end do
    deallocate(snph,csph)
!$omp end parallel

    tmp = 0.0d0
!$omp parallel do reduction(+:tmp)
    do i=1,nsi
       tmp = tmp + zo(kd(i))
    end do
    do i=1,nsi
       engatm(i) = engatm(i) - 0.5d0*const*zo(kd(i))*tmp*gamma*gamma
    end do

    deallocate(mg,snpx,cspx,snpy,cspy,snpz,cspz)

    qrd = ewaldmax*2.0d0*gamma
    call LatticeInterval(aa,h)
!
! because dst vector is in
! aa*[-0.5:0.5; -0.5:0.5; -0.5:0.5] range
! add dmaxaa(aa) to qrd
!
    qrdph = qrd + dmaxaa(aa)
    nx = ceiling(qrdph/h(1))
    ny = ceiling(qrdph/h(2))
    nz = ceiling(qrdph/h(3))

    nrm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(ucr(4,0:nrm))
    ir = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             tucr(1) = aa(1,1)*m1 + aa(1,2)*m2 + aa(1,3)*m3
             tucr(2) = aa(2,1)*m1 + aa(2,2)*m2 + aa(2,3)*m3
             tucr(3) = aa(3,1)*m1 + aa(3,2)*m2 + aa(3,3)*m3
             tucr(4) = tucr(1)*tucr(1) + tucr(2)*tucr(2) + tucr(3)*tucr(3)
             tucr(4) = sqrt(tucr(4))
             if (tucr(4).lt.qrdph) then
                ucr(:,ir) = tucr
                ir=ir+1
             end if
          end do
       end do
    end do
    nrm = ir - 1

    const = 1.0d0/sqrt(pi)/gamma

    do i=1,nsi
       engatm(i) = engatm(i) - 0.5d0*const*zo(kd(i))*zo(kd(i))
    end do

!$omp parallel do private(tmp2,esq,ey,i),reduction(+:engatm)
!poption psum(engatm)
    do ir=1,nrm
       esq = exp(-ucr(4,ir)*ucr(4,ir)*0.25d0/(gamma*gamma))
       ey = erfc(ucr(4,ir)/2.0d0/gamma)
       tmp2 = 0.5d0/ucr(4,ir)*ey
       do i=1,nsi
          engatm(i) = engatm(i) + tmp2*zo(kd(i))*zo(kd(i))
       end do
    end do

    engatmr = 0.0d0
!$omp parallel do private(dst,cdst,tcd,j,ir,tzz,tcdsq), &
!$omp private(et,eu,eu2,eu4,ey,esq,ey0,ey1,ey2,ey3), &
!$omp reduction(+:engatmr),schedule(static,1)
!poption psum(engatmr),tlocal(dst,cdst,tcd),cyclic
    do i=1,nsi
       do j=i+1,nsi
          dst(1) = asi(1,i) - asi(1,j)
          dst(2) = asi(2,i) - asi(2,j)
          dst(3) = asi(3,i) - asi(3,j)
          dst(1) = dst(1) - nint(dst(1))
          dst(2) = dst(2) - nint(dst(2))
          dst(3) = dst(3) - nint(dst(3))
          cdst(1) = aa(1,1)*dst(1) + aa(1,2)*dst(2) + aa(1,3)*dst(3)
          cdst(2) = aa(2,1)*dst(1) + aa(2,2)*dst(2) + aa(2,3)*dst(3)
          cdst(3) = aa(3,1)*dst(1) + aa(3,2)*dst(2) + aa(3,3)*dst(3)
          tzz = zo(kd(i))*zo(kd(j))
          do ir=0,nrm
             tcd(1) = cdst(1) - ucr(1,ir)
             tcd(2) = cdst(2) - ucr(2,ir)
             tcd(3) = cdst(3) - ucr(3,ir)
             tcdsq = tcd(1)*tcd(1) + tcd(2)*tcd(2) + tcd(3)*tcd(3)

             if (tcdsq .ge. qrd*qrd) cycle

             tcd(4) = sqrt(tcdsq)
!
! erfc from a package, gamerf obtained from
!   http://www.kurims.kyoto-u.ac.jp/~ooura/index-j.html
! at 2007/4/20.
!
!  Copyright(C) 1996 Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp).
!  You may use, copy, modify this code for any purpose and 
!  without fee. You may distribute this ORIGINAL package.
!
             esq = exp(-tcdsq*0.25d0/(gamma*gamma))

             et = pa / (pa + tcd(4)*0.5d0/gamma)
             eu = et - 0.5d0
             eu2 = eu*eu
             eu4 = eu2*eu2

             ey0 = p20*eu4 + p16
             ey1 = p21*eu4 + p17
             ey2 = p22*eu4 + p18

             ey0 = ey0*eu4 + p12
             ey1 = ey1*eu4 + p13
             ey2 = ey2*eu4 + p14
             ey3 = p19*eu4 + p15

             ey0 = ey0*eu4 + p8
             ey1 = ey1*eu4 + p9
             ey2 = ey2*eu4 + p10
             ey3 = ey3*eu4 + p11

             ey0 = ey0*eu4 + p4
             ey1 = ey1*eu4 + p5
             ey2 = ey2*eu4 + p6
             ey3 = ey3*eu4 + p7

             ey0 = ey0*eu4 + p0
             ey1 = ey1*eu4 + p1
             ey2 = ey2*eu4 + p2
             ey3 = ey3*eu4 + p3

             ey = ((ey3*eu + ey2)*eu2 + ey1*eu + ey0)*et*esq
!
! end erfc
!
             engatmr(i) = engatmr(i) + tzz/tcd(4)*ey
             engatmr(j) = engatmr(j) + tzz/tcd(4)*ey
          end do
       end do
    end do

    engatm = engatm + 0.5d0*engatmr

    deallocate(ucr)

    return
  end subroutine eatmewald

  subroutine seatmewald(aa,nsi,kd,asi,engatm,sengatm)
    use util
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(out):: engatm(nsi),sengatm(6,nsi)

    integer:: i,j,k,ig,ir,m1,m2,m3
    real(kind=8):: tmp,tmp2,ph,h(3),va,gamma,const,fac,qrd,qrdph,tzz
    real(kind=8):: bb(3,3),ucg(4),tucr(4),dst(3),cdst(3),tcd(4),tcdsq
    integer:: nrm,ngm,nx,ny,nz
    integer,allocatable:: mg(:,:)
    real(kind=8),allocatable:: snph(:),csph(:),ucr(:,:)
    real(kind=8):: engatmr(nsi),sengatmr(6,nsi)
    real(kind=8),allocatable,dimension(:,:):: snpx,cspx,snpy,cspy,snpz,cspz
    real(kind=8):: ct,st,c1,s1
    complex(kind=8):: sf
    real(kind=8):: et,eu,eu2,eu4,ey0,ey1,ey2,ey3,ey,esq

    if (maxval(zo).eq.0.0d0 .and. minval(zo).eq.0.0d0) then
       engatm = 0.0d0
       sengatm = 0.0d0
       return
    end if

    va = cellvol(aa)
    gamma = (va**(1.0d0/3.0d0))*0.5d0/sqrt(pi)
    gamma = gamma/( (alpha*nsi)**(1.0d0/6.0d0) )
    call recp_uvec(aa,bb)

    qrd = ewaldmax/gamma
    call LatticeInterval(bb,h)
    nx = ceiling(qrd/h(1))
    ny = ceiling(qrd/h(2))
    nz = ceiling(qrd/h(3))

    ngm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(mg(3,0:ngm), &
         snpx(nsi,-nx:nx),cspx(nsi,-nx:nx), &
         snpy(nsi,-ny:ny),cspy(nsi,-ny:ny), &
         snpz(nsi,-nz:nz),cspz(nsi,-nz:nz))
    ig = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             ucg(1) = bb(1,1)*m1 + bb(1,2)*m2 + bb(1,3)*m3
             ucg(2) = bb(2,1)*m1 + bb(2,2)*m2 + bb(2,3)*m3
             ucg(3) = bb(3,1)*m1 + bb(3,2)*m2 + bb(3,3)*m3
             ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
             if (ucg(4).lt.qrd*qrd) then
                mg(1,ig) = m1
                mg(2,ig) = m2
                mg(3,ig) = m3
                ig=ig+1
             end if
          end do
       end do
    end do
    ngm = ig - 1

!$omp parallel do private(k,ph,c1,s1)
    do i=1,nsi
       ph = 2.0d0*pi*asi(1,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(1,i)*nx
       cspx(i,-nx) = cos(ph)
       snpx(i,-nx) = sin(ph)
       do k=-nx+1,nx
          cspx(i,k) = cspx(i,k-1)*c1 - snpx(i,k-1)*s1
          snpx(i,k) = snpx(i,k-1)*c1 + cspx(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(2,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(2,i)*ny
       cspy(i,-ny) = cos(ph)
       snpy(i,-ny) = sin(ph)
       do k=-ny+1,ny
          cspy(i,k) = cspy(i,k-1)*c1 - snpy(i,k-1)*s1
          snpy(i,k) = snpy(i,k-1)*c1 + cspy(i,k-1)*s1
       end do

       ph = 2.0d0*pi*asi(3,i)
       c1 = cos(ph)
       s1 = sin(ph)
       ph = -2.0d0*pi*asi(3,i)*nz
       cspz(i,-nz) = cos(ph)
       snpz(i,-nz) = sin(ph)
       do k=-nz+1,nz
          cspz(i,k) = cspz(i,k-1)*c1 - snpz(i,k-1)*s1
          snpz(i,k) = snpz(i,k-1)*c1 + cspz(i,k-1)*s1
       end do
    end do

    engatm = 0.0d0
    sengatm = 0.0d0

    const = 4.0d0*pi/va
!$omp parallel private(ucg,dst,snph,csph,ig,i,tmp,fac,ct,st,sf)
    allocate(snph(nsi),csph(nsi))
!$omp do reduction(+:engatm,sengatm)
!poption psum(engatm(nsi),sengatm(6,nsi)),tlocal(ucg,dst,snph,csph)
    do ig=1,ngm
       ucg(1) = bb(1,1)*mg(1,ig) + bb(1,2)*mg(2,ig) + bb(1,3)*mg(3,ig)
       ucg(2) = bb(2,1)*mg(1,ig) + bb(2,2)*mg(2,ig) + bb(2,3)*mg(3,ig)
       ucg(3) = bb(3,1)*mg(1,ig) + bb(3,2)*mg(2,ig) + bb(3,3)*mg(3,ig)
       ucg(4) = ucg(1)*ucg(1) + ucg(2)*ucg(2) + ucg(3)*ucg(3)
       tmp = const/ucg(4)*exp(-gamma*gamma*ucg(4))
       sf = (0.0d0, 0.0d0)
       do i=1,nsi
          ct = cspx(i,mg(1,ig))*cspy(i,mg(2,ig)) &
               - snpx(i,mg(1,ig))*snpy(i,mg(2,ig))
          st = cspx(i,mg(1,ig))*snpy(i,mg(2,ig)) &
               + snpx(i,mg(1,ig))*cspy(i,mg(2,ig))
          csph(i) = ct*cspz(i,mg(3,ig)) - st*snpz(i,mg(3,ig))
          snph(i) = ct*snpz(i,mg(3,ig)) + st*cspz(i,mg(3,ig))

          sf = sf + zo(kd(i))*dcmplx(csph(i),-snph(i))
       end do
       do i=1,nsi
          fac = zo(kd(i))*( csph(i)*dble(sf) - snph(i)*aimag(sf) )*tmp
          engatm(i) = engatm(i) + 0.5d0*fac
          sengatm(1,i) = sengatm(1,i) + 0.5d0*fac
          sengatm(2,i) = sengatm(2,i) + 0.5d0*fac
          sengatm(3,i) = sengatm(3,i) + 0.5d0*fac
          fac = fac*(1.0d0/ucg(4) + gamma*gamma)
          sengatm(1,i) = sengatm(1,i) - fac*ucg(1)*ucg(1)
          sengatm(2,i) = sengatm(2,i) - fac*ucg(2)*ucg(2)
          sengatm(3,i) = sengatm(3,i) - fac*ucg(3)*ucg(3)
          sengatm(4,i) = sengatm(4,i) - fac*2.0d0*ucg(2)*ucg(3)
          sengatm(5,i) = sengatm(5,i) - fac*2.0d0*ucg(3)*ucg(1)
          sengatm(6,i) = sengatm(6,i) - fac*2.0d0*ucg(1)*ucg(2)
       end do
    end do
!$omp end do
    deallocate(snph,csph)
!$omp end parallel

    tmp = 0.0d0
!$omp parallel do reduction(+:tmp)
    do i=1,nsi
       tmp = tmp + zo(kd(i))
    end do
    do i=1,nsi
       engatm(i) = engatm(i) - 0.5d0*const*zo(kd(i))*tmp*gamma*gamma
       sengatm(1:3,i) = sengatm(1:3,i) - 0.5d0*const*zo(kd(i))*tmp*gamma*gamma
    end do
    deallocate(mg,snpx,cspx,snpy,cspy,snpz,cspz)

    qrd = ewaldmax*2.0d0*gamma
    call LatticeInterval(aa,h)
!
! because dst vector is in
! aa*[-0.5:0.5; -0.5:0.5; -0.5:0.5] range
! add dmaxaa(aa) to qrd
!
    qrdph = qrd + dmaxaa(aa)
    nx = ceiling(qrdph/h(1))
    ny = ceiling(qrdph/h(2))
    nz = ceiling(qrdph/h(3))

    nrm = (2*nx+1)*(2*ny+1)*(2*nz+1)-1
    allocate(ucr(4,0:nrm))
    ir = 0
    do k=0,2*nz
       do j=0,2*ny
          do i=0,2*nx
             m1 = mod(i+nx,2*nx+1)-nx
             m2 = mod(j+ny,2*ny+1)-ny
             m3 = mod(k+nz,2*nz+1)-nz
             tucr(1) = aa(1,1)*m1 + aa(1,2)*m2 + aa(1,3)*m3
             tucr(2) = aa(2,1)*m1 + aa(2,2)*m2 + aa(2,3)*m3
             tucr(3) = aa(3,1)*m1 + aa(3,2)*m2 + aa(3,3)*m3
             tucr(4) = tucr(1)*tucr(1) + tucr(2)*tucr(2) + tucr(3)*tucr(3)
             tucr(4) = sqrt(tucr(4))
             if (tucr(4).lt.qrdph) then
                ucr(:,ir) = tucr
                ir=ir+1
             end if
          end do
       end do
    end do
    nrm = ir - 1

    const = 1.0d0/sqrt(pi)/gamma

    tmp = 0.0d0
!$omp parallel do reduction(+:tmp)
!poption psum(tmp)
    do i=1,nsi
       tmp = tmp + zo(kd(i))*zo(kd(i))
    end do
    do i=1,nsi
       engatm(i) = engatm(i) - 0.5d0*const*zo(kd(i))*zo(kd(i))
    end do

!$omp parallel do private(tmp2,esq,ey,fac),reduction(+:engatm,sengatm)
!poption psum(engatm(nsi),sengatm(6,nsi))
    do ir=1,nrm
       esq = exp(-ucr(4,ir)*ucr(4,ir)*0.25d0/(gamma*gamma))
       ey = erfc(ucr(4,ir)/2.0d0/gamma)
       tmp2 = 0.5d0/ucr(4,ir)*ey
       do i=1,nsi
          engatm(i) = engatm(i) + tmp2*zo(kd(i))*zo(kd(i))
          fac = 0.5d0*zo(kd(i))*zo(kd(i))*(const*esq + 1.0d0/ucr(4,ir)*ey)/(ucr(4,ir)*ucr(4,ir))
          sengatm(1,i) = sengatm(1,i) + fac*ucr(1,ir)*ucr(1,ir)
          sengatm(2,i) = sengatm(2,i) + fac*ucr(2,ir)*ucr(2,ir)
          sengatm(3,i) = sengatm(3,i) + fac*ucr(3,ir)*ucr(3,ir)
          sengatm(4,i) = sengatm(4,i) + fac*2.0d0*ucr(2,ir)*ucr(3,ir)
          sengatm(5,i) = sengatm(5,i) + fac*2.0d0*ucr(3,ir)*ucr(1,ir)
          sengatm(6,i) = sengatm(6,i) + fac*2.0d0*ucr(1,ir)*ucr(2,ir)
       end do
    end do

    engatmr = 0.0d0
    sengatmr = 0.0d0
!$omp parallel do private(dst,cdst,tcd,j,ir,fac,tzz,tcdsq), &
!$omp private(et,eu,eu2,eu4,ey,esq,ey0,ey1,ey2,ey3), &
!$omp reduction(+:engatmr,sengatmr),schedule(static,1)
!poption psum(engatmr,sengatmr),tlocal(dst,cdst,tcd),cyclic
    do i=1,nsi
       do j=i+1,nsi
          dst(1) = asi(1,i) - asi(1,j)
          dst(2) = asi(2,i) - asi(2,j)
          dst(3) = asi(3,i) - asi(3,j)
          dst(1) = dst(1) - nint(dst(1))
          dst(2) = dst(2) - nint(dst(2))
          dst(3) = dst(3) - nint(dst(3))
          cdst(1) = aa(1,1)*dst(1) + aa(1,2)*dst(2) + aa(1,3)*dst(3)
          cdst(2) = aa(2,1)*dst(1) + aa(2,2)*dst(2) + aa(2,3)*dst(3)
          cdst(3) = aa(3,1)*dst(1) + aa(3,2)*dst(2) + aa(3,3)*dst(3)
          tzz = zo(kd(i))*zo(kd(j))
          do ir=0,nrm
             tcd(1) = cdst(1) - ucr(1,ir)
             tcd(2) = cdst(2) - ucr(2,ir)
             tcd(3) = cdst(3) - ucr(3,ir)
             tcdsq = tcd(1)*tcd(1) + tcd(2)*tcd(2) + tcd(3)*tcd(3)

             if (tcdsq .ge. qrd*qrd) cycle

             tcd(4) = sqrt(tcdsq)
!
! erfc from a package, gamerf obtained from
!   http://www.kurims.kyoto-u.ac.jp/~ooura/index-j.html
! at 2007/4/20.
!
!  Copyright(C) 1996 Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp).
!  You may use, copy, modify this code for any purpose and 
!  without fee. You may distribute this ORIGINAL package.
!
             esq = exp(-tcdsq*0.25d0/(gamma*gamma))

             et = pa / (pa + tcd(4)*0.5d0/gamma)
             eu = et - 0.5d0
             eu2 = eu*eu
             eu4 = eu2*eu2

             ey0 = p20*eu4 + p16
             ey1 = p21*eu4 + p17
             ey2 = p22*eu4 + p18

             ey0 = ey0*eu4 + p12
             ey1 = ey1*eu4 + p13
             ey2 = ey2*eu4 + p14
             ey3 = p19*eu4 + p15

             ey0 = ey0*eu4 + p8
             ey1 = ey1*eu4 + p9
             ey2 = ey2*eu4 + p10
             ey3 = ey3*eu4 + p11

             ey0 = ey0*eu4 + p4
             ey1 = ey1*eu4 + p5
             ey2 = ey2*eu4 + p6
             ey3 = ey3*eu4 + p7

             ey0 = ey0*eu4 + p0
             ey1 = ey1*eu4 + p1
             ey2 = ey2*eu4 + p2
             ey3 = ey3*eu4 + p3

             ey = ((ey3*eu + ey2)*eu2 + ey1*eu + ey0)*et*esq
!
! end erfc
!
             engatmr(i) = engatmr(i) + tzz/tcd(4)*ey
             engatmr(j) = engatmr(j) + tzz/tcd(4)*ey

             fac = tzz*(const*esq + 1.0d0/tcd(4)*ey)
             sengatmr(1,i) = sengatmr(1,i) + fac*tcd(1)*tcd(1)/tcdsq
             sengatmr(2,i) = sengatmr(2,i) + fac*tcd(2)*tcd(2)/tcdsq
             sengatmr(3,i) = sengatmr(3,i) + fac*tcd(3)*tcd(3)/tcdsq
             sengatmr(4,i) = sengatmr(4,i) + fac*2.0d0*tcd(2)*tcd(3)/tcdsq
             sengatmr(5,i) = sengatmr(5,i) + fac*2.0d0*tcd(3)*tcd(1)/tcdsq
             sengatmr(6,i) = sengatmr(6,i) + fac*2.0d0*tcd(1)*tcd(2)/tcdsq

             sengatmr(1,j) = sengatmr(1,j) + fac*tcd(1)*tcd(1)/tcdsq
             sengatmr(2,j) = sengatmr(2,j) + fac*tcd(2)*tcd(2)/tcdsq
             sengatmr(3,j) = sengatmr(3,j) + fac*tcd(3)*tcd(3)/tcdsq
             sengatmr(4,j) = sengatmr(4,j) + fac*2.0d0*tcd(2)*tcd(3)/tcdsq
             sengatmr(5,j) = sengatmr(5,j) + fac*2.0d0*tcd(3)*tcd(1)/tcdsq
             sengatmr(6,j) = sengatmr(6,j) + fac*2.0d0*tcd(1)*tcd(2)/tcdsq
          end do
       end do
    end do

    engatm = engatm + 0.5d0*engatmr
    sengatm = sengatm + 0.5d0*sengatmr

    deallocate(ucr)

    return
  end subroutine seatmewald

  subroutine setfewald(inkd,izo)
    implicit none
    integer:: inkd
    real(kind=8):: izo(inkd)
    integer:: ikd

    if (allocated(zo)) deallocate(zo)
    nkd = inkd
    allocate(zo(nkd))

    do ikd=1,nkd
       zo(ikd) = izo(ikd)
    end do
    return
  end subroutine setfewald

  real(kind=8) function dmaxaa(aa)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8):: v(3),tmp

    tmp = 0.0d0

    v =  aa(:,1)+aa(:,2)+aa(:,3)
    tmp = max(tmp,sqrt(v(1)*v(1)+v(2)*v(2)+v(3)*v(3)))
    v = -aa(:,1)+aa(:,2)+aa(:,3)
    tmp = max(tmp,sqrt(v(1)*v(1)+v(2)*v(2)+v(3)*v(3)))
    v =  aa(:,1)-aa(:,2)+aa(:,3)
    tmp = max(tmp,sqrt(v(1)*v(1)+v(2)*v(2)+v(3)*v(3)))
    v = -aa(:,1)-aa(:,2)+aa(:,3)
    tmp = max(tmp,sqrt(v(1)*v(1)+v(2)*v(2)+v(3)*v(3)))

    dmaxaa = 0.5d0*tmp
    return
  end function dmaxaa
end module refewald
