!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module replica
  type replica_struct
     logical:: ifreplm
	 integer:: comt
	 integer:: nrepl, rank
  end type replica_struct
contains
  subroutine setup_replica(nfin,commi,commo,repls)
    use mpi
#ifdef __INTEL_COMPILER
    use ifport
#endif
#ifdef __FUJITSU
    use service_routines
#endif
    implicit none
    integer,intent(in):: nfin
    integer:: commi, commo
	type(replica_struct):: repls

    character(len=256):: buf, buf2, dname, ifname, ofname
    integer,allocatable:: color(:),key(:)
    integer:: nproc, myrank, mylrnk, ierr
    integer:: i

    call MPI_COMM_SIZE(commi,nproc,ierr)
    call MPI_COMM_RANK(commi,myrank,ierr)
    allocate(color(0:nproc-1),key(0:nproc-1))
    read(nfin,*) buf
    if (trim(buf) .eq. 'REPLICA') then
       repls%ifreplm = .true.
       ierr = 0
       do i=0,nproc-1
          read(nfin,*) color(i),key(i)
          if (color(i).le.-1 .or. color(i).ge.nproc) then
             if (myrank.eq.0) write(6,*) 'color out of range for proc = ',i
             ierr = 1
          end if
          if (key(i).le.-1 .or. key(i).ge.nproc) then
             if (myrank.eq.0) write(6,*) 'key out of range for proc = ',i
             ierr = 1
          end if
       end do
       if (ierr.eq.1) then
          call MPI_FINALIZE(ierr)
          stop
       end if
    else
       repls%ifreplm = .false.
       rewind(nfin)
       do i=0,nproc-1
          color(i) = 0
          key(i) = i
       end do
    end if
    if (repls%ifreplm) then
       call MPI_COMM_SPLIT(commi,color(myrank),key(myrank),commo,ierr)
       call MPI_COMM_SPLIT(commi,key(myrank),color(myrank),repls%comt,ierr)
       call MPI_COMM_SIZE(repls%comt,repls%nrepl,ierr)
       call MPI_COMM_RANK(repls%comt,repls%rank,ierr)
    else
       call MPI_COMM_DUP(commi,commo,ierr)
       repls%comt = MPI_COMM_SELF
       repls%nrepl = 1
       repls%rank = 0
    end if

    if (repls%ifreplm) then
       read(nfin,*) buf, ifname, ofname
       write(buf2,'(i5.5)') color(myrank)
       dname = trim(buf)//buf2
       write(6,'(1x,''proc #'',i5,'' chdir to '',a)') myrank, trim(dname)
       ierr = chdir(dname)
       if (ierr .ne. 0) then
          write(6,'(1x,a,i5)') 'cannot chdir: proc = ', myrank
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       end if
       close(nfin)
       close(6)
       open(nfin, file=ifname, status='old', &
            action='read', form='formatted', position='rewind')
       open(6, file=ofname, status='unknown', &
            action='write', form='formatted', position='rewind')
       call MPI_COMM_RANK(commo,mylrnk,ierr)
       if (mylrnk.eq.0) then
          write(6,'(1x,a)') 'replica mode'
          write(6,'(1x,a,i5)') 'color = ', color(myrank)
       end if
    end if
    deallocate(color,key)
    return
  end subroutine setup_replica
end module replica
