!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module stmitgdos
contains
  subroutine vbstd(nfstm,dp,lpt,pby &
       ,ndiag,itsb,qbtp,ekbt,eferm,vs,nbias &
       ,nband,laprj,nsi,nkd,va,ng &
       ,ncomp,nspin,ndens,spaxs,ndx,ndy,ndz,nw,nb1,nb2 &
       ,aa,bb,asi,kd,ee,wgtk,dd &
       ,mgs,xctp,bspotfc,bspots,ifinvsym,mssg &
       ,myrank,nproc,comm)
    !
    use util
    use types
    use soccup_num
    use bsplpot
    use uspp_cr
    use bandchrg
    use converter
    use iounit, only: xrewind
    use diag_hml
    use types
    use meshgrid
    use exchange
    use xcf_types
    implicit none
    integer:: nfstm,itsb,ndiag,nband,laprj,nsi,nkd
    integer:: ncomp,nspin,ndens,ndx,ndy,ndz
    real(kind=8):: spaxs(3)
    integer:: ng,nw,nb1,nb2
    integer:: mssg
    real(kind=8):: ekbt,eferm,va
    real(kind=8):: aa(3,3),bb(3,3),asi(3,nsi)
    integer:: kd(nsi)
    type (meshgrid_struct):: mgs
    integer,intent(in):: xctp
    real(kind=8):: ee(nband,mgs%nks,nspin)
    real(kind=8),intent(in):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: qbtp(3)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    integer:: nbias
    real(kind=8):: vs(nbias)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    complex(kind=8):: lpt(mgs%nvxyz,nspin*ndens)
    logical:: ifinvsym
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    !
    type (wfn_for_k):: dp(mgs%nks,nspin)
    !
    integer:: myrank,nproc,comm
    !
    integer:: nks,nkmesh,nrcell,nkr,nng
    integer:: nrx,nry,nrz,nrxyz,nvx,nvy,nvz,nwx,nwy,nwz
!SOPTION CONTAREA(rco,nkm,kg,nsy,kz,nwk,idcwk,kf,kqg,g_count,g_displ)
    type (irdkpt_struct),pointer:: irdkpt(:)
    real(kind=8),pointer:: rco(:,:)
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: nsy(:),kz(:,:,:),nwk(:,:),idcwk(:,:)
    integer,pointer:: kf(:,:),kqg(:,:)
    integer:: ldkg,bkc,ekc
    integer,pointer:: g_count(:,:),g_displ(:,:)
    !
    ! CODATA 2014
    real(kind=8), parameter:: ehart = 27.21138602d0
    !
    real(kind=8),allocatable:: pe(:,:,:),wgtkd(:,:,:),bzwk(:)
    real(kind=8),allocatable:: qgk(:,:)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    type (qg_for_atm),allocatable:: dqgs(:,:)
    integer:: ldvpj
    real(kind=8),allocatable:: dvpj(:,:)
    real(kind=8),allocatable:: akg(:)
    !
    real(kind=8),allocatable:: rho(:,:),crho(:,:)
    real(kind=8),allocatable:: brho(:,:)
    real(kind=8):: bexa,bexc,bevxc(0:6)
    complex(kind=8),allocatable:: hyblpt(:,:)
    integer:: ierr
    integer:: lpnum,maxr
    integer:: iks,is,ivpj,ikd,ibias,nob,idens
    real(kind=8):: pwchrg, crchrg, tchrg
    real(kind=8):: es
    integer:: bkm,ekm
    integer:: ifgam
    !
    nks = mgs%nks
    nkmesh = mgs%nkmesh
    nrcell = mgs%nrcell
    nkr = mgs%nkr
    nng = mgs%nng
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz
    nrxyz = mgs%nrxyz
    nwx = mgs%nwx
    nwy = mgs%nwy
    nwz = mgs%nwz
    nvx = mgs%nvx
    nvy = mgs%nvy
    nvz = mgs%nvz

    irdkpt => mgs%irdkpt
    rco => mgs%rco
    nkm => mgs%nkm
    kg => mgs%kg
    nsy => mgs%nsy
    kz => mgs%kz
    nwk => mgs%nwk
    idcwk => mgs%idcwk
    kf => mgs%kf
    kqg => mgs%kqg

    ldkg = mgs%ldkg
    bkc = mgs%bkc
    ekc = mgs%ekc
    g_count => mgs%g_count
    g_displ => mgs%g_displ
    
    lpnum = getlpnum(nkd,bspots)
    maxr = getmaxr(nkd,bspots)
    ldvpj=maxval(g_count(myrank,:))*maxr*lpnum*nkd
    allocate(dvpj(ldvpj,nks),stat=ierr)
    call chkalloc('vbstd','dvpj',ierr)
    allocate(akg(ldkg),stat=ierr)
    call chkalloc('vbstd','akg',ierr)
    do iks=1,nks
       bkm=g_displ(myrank,iks)+1
       ekm=g_displ(myrank,iks)+g_count(myrank,iks)
       call kg2akg(bb,irdkpt(iks)%sk,g_count(myrank,iks),kg(:,bkm:ekm,iks),akg)
       ivpj=1
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),g_count(myrank,iks),akg, &
               g_count(myrank,iks),maxr,lpnum,dvpj(ivpj,iks))
          ivpj=ivpj+g_count(myrank,iks)*maxr*lpnum
       end do
    end do
    deallocate(akg)
    allocate(dqgs(nkd,ndens),stat=ierr)
    call chkalloc('vbscd','dqgs',ierr)
    do idens=1,ndens
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd,idens)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('vbscd','qg',ierr)
       end do
    end do
    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('vbscd','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
    do idens=1,ndens
       do ikd=1,nkd
          qg => dqgs(ikd,idens)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,idens,qg)
       end do
    end do
    deallocate(qgk)
    !
    allocate(pe(nkr,nb1:nb2,nspin), &
         wgtkd(nb1:nb2,nks,nspin),bzwk(nks),stat=ierr)
    call chkalloc('vbstd','pe,wgtkd,bzwk',ierr)
    !
    bzwk(:) = irdkpt(:)%ngir
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if
    !
    if (mssg.gt.-1) then
       write(6,*)'############### VBSTD START ###############'
    end if
    !
    allocate(rho(nrxyz,nspin*ndens),crho(nrxyz,nspin*ndens),stat=ierr)
    call chkalloc('vbstd','rho,crho',ierr)
    if (mssg.gt.-1) write(6,*) 'NDIAG=',ndiag, 'ITSB=',itsb
    !
    allocate(hyblpt(mgs%nvxyz,nspin*ndens),stat=ierr)
    call chkalloc('vbstd','hybvg',ierr)
    if (xcf_exch_factor(xctp) .ne. 0.0d0) then
       allocate(brho(nrxyz,nspin*ndens),stat=ierr)
       call chkalloc('vbstd','brho',ierr)
       call yps3(nspin,dp,brho,bzwk,wgtk,va,aa,mgs%nkmesh &
            ,ncomp,ndens,mgs%nrx,mgs%nry,mgs%nrz,mgs%nrxyz &
            ,mgs%ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg &
            ,mgs%nrcell,mgs%nks,mgs%nkm,mgs%kg,mgs%ldkg &
            ,nb1,nb2,mgs%nwk,mgs%idcwk,irdkpt &
            ,(ifinvsym .or. xcf_exch_factor(xctp).ne.0.0d0) &
            ,1,myrank,nproc,mgs%g_count,mgs%g_displ,comm)
       call pothybrid(va,bb,nspin,ndens,spaxs,brho,lpt,bexa,bexc &
            ,bevxc,mgs%nkfxa,mgs%kf,nrx*2,nry*2,nrz*2,nvx,nvy,nvz &
            ,hyblpt,xctp,mssg,comm)
       if (mssg.gt.-1) write(6,*) 'local potential for a hybrid functional'
       deallocate(brho)
       if ( (.not. ifinvsym) .and. nks.eq.1 &
            .and. maxval(abs(irdkpt(1)%sk)).eq.0.0d0 ) then
          ifgam = 1
       else
          ifgam = 0
       end if
       call potexch_init(ifgam,nwx,nwy,nwz,mssg)
    else
       hyblpt = lpt
    end if
    call ps2vcg(dp,ee,dd,nband,nks &
         ,ncomp,nspin,ndens &
         ,dvpj,ldvpj,qbtp &
         ,hyblpt,nvx,nvy,nvz,nw &
         ,ndiag,itsb,bb,va,asi,nsi,nkd,kd,bspots,lpnum,maxr,laprj,pby &
         ,xctp,mgs,nb1,nb2,bzwk,wgtk &
         ,ifinvsym,nproc,comm,mssg)
    if (xcf_exch_factor(xctp) .ne. 0.0d0) then
       call potexch_done()
    end if
    call prntime('PS2VCG ',mssg)
    !
    if (myrank.eq.0) then
       call xrewind(nfstm)
       write(nfstm) 'stm file'
       write(nfstm) aa,nspin
       write(nfstm) nrx,nry,nrz
       write(nfstm) nbias
    end if
    !
    do ibias=1,nbias
       es = eferm + vs(ibias)/ehart
       if (mssg.gt.0) then
          write(6,*) 'STM density at Vs(eV) =', vs(ibias)
          write(6,*)  'EF=', eferm, ' ES= ',es
          write(6,*) 'NB1,NB2=', nb1,nb2
       end if
       !
       call sdoss(nband,ee,rco,nkmesh,pe,es,eferm,ndx,ndy,ndz &
            ,nb1,nb2,nkr,nng,nsy,kz,nks,ncomp,nspin,ekbt,wgtkd,mssg)
       call prntime('SDOSS  ',mssg)
       !
       call yps3(nspin,dp,rho,bzwk,wgtkd,va,aa,nkmesh &
            ,ncomp,1,nrx,nry,nrz,nrxyz &
            ,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg &
            ,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
            ,(ifinvsym .or. xcf_exch_factor(xctp).ne.0.0d0) &
            ,0,myrank,nproc,g_count,g_displ,comm)
       call prntime('YPS3   ',mssg)
       !     
       call crgen(nspin,bkc,ekc,kqg,pby,1,crho &
            ,bzwk,wgtkd,nsi,nkd,kd &
            ,laprj,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg,aa,nrcell,nkmesh,nks,asi &
            ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
            ,dqgs,bspots,0,mssg,comm)
       call prntime('CRGEN  ',mssg)

       do is=1,nspin
          call calchr(rho(1,is),nrxyz,va,pwchrg)
          call calchr(crho(1,is),nrxyz,va,crchrg)
          rho(:,is) = rho(:,is)+crho(:,is)
          tchrg = pwchrg + crchrg
          if (mssg.gt.0) write(6,500) tchrg,crchrg,pwchrg
500       format(' TOTAL CHARGE=',f25.20,' CORE CHARGE=',f25.20/, &
               39x,' PW   CHARGE=',f25.20)
       end do
       !
       if (myrank.eq.0) then
          write(nfstm) vs(ibias),rho
       end if
    end do
    close(nfstm)
    !
    deallocate(hyblpt)
    deallocate(pe,wgtkd,bzwk)
    deallocate(rho,dvpj)
    do idens=1,ndens
       do ikd=1,nkd
          deallocate(dqgs(ikd,idens)%qg)
       end do
    end do
    deallocate(dqgs)
    !
    return
  end subroutine vbstd
end module stmitgdos
