!
! Copyright (c) 2017 Yoshihide Yoshimoto
!
module subr_distpart
contains
  subroutine read_trinkle(bname,nsi,idxcar2xtp,engatm,sengatm)
    use types
    implicit none
    character(len=256):: bname
    integer,intent(in):: nsi
    integer,intent(in):: idxcar2xtp(nsi)
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm
    integer,parameter:: nftrink = 200

    integer:: it,idum,itcar
    real(kind=8):: rdum1,rdum2
    character(len=256):: fname

    fname=trim(bname)//'.wgt-bader'
    open(nftrink, file=fname)
    read(nftrink, *)
    do itcar=1,nsi
       it = idxcar2xtp(itcar)
       read(nftrink, *) idum, rdum1, engatm%ekin(it), engatm%exc(it), engatm%eht(it), rdum2
    end do
    close(nftrink)

    fname=trim(bname)//'.wgt-evxc-bader'
    open(nftrink, file=fname)
    read(nftrink, *)
    do itcar=1,nsi
       it = idxcar2xtp(itcar)
       read(nftrink, *) idum, rdum1, engatm%evxc(:,it), rdum2
    end do
    close(nftrink)

    fname=trim(bname)//'.wgt-seht-bader'
    open(nftrink, file=fname)
    read(nftrink, *)
    do itcar=1,nsi
       it = idxcar2xtp(itcar)
       read(nftrink, *) idum, rdum1, sengatm%seht(:,it), rdum2
    end do
    close(nftrink)

    fname=trim(bname)//'.wgt-sekin-bader'
    open(nftrink, file=fname)
    read(nftrink, *)
    do itcar=1,nsi
       it = idxcar2xtp(itcar)
       read(nftrink, *) idum, rdum1, sengatm%sekin(:,it), rdum2
    end do
    close(nftrink)
    
    return
  end subroutine read_trinkle
  
  subroutine write_distpart(nkd,nsi,kd,engatm,sengatm,bspots,nf)
    use types
    use bsplpot
    integer,intent(in):: nkd,nsi,kd(nsi),nf
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm
    type (bspot_for_atm):: bspots(nkd)

    integer:: it,ikd
    real(kind=8):: sum_ekin, sum_exc, sum_eht, sum_evxc(0:6)
    real(kind=8):: sum_sekin(6), sum_seht(6)
    real(kind=8):: sum_etot, sum_setot(6)

    logical:: ifcoheng

    ifcoheng = .true.
    do ikd=1,nkd
       ifcoheng = ifcoheng .and. bspots(ikd)%ifionchg
    end do

    sum_etot = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# energy per atom'
    if (ifcoheng) then
       write(nf,'(a)') '# total energy, cohesive energy'
       do it=1,nsi
          write(nf,'(6x,2(1x,e21.12))') &
               engatm%total(it), engatm%total(it) - bspots(kd(it))%engatom
          sum_etot = sum_etot + engatm%total(it)
       end do
    else
       write(nf,'(a)') '# total energy'
       do it=1,nsi
          write(nf,'(6x,1x,e21.12)') engatm%total(it)
          sum_etot = sum_etot + engatm%total(it)
       end do
    end if
    write(nf,'(''# Sum '',1x,e21.12)') sum_etot

    sum_ekin = 0.0d0
    sum_exc = 0.0d0
    sum_eht = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# energy components per atom'
    write(nf,'(a)') '# kinetic, exch-corr, Hartree'
    do it=1,nsi
       write(nf,'(6x,3(1x,e21.12))') &
            engatm%ekin(it), engatm%exc(it), engatm%eht(it)
       sum_ekin = sum_ekin + engatm%ekin(it) 
       sum_exc = sum_exc + engatm%exc(it)
       sum_eht = sum_eht + engatm%eht(it)
    end do
    write(nf,'(''# Sum '',3(1x,e21.12))') sum_ekin, sum_exc, sum_eht

    sum_setot = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# stress per atom'
    write(nf,'(a)') '# total(1:6)'
    do it=1,nsi
       write(nf,'(6x,6(1x,e21.12))') &
            sengatm%total(:,it)
       sum_setot = sum_setot + sengatm%total(:,it)
    end do
    write(nf,'(''# Sum '',6(1x,e21.12))') sum_setot

    sum_sekin = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# stress per atom'
    write(nf,'(a)') '# kinetic(1:6)'
    do it=1,nsi
       write(nf,'(6x,6(1x,e21.12))') &
            sengatm%sekin(:,it)
       sum_sekin = sum_sekin + sengatm%sekin(:,it)
    end do
    write(nf,'(''# Sum '',6(1x,e21.12))') sum_sekin

    sum_seht = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# stress per atom'
    write(nf,'(a)') '# Hartree(1:6)'
    do it=1,nsi
       write(nf,'(6x,6(1x,e21.12))') &
            sengatm%seht(:,it)
       sum_seht = sum_seht + sengatm%seht(:,it)
    end do
    write(nf,'(''# Sum '',6(1x,e21.12))') sum_seht

    sum_evxc = 0.0d0
    write(nf,*)
    write(nf,'(a)') '# energy per atom'
    write(nf,'(a)') '# Evxc(0:6)'
    do it=1,nsi
       write(nf,'(6x,7(1x,e21.12))') &
            engatm%evxc(:,it)
       sum_evxc = sum_evxc + engatm%evxc(:,it)
    end do
    write(nf,'(''# Sum '',7(1x,e21.12))') sum_evxc

    return
  end subroutine write_distpart
end module subr_distpart
