!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module subr_ekinor
  public calc_ekinor
contains
  subroutine calc_ekinor(nspin,ncomp,dp,wgtk &
       ,aa,mgs,nb1,nb2,ekinor,sekinor,myrank,nproc,comm)
    use types
    use util
    use subr_kpoint
    use meshgrid
    implicit none
    integer:: nb1,nb2
    integer:: nspin,ncomp
    type (meshgrid_struct):: mgs
    real(kind=8),intent(out):: ekinor(mgs%nrxyz)
    real(kind=8),intent(out):: sekinor(mgs%nrxyz,6)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: aa(3,3)
    integer:: myrank,nproc,comm
    !
    type (wfn_for_k):: dp(mgs%nks,nspin)

    integer:: nks,nkmesh,nrx,nry,nrz,nrxyz,nrcell,ng,nnp
    integer,pointer:: pg(:,:),rg(:,:,:)
    integer:: ldkg
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: nwk(:,:),idcwk(:,:)
    integer,pointer:: g_count(:,:),g_displ(:,:)
    type (irdkpt_struct),pointer:: irdkpt(:)

    real(kind=8):: va,bb(3,3),bzwk(mgs%nks)
    integer:: is

    va = cellvol(aa)
    call recp_uvec(aa,bb)

    nks = mgs%nks
    nkmesh = mgs%nkmesh
    nrcell = mgs%nrcell
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz
    nrxyz = mgs%nrxyz

    ng = mgs%ng
    nnp = mgs%nnp
    rg => mgs%rg
    pg => mgs%pg

    irdkpt => mgs%irdkpt
    nkm => mgs%nkm
    kg => mgs%kg
    nwk => mgs%nwk
    idcwk => mgs%idcwk

    ldkg = mgs%ldkg
    g_count => mgs%g_count
    g_displ => mgs%g_displ

    bzwk(:) = irdkpt(:)%ngir
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if

    ekinor = 0.0d0
    sekinor = 0.0d0
    if (nspin.eq.1) then
       call ekinor_core(dp,ekinor,sekinor,bzwk,wgtk,va,aa,bb,nkmesh &
            ,ncomp,nrx,nry,nrz,nrxyz &
            ,ng,nnp,rg,pg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
            ,myrank,nproc,g_count,g_displ,comm)
    else
       do is=1,nspin
          call ekinor_core(dp(1,is),ekinor,sekinor,bzwk,wgtk(nb1,1,is) &
               ,va,aa,bb,nkmesh,ncomp,nrx,nry,nrz,nrxyz &
               ,ng,nnp,rg,pg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
               ,myrank,nproc,g_count,g_displ,comm)
       end do
    end if
    return
  end subroutine calc_ekinor

  subroutine ekinor_core(dp,ekinor,sekinor,bzwk,wgtk &
       ,va,aa,bb,nkmesh,ncomp,nrx,nry,nrz,nrxyz &
       ,ng,nnp,rg,pg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
       ,myrank,nproc,g_count,g_displ,comm)
    !
    !                     1992. JAN. 30. BY J.YAMAUCHI
    !
    use mpi
    use xmpi
    use fft_3d
    use symchrg
    use types
    use timeprof
    use subr_kpoint
    implicit none
    integer:: nkmesh,nrx,nry,nrz,nrxyz,ng,nnp,nrcell
    integer:: rg(3,3,ng),pg(3,ng)
    integer:: nks,nb1,nb2
    integer:: ncomp
    real(kind=8),intent(out):: ekinor(nrx*2,nry*2,nrz*2)
    real(kind=8),intent(out):: sekinor(nrx*2,nry*2,nrz*2,6)
    real(kind=8):: bzwk(nks),wgtk(nb1:nb2,nks)
    real(kind=8):: va,aa(3,3),bb(3,3)
    integer:: nkm(nks),ldkg,kg(3,ldkg,nks)
    integer,intent(in):: nwk(nrxyz,3),idcwk(nrcell,2)
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer:: myrank,nproc,comm
    integer:: mpi_typ_wfn
    integer:: g_count(0:nproc-1,nks),g_displ(0:nproc-1,nks)
    !
    type (wfn_for_k):: dp(nks)
    !
    type (fft3_struct):: fs
    real(kind=8),allocatable:: xekinor(:,:,:),pxekinor(:,:,:)
    real(kind=8),allocatable:: xsekinor(:,:,:,:),pxsekinor(:,:,:,:)
    integer:: ierr,ks,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    real(kind=8):: fac
    integer:: bkm,ekm,irnk
    integer,allocatable:: b_count(:),b_displ(:)
!soption contarea(wf)
    complex(kind=8),pointer:: wf(:,:,:)
    integer:: i1,i2,i3

    call tpf_begin(tpf_bchrg)
    !
    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2

    nlx = FFT_PAD_NX(nrx2)
    nly = FFT_PAD_NY(nry2)
    nlz = FFT_PAD_NZ(nrz2)

    nlxyz=nlx*nly*nlz
    !
    allocate(xekinor(nlx,nly,nlz),pxekinor(nlx,nly,nlz),stat=ierr)
    call chkalloc('ekinor_core','tmp',ierr)
    allocate(xsekinor(6,nlx,nly,nlz),pxsekinor(6,nlx,nly,nlz),stat=ierr)
    call chkalloc('ekinor_core','tmp2',ierr)
    !
    allocate(b_count(0:nproc-1),b_displ(0:nproc-1),stat=ierr)
    call chkalloc('ekinor_core','b_count',ierr)

    b_count = nb2/nproc
    do irnk=0,mod(nb2,nproc)-1
       b_count(irnk) = b_count(irnk)+1
    end do
    b_displ(0) = 0
    do irnk=0,nproc-2
       b_displ(irnk+1) = b_displ(irnk) + b_count(irnk)
    end do
    !
    call fft3_init(nrx2,nry2,nrz2,nlx,nly,nlz,fs)

    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    pxekinor = 0.0d0
    pxsekinor = 0.0d0
    do ks=1,nks
       bkm = g_displ(myrank,ks)+1
       ekm = g_displ(myrank,ks)+g_count(myrank,ks)
       wf => dp(ks)%wf

       call calc_ekinor1(ncomp,bkm,ekm,nkm(ks),wf,kg(:,:,ks) &
            ,fs,bzwk(ks),wgtk(nb1,ks),pxekinor,pxsekinor &
            ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
            ,nb1,nb2,irdkpt(ks)%sk,bb,myrank,nproc &
            ,b_count,b_displ,g_count(0,ks),g_displ(0,ks),mpi_typ_wfn,comm)
    end do

    call xmpi_alldsum(pxekinor,xekinor,nlxyz,comm,ierr)
    call xmpi_alldsum(pxsekinor,xsekinor,6*nlxyz,comm,ierr)
    
    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)
    !
    ! after ysymch(), the correct kinetic energy density is obtained.
    ! after symstrsor(), the correct kinetic stress density is obtained.
    !
    call ysymch(xekinor,nwk,idcwk,nlx,nly,nlz,nrxyz,ng,nrcell)
    call symstrsor(xsekinor,nlx,nly,nlz,nrx2,nry2,nrz2,ng,nnp,rg,pg,aa)

    fac=1.0d0/dble(nkmesh)/va
    ekinor = ekinor + xekinor(1:nrx2,1:nry2,1:nrz2)*fac
!$omp parallel do private(i1,i2,i3)
    do i3 = 1,nrz*2
       do i2 = 1,nry*2
          do i1 = 1,nrx*2
             sekinor(i1,i2,i3,1) = sekinor(i1,i2,i3,1) &
                  + xsekinor(1,i1,i2,i3)*fac
             sekinor(i1,i2,i3,2) = sekinor(i1,i2,i3,2) &
                  + xsekinor(2,i1,i2,i3)*fac
             sekinor(i1,i2,i3,3) = sekinor(i1,i2,i3,3) &
                  + xsekinor(3,i1,i2,i3)*fac
             sekinor(i1,i2,i3,4) = sekinor(i1,i2,i3,4) &
                  + xsekinor(4,i1,i2,i3)*fac
             sekinor(i1,i2,i3,5) = sekinor(i1,i2,i3,5) &
                  + xsekinor(5,i1,i2,i3)*fac
             sekinor(i1,i2,i3,6) = sekinor(i1,i2,i3,6) &
                  + xsekinor(6,i1,i2,i3)*fac
          end do
       end do
    end do
    !
    call fft3_done(fs)
    deallocate(b_count,b_displ)
    deallocate(xekinor,pxekinor,xsekinor,pxsekinor)
    !
    call tpf_end(tpf_bchrg)
    return
  end subroutine ekinor_core

  subroutine calc_ekinor1(ncomp,bkm,ekm,nkm,dp,kg,fs,bzwk,wgtk &
       ,ekinor,sekinor &
       ,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz &
       ,nb1,nb2,sk,bb,myrank,nproc &
       ,b_count,b_displ,g_count,g_displ,mpi_typ_wfn,comm)
    use mpi
    use fft_3d
    use timeprof
    implicit none
    type (fft3_struct):: fs
    integer:: ncomp
    integer:: nrx2,nry2,nrz2,nb1,nb2
    integer:: nlx,nly,nlz,nlxyz
    real(kind=8):: sk(3),bb(3,3)
    real(kind=8):: ekinor(nlxyz),sekinor(6,nlxyz),bzwk,wgtk(nb1:nb2)
    integer:: nkm,bkm,ekm,kg(3,nkm)
    complex(kind=8):: dp(ncomp,bkm:ekm,nb2)
    integer:: mpi_typ_wfn
    integer:: myrank,nproc,comm
    integer,dimension(0:nproc-1):: b_count,b_displ,g_count,g_displ
    !
    real(kind=8),allocatable:: ztmp(:,:,:),wtmp(:,:)
    integer:: ierr,ib,irnk,ig

    complex(kind=8),allocatable:: tp(:,:)
    integer,allocatable:: s_count(:),s_displ(:)
    integer,allocatable:: r_count(:),r_displ(:)
    real(kind=8):: fac
    integer:: ic,iax
    !
    allocate(FFT_A1M_DEFSZ(ztmp,nlxyz,3), &
         FFT_A1D_DEFSZ(wtmp,nlxyz),stat=ierr)
    call chkalloc('calc_ekior1','tmp',ierr)
    !
    allocate(s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('calc_ekior1','s_count',ierr)

    allocate(tp(ncomp,nkm*b_count(myrank)),stat=ierr)
    call chkalloc('calc_ekior1','tp',ierr)

    s_count = g_count(myrank)*b_count
    s_displ = g_count(myrank)*b_displ

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER &
         ,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,mpi_typ_wfn, &
         tp,r_count,r_displ,mpi_typ_wfn,comm,ierr)
    call tpf_end(tpf_mpi)
    !
!CDIR RESERVE
    do ib=0,b_count(myrank)-1
       do ic=1,ncomp
          do iax=1,3
!$omp parallel do
             do ig=1,nlxyz
                FFT_A1M_RL(ztmp,ig,iax)=0.0d0
                FFT_A1M_IM(ztmp,ig,iax)=0.0d0
             end do
          end do
          call gradwfgen(ztmp,sk,bb,kg,tp,nkm*b_count(myrank), &
               nrx2,nry2,nrz2,nlx,nly,nlz,ncomp, &
               nkm,ic,ib,nproc,g_count,g_displ,r_displ)
          do iax=1,3
             call fft3_bw(fs,FFT_A1M_RL(ztmp,1,iax),wtmp)
          end do
          if (ib+b_displ(myrank)+1 .lt. nb1) then
             fac = bzwk
          else
             fac = wgtk(ib+b_displ(myrank)+1)
          end if
          call gradwf2ekinor(ztmp,ekinor,sekinor,fac,nlxyz)
       end do
    end do
!CDIR RELEASE
    deallocate(tp)
    deallocate(s_count,s_displ,r_count,r_displ)
    deallocate(ztmp,wtmp)
    return
  end subroutine calc_ekinor1

  subroutine gradwfgen(gradwf,sk,bb,kg,tp,ntp,nrx2,nry2,nrz2,nlx,nly,nlz, &
       ncomp,nkm,ic,ib,nproc,g_count,g_displ,r_displ)
    use timeprof
    implicit none
    real(kind=8),intent(in):: sk(3),bb(3,3)
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,ncomp,nkm,ntp,ic,ib
    integer,intent(in):: nproc
    integer,intent(in):: g_count(0:nproc-1),g_displ(0:nproc-1)
    integer,intent(in):: r_displ(0:nproc-1)
    !
    ! gradwf must be zero cleared before passed to this subroutine
    !
    real(kind=8):: FFT_A3M_DEFSZ(gradwf,nlx,nly,nlz,3)
    complex(kind=8),intent(in):: tp(ncomp,ntp)
    integer,intent(in):: kg(3,nkm)

    integer:: k1,k2,k3,ig,jg,k,irnk
    real(kind=8):: fac(3)
    integer:: j1,j2,j3,iax

    call tpf_begin(tpf_scgt)

    do irnk=0,nproc-1
       jg=r_displ(irnk)+g_count(irnk)*ib
       ig=g_displ(irnk)
!$omp parallel do private(k1,k2,k3,j1,j2,j3,fac)
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
!POPTION INDEP(rho)
       do k=1,g_count(irnk)
          k1 = kg(1,ig+k)
          k2 = kg(2,ig+k)
          k3 = kg(3,ig+k)
          fac(1) = bb(1,1)*(sk(1)+k1) &
               + bb(1,2)*(sk(2)+k2) + bb(1,3)*(sk(3)+k3)
          fac(2) = bb(2,1)*(sk(1)+k1) &
               + bb(2,2)*(sk(2)+k2) + bb(2,3)*(sk(3)+k3)
          fac(3) = bb(3,1)*(sk(1)+k1) &
               + bb(3,2)*(sk(2)+k2) + bb(3,3)*(sk(3)+k3)
          j1 = mod(nrx2+k1,nrx2)+1
          j2 = mod(nry2+k2,nry2)+1
          j3 = mod(nrz2+k3,nrz2)+1
          do iax=1,3
             FFT_A3M_RL(gradwf,j1,j2,j3,iax) = -fac(iax)*aimag(tp(ic,jg+k))
             FFT_A3M_IM(gradwf,j1,j2,j3,iax) = fac(iax)*dble(tp(ic,jg+k))
          end do
       end do
    end do
    call tpf_end(tpf_scgt)
    return
  end subroutine gradwfgen

  subroutine gradwf2ekinor(gradwf,ekinor,sekinor,rw,nlxyz)
    implicit none
    integer,intent(in):: nlxyz
    real(kind=8):: rw
    real(kind=8):: ekinor(nlxyz),sekinor(6,nlxyz)
    real(kind=8):: FFT_A1M_DEFSZ(gradwf,nlxyz,3)

    integer:: i

!$omp parallel do private(i)
    do i=1,nlxyz
       ekinor(i) = ekinor(i) + 0.5d0*rw*( &
             FFT_A1M_RL(gradwf,i,1)*FFT_A1M_RL(gradwf,i,1) &
           + FFT_A1M_IM(gradwf,i,1)*FFT_A1M_IM(gradwf,i,1) &
           + FFT_A1M_RL(gradwf,i,2)*FFT_A1M_RL(gradwf,i,2) &
           + FFT_A1M_IM(gradwf,i,2)*FFT_A1M_IM(gradwf,i,2) &
           + FFT_A1M_RL(gradwf,i,3)*FFT_A1M_RL(gradwf,i,3) &
           + FFT_A1M_IM(gradwf,i,3)*FFT_A1M_IM(gradwf,i,3) )

       sekinor(1,i) = sekinor(1,i) + rw*( &
            FFT_A1M_RL(gradwf,i,1)*FFT_A1M_RL(gradwf,i,1) &
          + FFT_A1M_IM(gradwf,i,1)*FFT_A1M_IM(gradwf,i,1) &
            )
       sekinor(2,i) = sekinor(2,i) + rw*( &
            FFT_A1M_RL(gradwf,i,2)*FFT_A1M_RL(gradwf,i,2) &
          + FFT_A1M_IM(gradwf,i,2)*FFT_A1M_IM(gradwf,i,2) &
            )
       sekinor(3,i) = sekinor(3,i) + rw*( &
            FFT_A1M_RL(gradwf,i,3)*FFT_A1M_RL(gradwf,i,3) &
          + FFT_A1M_IM(gradwf,i,3)*FFT_A1M_IM(gradwf,i,3) &
            )
       sekinor(4,i) = sekinor(4,i) + rw*( &
            FFT_A1M_RL(gradwf,i,2)*FFT_A1M_RL(gradwf,i,3) &
          + FFT_A1M_IM(gradwf,i,2)*FFT_A1M_IM(gradwf,i,3) &
          + FFT_A1M_RL(gradwf,i,3)*FFT_A1M_RL(gradwf,i,2) &
          + FFT_A1M_IM(gradwf,i,3)*FFT_A1M_IM(gradwf,i,2) &
            )
       sekinor(5,i) = sekinor(5,i) + rw*( &
            FFT_A1M_RL(gradwf,i,3)*FFT_A1M_RL(gradwf,i,1) &
          + FFT_A1M_IM(gradwf,i,3)*FFT_A1M_IM(gradwf,i,1) &
          + FFT_A1M_RL(gradwf,i,1)*FFT_A1M_RL(gradwf,i,3) &
          + FFT_A1M_IM(gradwf,i,1)*FFT_A1M_IM(gradwf,i,3) &
            )
       sekinor(6,i) = sekinor(6,i) + rw*( &
            FFT_A1M_RL(gradwf,i,1)*FFT_A1M_RL(gradwf,i,2) &
          + FFT_A1M_IM(gradwf,i,1)*FFT_A1M_IM(gradwf,i,2) &
          + FFT_A1M_RL(gradwf,i,2)*FFT_A1M_RL(gradwf,i,1) &
          + FFT_A1M_IM(gradwf,i,2)*FFT_A1M_IM(gradwf,i,1) &
            )
    end do
    return
  end subroutine gradwf2ekinor

end module subr_ekinor
