!
! Copyright (c) 2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module subr_engpart
  type eng_part_struct
     real(kind=8):: qfengpar
     integer:: mode, mhtor
     integer:: num_bin_wgt
     real(kind=8),pointer:: refsz(:)
  end type eng_part_struct
contains
  subroutine init_engparc(engparc)
    implicit none
    type (eng_part_struct):: engparc
    integer:: ierr
    engparc%mode = 0
    engparc%qfengpar = -1.0
    engparc%mhtor = 1
    engparc%num_bin_wgt = 10
    nullify(engparc%refsz)
    return
  end subroutine init_engparc

  subroutine reset_engparc(engparc)
    implicit none
    type (eng_part_struct):: engparc
    integer:: ierr
    engparc%mode = 0
    engparc%qfengpar = -1.0
    engparc%mhtor = 1
    engparc%num_bin_wgt = 10
    if (associated(engparc%refsz)) then
       deallocate(engparc%refsz,stat=ierr)
       if (ierr.ne.0) then
          write(6,*) 'reset_engparc() failed to deallocate(refsz)' // &
               ' although refsz is associated'
          stop
       end if
    end if
    return
  end subroutine reset_engparc
  
  subroutine set1partwgt(nx2,ny2,nz2,idxpar,it,wgtpar)
    implicit none
    integer,intent(in):: nx2,ny2,nz2
    integer,intent(in):: idxpar(nx2,ny2,nz2)
    integer,intent(in):: it
    real(kind=8):: wgtpar(nx2,ny2,nz2)

    integer:: j1,j2,j3

!$omp parallel do private(j1,j2,j3)
    do j3=1,nz2
       do j2=1,ny2
          do j1=1,nx2
             if (idxpar(j1,j2,j3).eq.it) then
                wgtpar(j1,j2,j3) = 1.0d0
             else
                wgtpar(j1,j2,j3) = 0.0d0
             end if
          end do
       end do
    end do

    return
  end subroutine set1partwgt

  subroutine set1partwgt_fltr(bb,nkfxa,kf,nx2,ny2,nz2,idxpar,qf,it,wgtpar)
    use fft_3d
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nkfxa
    integer,intent(in):: kf(3,nkfxa)
    integer,intent(in):: nx2,ny2,nz2
    integer,intent(in):: idxpar(nx2,ny2,nz2)
    real(kind=8),intent(in):: qf
    integer,intent(in):: it
    real(kind=8):: wgtpar(nx2,ny2,nz2)

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    type (fft3_struct):: fs
    real(kind=8),allocatable:: ztmp(:,:,:,:),wtmp(:,:,:,:)
    integer:: ig,j1,j2,j3,k1,k2,k3
    real(kind=8):: fac,sq,x(3)

    call fft3_init(nx2,ny2,nz2,nx2,ny2,nz2,fs)
    allocate(FFT_A3D_DEFSZ(ztmp,nx2,ny2,nz2), &
         FFT_A3D_DEFSZ(wtmp,nx2,ny2,nz2))
    wtmp = 0.0d0

    ztmp = 0.0d0
    do j3=1,nz2
       do j2=1,ny2
          do j1=1,nx2
             if (idxpar(j1,j2,j3).eq.it) then
                FFT_A3D_RL(ztmp,j1,j2,j3) = 1.0d0
             end if
          end do
       end do
    end do

    call fft3_fw(fs,ztmp,wtmp)
    wtmp = 0.0d0
    do ig=1,nkfxa
       k1 = kf(1,ig)
       k2 = kf(2,ig)
       k3 = kf(3,ig)
       j1 = mod(nx2+k1,nx2) + 1
       j2 = mod(ny2+k2,ny2) + 1
       j3 = mod(nz2+k3,nz2) + 1
       x(1) = bb(1,1)*k1 + bb(1,2)*k2 + bb(1,3)*k3
       x(2) = bb(2,1)*k1 + bb(2,2)*k2 + bb(2,3)*k3
       x(3) = bb(3,1)*k1 + bb(3,2)*k2 + bb(3,3)*k3
       sq = sqrt(x(1)*x(1)+x(2)*x(2)+x(3)*x(3))
       !
       ! Hann window
       !
       !if (sq.lt.qf) then
       !   fac = 0.5d0*(1.0d0+cos(pi*sq/qf))
       !else
       !   fac = 0.0d0
       !end if
       !
       ! Blackman window
       !
       if (sq.lt.qf) then
          fac = 0.42d0 + 0.5d0*cos(pi*sq/qf) + 0.08d0*cos(2.0d0*pi*sq/qf)
       else
          fac = 0.0d0
       end if
       FFT_A3D_RL(wtmp,j1,j2,j3) = fac*FFT_A3D_RL(ztmp,j1,j2,j3)
       FFT_A3D_IM(wtmp,j1,j2,j3) = fac*FFT_A3D_IM(ztmp,j1,j2,j3)
    end do
    call fft3_bw(fs,wtmp,ztmp)

    wgtpar(:,:,:) = FFT_A3D_RL(wtmp,:,:,:)

    call fft3_done(fs)
    return
  end subroutine set1partwgt_fltr

  subroutine setpartwgt(bb,nkfxa,kf,nx2,ny2,nz2,idxpar,nsi,qf,wgtpar)
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nkfxa
    integer,intent(in):: kf(3,nkfxa)
    integer,intent(in):: nx2,ny2,nz2
    integer,intent(in):: idxpar(nx2,ny2,nz2)
    integer,intent(in):: nsi
    real(kind=8),intent(in):: qf
    real(kind=8):: wgtpar(nx2,ny2,nz2,nsi)

    integer:: it

    do it=1,nsi
       call set1partwgt_fltr(bb,nkfxa,kf,nx2,ny2,nz2,idxpar,qf, &
            it,wgtpar(:,:,:,it))
    end do

    return
  end subroutine setpartwgt

  subroutine calcvoronorm(aa,nsi,kd,asi,nkd,qf,nx2,ny2,nz2,refsz,voronorm)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi),nkd
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(in):: qf
    integer,intent(in):: nx2,ny2,nz2
    real(kind=8),intent(in):: refsz(nkd)
    real(kind=8):: voronorm(nx2,ny2,nz2)

    integer:: it,j1,j2,j3,isft
    real(kind=8):: x(3),d(3),cd(3),trsq,r,fac

    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    integer,parameter:: sft(3,8) = reshape( &
         (/   0, 0, 0, -1, 0, 0,  0,-1, 0, &
             -1,-1, 0,  0, 0,-1, -1, 0,-1, &
              0,-1,-1, -1,-1,-1 /), (/ 3, 8/) )

    fac = qf/pi
    voronorm = 0.0d0
!$omp parallel do private(it,j1,j2,j3,isft,x,d,cd,trsq,r)
    do j3=1,nz2
       do j2=1,ny2
          do j1=1,nx2
             x(1) = dble(j1-1)/nx2
             x(2) = dble(j2-1)/ny2
             x(3) = dble(j3-1)/nz2
             do it=1,nsi
                r = huge(1.0d0)
                do isft = 1,8
                   d = modulo(x - asi(:,it),1.0d0) + sft(:,isft)
                   cd = matmul(aa,d)
                   trsq = cd(1)*cd(1) + cd(2)*cd(2) + cd(3)*cd(3)
                   r = min(r,sqrt(trsq))
                end do
                voronorm(j1,j2,j3) = voronorm(j1,j2,j3) &
                     + exp(-fac*(r-refsz(kd(it))))
             end do
          end do
       end do
    end do
    return
  end subroutine calcvoronorm

  subroutine set1partwgt_voro(aa,nsi,kd,asi,nkd,nx2,ny2,nz2,refsz, &
       voronorm,qf,it,wgtpar)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi),nkd
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: nx2,ny2,nz2
    real(kind=8),intent(in):: refsz(nkd)
    real(kind=8),intent(in):: voronorm(nx2,ny2,nz2)
    real(kind=8),intent(in):: qf
    integer,intent(in):: it
    real(kind=8):: wgtpar(nx2,ny2,nz2)
    
    integer:: j1,j2,j3,isft
    real(kind=8):: x(3),d(3),cd(3),trsq,r,fac

    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    integer,parameter:: sft(3,8) = reshape( &
         (/   0, 0, 0, -1, 0, 0,  0,-1, 0, &
             -1,-1, 0,  0, 0,-1, -1, 0,-1, &
              0,-1,-1, -1,-1,-1 /), (/ 3, 8/) )

    fac = qf/pi
!$omp parallel do private(j1,j2,j3,x,r,isft,d,cd,trsq)
    do j3=1,nz2
       do j2=1,ny2
          do j1=1,nx2
             x(1) = dble(j1-1)/nx2
             x(2) = dble(j2-1)/ny2
             x(3) = dble(j3-1)/nz2
             r = huge(1.0d0)
             do isft = 1,8
                d = modulo(x - asi(:,it),1.0d0) + sft(:,isft)
                cd = matmul(aa,d)
                trsq = cd(1)*cd(1) + cd(2)*cd(2) + cd(3)*cd(3)
                r = min(r,sqrt(trsq))
             end do
             wgtpar(j1,j2,j3) = exp(-fac*(r-refsz(kd(it))))/voronorm(j1,j2,j3)
          end do
       end do
    end do
    return
  end subroutine set1partwgt_voro

  subroutine statwgt(nx2,ny2,nz2,nbin,wgtpar,wgtstat)
    implicit none
    integer,intent(in):: nx2,ny2,nz2,nbin
    real(kind=8),intent(in):: wgtpar(nx2,ny2,nz2)
    real(kind=8):: wgtstat(nbin)
    integer:: j1,j2,j3,jbin

!$omp parallel do private(j1,j2,j3,jbin),reduction(+:wgtstat)
    do j3=1,nz2
       do j2=1,ny2
          do j1=1,nx2
             jbin = ceiling(wgtpar(j1,j2,j3)*nbin)
             jbin = min(jbin,nbin) ! make it sure
             wgtstat(jbin) = wgtstat(jbin) + 1.0d0
          end do
       end do
    end do
    return
  end subroutine statwgt

  subroutine calctotengatm(nsi,engatm,sengatm,sum_etot,sum_setot)
    use types
    implicit none
    integer,intent(in):: nsi
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm

    integer:: it
    real(kind=8):: sum_etot,sum_setot(6)
    
    sum_etot = 0.0d0
    sum_setot = 0.0d0
    do it=1,nsi
       engatm%total(it) = engatm%eht(it) + engatm%exc(it) &
            + engatm%epotl(it) + engatm%epotnl(it) + engatm%ekin(it) &
            + engatm%ewald(it) ! + engatm%exx(it) + engatm%axeng(it)

       sengatm%total(1:3,it) = sengatm%sekin(1:3,it) &
            + engatm%epotl(it) + engatm%epotnl(it) + engatm%eht(it) &
            - engatm%exc(it) &
            + engatm%evxc(0,it) + engatm%evxc(1:3,it) &
            + sengatm%sepotl(1:3,it) &
            + sengatm%selptcr(1:3,it) + sengatm%seht(1:3,it) &
            + sengatm%sepotnl(1:3,it) + sengatm%sewald(1:3,it)
       ! + sengatm%sexx(1:3,it)

       sengatm%total(4:6,it) = sengatm%sekin(4:6,it) &
            + engatm%evxc(4:6,it) &
            + sengatm%sepotl(4:6,it) &
            + sengatm%selptcr(4:6,it) + sengatm%seht(4:6,it) &
            + sengatm%sepotnl(4:6,it) + sengatm%sewald(4:6,it)
       ! + sengatm%sexx(4:6,it)

       ! sengatm%total(1:6,it) = sengatm%total(1:6,it) + sengatm%axseng(1:6,it)

       sum_etot = sum_etot + engatm%total(it)
       sum_setot = sum_setot + sengatm%total(:,it)
    end do
    return
  end subroutine calctotengatm

  subroutine calcvoroengpart(aa,nsi,kd,asi,nkd,nrx2,nry2,nrz2, &
       ekinor,excor,evxcor,hrtor,sekinor,shrtor,rho, &
       engatm,sengatm,engparc,mssg)
    use util
    use types
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi,kd(nsi),nkd
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: nrx2,nry2,nrz2
    real(kind=8),intent(in):: ekinor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: excor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: evxcor(nrx2,nry2,nrz2,0:6)
    real(kind=8),intent(in):: hrtor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: sekinor(nrx2,nry2,nrz2,6)
    real(kind=8),intent(in):: shrtor(nrx2,nry2,nrz2,6)
    real(kind=8),intent(in):: rho(nrx2,nry2,nrz2)

    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm
    type (eng_part_struct):: engparc
    integer,intent(in):: mssg
    
    real(kind=8),allocatable:: voronorm(:,:,:),wgtpar(:,:,:)

    real(kind=8):: va
    integer:: it,j1,j2,j3,nrxyz
    real(kind=8):: tekin,texc,tevxc(0:6),teht,trho
    real(kind=8):: tsekin(6),tseht(6)
    real(kind=8):: sum_etot,sum_setot(6)
    real(kind=8):: wgtstat(engparc%num_bin_wgt)

    va = cellvol(aa)
    nrxyz = nrx2*nry2*nrz2
    
    allocate(voronorm(nrx2,nry2,nrz2),wgtpar(nrx2,nry2,nrz2))

    call calcvoronorm(aa,nsi,kd,asi,nkd,engparc%qfengpar,nrx2,nry2,nrz2, &
         engparc%refsz,voronorm)
    
    do it = 1,nsi
       call set1partwgt_voro(aa,nsi,kd,asi,nkd,nrx2,nry2,nrz2,engparc%refsz, &
            voronorm,engparc%qfengpar,it,wgtpar)
       call statwgt(nrx2,nry2,nrz2,engparc%num_bin_wgt,wgtpar,wgtstat)
       tekin = 0.0d0
       texc = 0.0d0
       tevxc = 0.0d0
       teht = 0.0d0
       tsekin = 0.0d0
       tseht = 0.0d0
       trho = 0.0d0
!$omp parallel do private(j1,j2,j3), &
!$omp reduction(+:tekin,texc,tevxc,teht,tsekin,tseht,trho)
       do j3=1,nrz2
          do j2=1,nry2
             do j1=1,nrx2
                tekin = tekin + wgtpar(j1,j2,j3)*ekinor(j1,j2,j3)
                texc = texc + wgtpar(j1,j2,j3)*excor(j1,j2,j3)
                tevxc = tevxc + wgtpar(j1,j2,j3)*evxcor(j1,j2,j3,:)
                teht = teht + wgtpar(j1,j2,j3)*hrtor(j1,j2,j3)

                tsekin = tsekin + wgtpar(j1,j2,j3)*sekinor(j1,j2,j3,:)
                tseht = tseht + wgtpar(j1,j2,j3)*shrtor(j1,j2,j3,:)

                trho = trho + wgtpar(j1,j2,j3)*rho(j1,j2,j3)
             end do
          end do
       end do
       engatm%ekin(it) = tekin
       engatm%exc(it) = texc
       engatm%evxc(:,it) = tevxc
       engatm%eht(it) = teht

       sengatm%sekin(:,it) = tsekin
       sengatm%seht(:,it) = tseht

       trho = trho*va/nrxyz
    end do
    engatm%ekin = engatm%ekin*va/nrxyz
    engatm%exc = engatm%exc*va/nrxyz
    engatm%evxc = engatm%evxc*va/nrxyz
    engatm%eht = engatm%eht*va/nrxyz
    sengatm%sekin = sengatm%sekin*va/nrxyz
    sengatm%seht = sengatm%seht*va/nrxyz

    wgtstat = wgtstat/(nrxyz)
    deallocate(voronorm,wgtpar)

    call calctotengatm(nsi,engatm,sengatm,sum_etot,sum_setot)

    if (mssg.gt.-1) then
       write(6,'(''# sum total energy '',1x,e21.12)') sum_etot
       write(6,'(''# sum stress'',6(1x,e21.12))') sum_setot
       write(6,'(/a)') '# wgtstat'
       write(6,'(1x,f9.3,9(1x,f5.3))') wgtstat
    end if
    
    return
  end subroutine calcvoroengpart

  subroutine calcidxparengpart(aa,nsi,nrx2,nry2,nrz2, &
       ekinor,excor,evxcor,hrtor,sekinor,shrtor,rho, &
       engatm,sengatm,idxpar,idxcar2xtp,engparc,mssg)
    use util
    use types
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: nsi
    integer,intent(in):: nrx2,nry2,nrz2
    real(kind=8),intent(in):: ekinor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: excor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: evxcor(nrx2,nry2,nrz2,0:6)
    real(kind=8),intent(in):: hrtor(nrx2,nry2,nrz2)
    real(kind=8),intent(in):: sekinor(nrx2,nry2,nrz2,6)
    real(kind=8),intent(in):: shrtor(nrx2,nry2,nrz2,6)
    real(kind=8),intent(in):: rho(nrx2,nry2,nrz2)
    integer,intent(in):: idxpar(nrx2,nry2,nrz2),idxcar2xtp(nsi)

    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm
    type (eng_part_struct):: engparc
    integer,intent(in):: mssg
    
    real(kind=8),allocatable:: wgtpar(:,:,:)

    real(kind=8):: va
    integer:: it,j1,j2,j3,nrxyz
    real(kind=8):: tekin,texc,tevxc(0:6),teht,trho
    real(kind=8):: tsekin(6),tseht(6)
    real(kind=8):: sum_etot,sum_setot(6)
    real(kind=8):: wgtstat(engparc%num_bin_wgt)
    integer:: itcar

    va = cellvol(aa)
    nrxyz = nrx2*nry2*nrz2
    allocate(wgtpar(nrx2,nry2,nrz2))

    engatm%ekin = 0.0d0
    engatm%exc = 0.0d0
    engatm%evxc = 0.0d0
    engatm%eht = 0.0d0
    sengatm%sekin = 0.0d0
    sengatm%seht = 0.0d0
    wgtstat = 0.0d0
    write(6,'(/1x,a)') 'atom #, weight for atom, rho for atom'
    do itcar=1,nsi
       it = idxcar2xtp(itcar)
       call set1partwgt(nrx2,nry2,nrz2,idxpar,itcar,wgtpar)
       call statwgt(nrx2,nry2,nrz2,engparc%num_bin_wgt,wgtpar,wgtstat)
       tekin = 0.0d0
       texc = 0.0d0
       tevxc = 0.0d0
       teht = 0.0d0
       tsekin = 0.0d0
       tseht = 0.0d0
       trho = 0.0d0
!$omp parallel do private(j1,j2,j3), &
!$omp reduction(+:tekin,texc,tevxc,teht,tsekin,tseht,trho)
       do j3=1,nrz2
          do j2=1,nry2
             do j1=1,nrx2
                tekin = tekin + wgtpar(j1,j2,j3)*ekinor(j1,j2,j3)
                texc = texc + wgtpar(j1,j2,j3)*excor(j1,j2,j3)
                tevxc = tevxc + wgtpar(j1,j2,j3)*evxcor(j1,j2,j3,:)
                teht = teht + wgtpar(j1,j2,j3)*hrtor(j1,j2,j3)

                tsekin = tsekin + wgtpar(j1,j2,j3)*sekinor(j1,j2,j3,:)
                tseht = tseht + wgtpar(j1,j2,j3)*shrtor(j1,j2,j3,:)

                trho = trho + wgtpar(j1,j2,j3)*rho(j1,j2,j3)
             end do
          end do
       end do
       engatm%ekin(it) = engatm%ekin(it) + tekin
       engatm%exc(it) = engatm%exc(it) + texc
       engatm%evxc(:,it) = engatm%evxc(:,it) + tevxc
       engatm%eht(it) = engatm%eht(it) + teht

       sengatm%sekin(:,it) = sengatm%sekin(:,it) + tsekin
       sengatm%seht(:,it) = sengatm%seht(:,it) + tseht

       trho = trho*va/nrxyz
       write(6,'(1x,i4,2(1x,f12.7))') it, sum(wgtpar)*va/nrxyz, trho
    end do
    engatm%ekin = engatm%ekin*va/nrxyz
    engatm%exc = engatm%exc*va/nrxyz
    engatm%evxc = engatm%evxc*va/nrxyz
    engatm%eht = engatm%eht*va/nrxyz
    sengatm%sekin = sengatm%sekin*va/nrxyz
    sengatm%seht = sengatm%seht*va/nrxyz

    wgtstat = wgtstat/(nrxyz)

    call calctotengatm(nsi,engatm,sengatm,sum_etot,sum_setot)

    if (mssg.gt.-1) then
       write(6,'(''# sum total energy '',1x,e21.12)') sum_etot
       write(6,'(''# sum stress'',6(1x,e21.12))') sum_setot
       write(6,'(/a)') '# wgtstat'
       write(6,'(1x,f9.3,9(1x,f5.3))') wgtstat
    end if
    
    return
  end subroutine calcidxparengpart
  
  subroutine calctrinkengpart(nsi,engatm,sengatm,mssg)
    use util
    use types
    implicit none
    integer,intent(in):: nsi
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm
    
    real(kind=8):: sum_etot,sum_setot(6)
    integer,intent(in):: mssg

    call calctotengatm(nsi,engatm,sengatm,sum_etot,sum_setot)
    if (mssg.gt.-1) then
       write(6,'(''# sum total energy '',1x,e21.12)') sum_etot
       write(6,'(''# sum stress'',6(1x,e21.12))') sum_setot
    end if
    return
  end subroutine calctrinkengpart

  subroutine energy_per_atom_store(nf,nsi,engatm,sengatm)
    use types
    use iounit
    implicit none
    integer,intent(in):: nf,nsi
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm

    call xrewind(nf)
    write(nf) nsi
    write(nf) engatm%ekin(:),engatm%epotl(:),engatm%epotnl(:), &
         engatm%eht(:),engatm%exc(:),engatm%evxc(:,:), &
         engatm%ewald(:)
    write(nf) sengatm%sekin(:,:),sengatm%sepotl(:,:), &
         sengatm%sepotnl(:,:),sengatm%seht(:,:), &
         sengatm%sewald(:,:),sengatm%selptcr(:,:)
    close(nf)

    return
  end subroutine energy_per_atom_store

  subroutine energy_per_atom_load(nf,nsi,engatm,sengatm)
    use types
    use iounit
    implicit none
    integer,intent(in):: nf,nsi
    type (energy_per_atom):: engatm
    type (senergy_per_atom):: sengatm

    integer:: insi

    call xrewind(nf)
    read(nf) insi
    if (insi.ne.nsi) then
       write(6,*) 'inconsistent number of atom in .loceng file'
       stop
    end if
    call energy_per_atom_alloc(nsi,engatm)
    call senergy_per_atom_alloc(nsi,sengatm)

    read(nf) engatm%ekin(:),engatm%epotl(:),engatm%epotnl(:), &
         engatm%eht(:),engatm%exc(:),engatm%evxc(:,:), &
         engatm%ewald(:)
    read(nf) sengatm%sekin(:,:),sengatm%sepotl(:,:), &
         sengatm%sepotnl(:,:),sengatm%seht(:,:), &
         sengatm%sewald(:,:),sengatm%selptcr(:,:)
    close(nf)

    return
  end subroutine energy_per_atom_load
end module subr_engpart
