!
! Copyright (c) 2014-2015 Yoshihide Yoshimoto
!
program tetrapdos
  use subr_tetra
  implicit none

  real(kind=8),allocatable:: aout(:,:,:),asta(:,:)
  real(kind=8),allocatable:: e(:,:),v(:,:),tskmsh(:,:,:,:,:,:)

  integer,parameter:: nfin = 59
  real(kind=8),parameter:: pi = 3.14159265358979323846d0

  integer:: nkmesh,ndx(3),nsmpl,mode,itwav,nwav
  integer:: nw,nspin
  character(len=256):: cbuf
  integer:: ie,is,nis,ix,iy,iz,ixp1,iyp1,izp1,ib,iis,im
  real(kind=8):: emax,emin,ef,de,fac,s
  integer:: ierr

  call getarg(1, cbuf)
  open(nfin, file=cbuf, status='old', action='read', &
       form='unformatted', iostat=ierr)
  if (ierr.ne.0) then
     write(6,*) 'error in opening '//trim(cbuf)
     stop
  end if

  call getarg(2, cbuf)
  read(cbuf,*) nsmpl
  call getarg(3, cbuf)
  read(cbuf,*) mode

  read(nfin) nkmesh,ndx,nw,nspin,itwav,nwav

  allocate(tskmsh(0:nwav,ndx(1),ndx(2),ndx(3),nw,nspin), &
       aout(0:nsmpl,0:nwav,nspin),asta(nwav,nspin))
  read(nfin) tskmsh

  emax = maxval(tskmsh(0,:,:,:,:,:))
  emin = minval(tskmsh(0,:,:,:,:,:))
  do iz = 1,ndx(3)
     do iy = 1,ndx(2)
        do ix = 1,ndx(1)
           if (tskmsh(0,ix,iy,iz,1,1).eq.huge(1.0d0)) then
              write(6,*) 'invalid data at', ix,iy,iz
           end if
        end do
     end do
  end do

  nis = 0
  do iz = 1,ndx(3)
     do iy = 1,ndx(2)
        do ix = 1,ndx(1)
           nis = nis + 1
        end do
     end do
  end do
  nis = nis*nw

  allocate(e(nis,8),v(nis,8))

  do iis=1,nspin
     do im=0,nwav
        is = 0
        do ib = 1,nw
           do iz = 1,ndx(3)
              izp1 = modulo(iz,ndx(3))+1
              do iy = 1,ndx(2)
                 iyp1 = modulo(iy,ndx(2))+1
                 do ix = 1,ndx(1)
                    ixp1 = modulo(ix,ndx(1))+1
                    is = is + 1
                    if (im.eq.0) then
                       e(is,1)=tskmsh(0,ix  ,iy,  iz,  ib,iis)
                       e(is,2)=tskmsh(0,ix  ,iyp1,iz,  ib,iis)
                       e(is,3)=tskmsh(0,ix  ,iy,  izp1,ib,iis)
                       e(is,4)=tskmsh(0,ix  ,iyp1,izp1,ib,iis)
                       e(is,5)=tskmsh(0,ixp1,iy,  iz,  ib,iis)
                       e(is,6)=tskmsh(0,ixp1,iyp1,iz,  ib,iis)
                       e(is,7)=tskmsh(0,ixp1,iy,  izp1,ib,iis)
                       e(is,8)=tskmsh(0,ixp1,iyp1,izp1,ib,iis)

                       v(is,:) = 1.0d0
                    else
                       v(is,1)=tskmsh(im,ix  ,iy,  iz,  ib,iis)
                       v(is,2)=tskmsh(im,ix  ,iyp1,iz,  ib,iis)
                       v(is,3)=tskmsh(im,ix  ,iy,  izp1,ib,iis)
                       v(is,4)=tskmsh(im,ix  ,iyp1,izp1,ib,iis)
                       v(is,5)=tskmsh(im,ixp1,iy,  iz,  ib,iis)
                       v(is,6)=tskmsh(im,ixp1,iyp1,iz,  ib,iis)
                       v(is,7)=tskmsh(im,ixp1,iy,  izp1,ib,iis)
                       v(is,8)=tskmsh(im,ixp1,iyp1,izp1,ib,iis)
                    end if
                 end do
              end do
           end do
        end do

        if (nspin.eq.1) then
           fac = 2.0d0/dble(ndx(1)*ndx(2)*ndx(3))
        else
           fac = 1.0d0/dble(ndx(1)*ndx(2)*ndx(3))
        end if
        de = (emax-emin)/nsmpl
        do ie = 0, nsmpl
           ef = emin + de*ie
           call tetra0(e(:,1),e(:,2),e(:,3),e(:,4), &
                e(:,5),e(:,6),e(:,7),e(:,8), &
                v(:,1),v(:,2),v(:,3),v(:,4), &
                v(:,5),v(:,6),v(:,7),v(:,8),ef,s,nis,mode)
           aout(ie,im,iis) = fac*s
        end do
     end do
  end do

  write(6,'(1x,a,e18.11)') '# epseng = ', 0.0d0
  write(6,'(1x,a,i4)') '# ifbunp = ', 3
  write(6,'(1x,a,i2)') '# nspin = ', nspin
  write(6,'(1x,a,i5,1x,i5)') '# itwav, nwav = ', itwav, nwav

  asta = 0.0d0
  de = (emax-emin)/nsmpl
  do ie = 0, nsmpl
     ef = emin + de*ie
     if (nspin.eq.1) then
        write(6,'(100(1x,e14.7))') ef, &
             (aout(ie,im,1),im=0,nwav)
     else
        write(6,'(100(1x,e14.7))') ef, &
             (aout(ie,im,1),im=0,nwav), &
             (aout(ie,im,2),im=0,nwav)
     end if
     asta(:,:) = asta(:,:) + aout(ie,1:,:)
  end do
  asta = asta*de
  write(6,'(1x,a)') '# integration of the pdos'
  if (nspin.eq.1) then
     write(6,'(1x,''#'',100(1x,e14.7))') &
          (asta(im,1),im=1,nwav)
  else
     write(6,'(1x,''#'',100(1x,e14.7))') &
          (asta(im,1),im=1,nwav), &
          (asta(im,2),im=1,nwav)
  end if

  write(6,*) '# tetrahedron data'
  write(6,'(a,i8,1x,i2)') &
       ' # nsmpl,mode = ',nsmpl,mode
  write(6,'(a,3(1x,i6))') ' # ndx,ndy,ndz =',ndx(1),ndx(2),ndx(3)
  write(6,'(a,1x,i3,1x,i6,1x,i2)') ' # nwav,nw,nspin =',nwav,nw,nspin

  stop

end program tetrapdos
