!
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
module types
  type wfn_for_k
     complex(kind=8),pointer:: wf(:,:,:)
     logical:: ifothnm
     logical:: ifrand
  end type wfn_for_k

  type qg_for_atm
     complex(kind=8),pointer:: qg(:,:)
  end type qg_for_atm

  type elec_energy
     real(kind=8):: eteig,eel,eht,exa,exc,evxc(0:6),ebnd,exx
     real(kind=8):: ekin,epot,epotl,epotnl
     real(kind=8):: total
  end type elec_energy

  type elec_senergy
     real(kind=8),dimension(6):: sekin,sepotl,sepotnl,sexx,total
     real(kind=8),dimension(6):: selptcr,seht
  end type elec_senergy

  type energy_per_atom
     real(kind=8),pointer:: ekin(:),epotl(:),epotnl(:),eht(:)
     real(kind=8),pointer:: exc(:),evxc(:,:),ewald(:)
     real(kind=8),pointer:: total(:)
  end type energy_per_atom

  type senergy_per_atom
     real(kind=8),pointer:: sekin(:,:),sepotl(:,:),sepotnl(:,:),selptcr(:,:)
     real(kind=8),pointer:: seht(:,:),sewald(:,:)
     real(kind=8),pointer:: total(:,:)
  end type senergy_per_atom
contains
  subroutine energy_per_atom_alloc(nsi,engatm)
    implicit none
    integer,intent(in):: nsi
    type (energy_per_atom):: engatm
    integer:: ierr

    allocate(engatm%ekin(nsi),engatm%epotl(nsi),engatm%epotnl(nsi), &
         engatm%eht(nsi),engatm%exc(nsi),engatm%evxc(0:6,nsi), &
         engatm%ewald(nsi),engatm%total(nsi),stat=ierr)
    call chkalloc('energy_per_atom','alloc',ierr)
    return
  end subroutine energy_per_atom_alloc

  subroutine energy_per_atom_dealloc(engatm)
    implicit none
    type (energy_per_atom):: engatm

    deallocate(engatm%ekin,engatm%epotl,engatm%epotnl, &
         engatm%eht,engatm%exc,engatm%evxc,engatm%ewald, &
         engatm%total)
    return
  end subroutine energy_per_atom_dealloc

  subroutine senergy_per_atom_alloc(nsi,sengatm)
    implicit none
    integer,intent(in):: nsi
    type (senergy_per_atom):: sengatm
    integer:: ierr

    allocate(sengatm%sekin(6,nsi),sengatm%sepotl(6,nsi), &
         sengatm%sepotnl(6,nsi),sengatm%seht(6,nsi), &
         sengatm%sewald(6,nsi),sengatm%selptcr(6,nsi), &
         sengatm%total(6,nsi),stat=ierr)
    call chkalloc('senergy_per_atom','alloc',ierr)
    return
  end subroutine senergy_per_atom_alloc

  subroutine senergy_per_atom_dealloc(sengatm)
    implicit none
    type (senergy_per_atom):: sengatm

    deallocate(sengatm%sekin,sengatm%sepotl,sengatm%sepotnl, &
         sengatm%seht,sengatm%sewald,sengatm%selptcr, &
         sengatm%total)
    return
  end subroutine senergy_per_atom_dealloc
end module types
