!
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
module utilmpi
contains
  subroutine copywfn(ncomp,nw,ssk,snkm,skg,sdwf,dsk,dnkm,dkg,ddwf,nproc, &
       sg_count,sg_displ,dg_count,dg_displ,ifinvsym,comm)
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: ncomp
    integer,intent(in):: nw,snkm,dnkm,skg(3,snkm),dkg(3,dnkm)
    real(kind=8),intent(in):: ssk(3),dsk(3)
    complex(kind=8),pointer:: sdwf(:,:,:)
    complex(kind=8),pointer:: ddwf(:,:,:)
    integer,intent(in):: nproc,comm
    integer,intent(in),dimension(0:nproc-1):: sg_count,sg_displ
    integer,intent(in),dimension(0:nproc-1):: dg_count,dg_displ
    logical,intent(in):: ifinvsym

    complex(kind=8),parameter:: unim = (0.0d0, 1.0d0)

    complex(kind=8),allocatable:: expwf(:,:,:,:),swf(:,:)
    integer:: nx,ny,nz,ig,ib,i1,i2,i3,bkm,ekm,myrank,ierr
    integer:: idg(3)
    real(kind=8):: norm,pnorm
    integer:: mpi_typ_wfn

    idg = nint(dsk-ssk)

    call MPI_COMM_RANK(comm,myrank,ierr)
    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    nx = 0
    ny = 0
    nz = 0
    do ig=1,snkm
       nx = max(nx,abs(skg(1,ig)))
       ny = max(ny,abs(skg(2,ig)))
       nz = max(nz,abs(skg(3,ig)))
    end do
    do ig=1,dnkm
       nx = max(nx,abs(dkg(1,ig)+idg(1)))
       ny = max(ny,abs(dkg(2,ig)+idg(2)))
       nz = max(nz,abs(dkg(3,ig)+idg(3)))
    end do

    allocate(expwf(ncomp,-nx:nx,-ny:ny,-nz:nz),swf(ncomp,snkm),stat=ierr)

    call chkalloc('copywfn','expwf,swf',ierr)

    do ib=1,nw
       expwf = 0.0d0

       bkm = sg_displ(myrank)+1
       call tpf_begin(tpf_mpi)
       call MPI_ALLGATHERV(sdwf(1,bkm,ib),sg_count(myrank), &
            mpi_typ_wfn,swf,sg_count,sg_displ, &
            mpi_typ_wfn,comm,ierr)
       call tpf_end(tpf_mpi)
!voption indep(expwf)
!poption indep(expwf)
       do ig=1,snkm
          i1=skg(1,ig)
          i2=skg(2,ig)
          i3=skg(3,ig)
          expwf(:,i1,i2,i3) = swf(:,ig)
       end do
       bkm = dg_displ(myrank)+1
       ekm = dg_displ(myrank)+dg_count(myrank)

       if (ncomp.eq.2 .or. maxval(abs(dsk)).gt.0.0d0 .or. ifinvsym) then
          do ig=bkm,ekm
             i1=dkg(1,ig)+idg(1)
             i2=dkg(2,ig)+idg(2)
             i3=dkg(3,ig)+idg(3)
             ddwf(:,ig,ib) = expwf(:,i1,i2,i3)
          end do
       else
          if (bkm.eq.1) then
             ddwf(1,1,ib) = dble(expwf(1,idg(1),idg(2),idg(3)))
             pnorm = ddwf(1,1,ib)**2
          else
             pnorm = 0.0d0
          end if
          !
          ! use Re \Psi(r)
          !
          do ig=max(bkm,2),ekm,2
             i1=dkg(1,ig)+idg(1)
             i2=dkg(2,ig)+idg(2)
             i3=dkg(3,ig)+idg(3)
             ddwf(1,ig,ib) = expwf(1,i1,i2,i3)
             i1=dkg(1,ig+1)+idg(1)
             i2=dkg(2,ig+1)+idg(2)
             i3=dkg(3,ig+1)+idg(3)
             ddwf(1,ig,ib) = 0.5d0*(ddwf(1,ig,ib)+conjg(expwf(1,i1,i2,i3)))
             ddwf(1,ig+1,ib) = conjg(ddwf(1,ig,ib))

             pnorm = pnorm + dble(ddwf(1,ig,ib))**2 + aimag(ddwf(1,ig,ib))**2 &
                  + dble(ddwf(1,ig+1,ib))**2 + aimag(ddwf(1,ig+1,ib))**2
          end do

          call tpf_begin(tpf_mpi)
          call MPI_ALLREDUCE(pnorm,norm,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
               comm,ierr)
          call tpf_end(tpf_mpi)

          if (norm.lt.0.5d0) then
             !
             ! use Im \Psi(r)
             !
             if (bkm.eq.1) then
                ddwf(1,1,ib) = aimag(expwf(1,idg(1),idg(2),idg(3)))
             end if
             do ig=max(bkm,2),ekm,2
                i1=dkg(1,ig)+idg(1)
                i2=dkg(2,ig)+idg(2)
                i3=dkg(3,ig)+idg(3)
                ddwf(1,ig,ib) = expwf(1,i1,i2,i3)
                i1=dkg(1,ig+1)+idg(1)
                i2=dkg(2,ig+1)+idg(2)
                i3=dkg(3,ig+1)+idg(3)
                ddwf(1,ig,ib) = 0.5d0*unim*( -ddwf(1,ig,ib)+conjg(expwf(1,i1,i2,i3)) )
                ddwf(1,ig+1,ib) = conjg(ddwf(1,ig,ib))
             end do
          end if
       end if
    end do
    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)
    deallocate(expwf,swf)
    return
  end subroutine copywfn

  subroutine rchkcons(t,n,name,comm)
    use mpi
    implicit none
    integer,intent(in):: n, comm
    real(kind=8),intent(in):: t(n)
    character(len=*),intent(in):: name
    real(kind=8),allocatable:: rt(:),rtmax(:)
    integer:: myrank,ierr
    real(kind=8):: tmax

    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(rt(n),rtmax(n))
    rt = t
    call MPI_BCAST(rt,n,MPI_DOUBLE_PRECISION,0,comm,ierr)
    rt = abs(t-rt)
    call MPI_REDUCE(rt,rtmax,n,MPI_DOUBLE_PRECISION,MPI_MAX,0,comm,ierr)
    if (myrank.eq.0) then
       tmax = maxval(rtmax)
       if (tmax.gt.0.0d0) then
          write(6,'(1x,''WARN: inconsistent '',a,e12.5)') name, tmax
       end if
    end if
    deallocate(rt,rtmax)
    return
  end subroutine rchkcons

  subroutine zchkcons(t,n,name,comm)
    use mpi
    implicit none
    integer,intent(in):: n, comm
    complex(kind=8),intent(in):: t(n)
    character(len=*),intent(in):: name
    complex(kind=8),allocatable:: rt(:)
    real(kind=8),allocatable:: rtmax(:),rttmp(:)
    integer:: myrank,ierr
    real(kind=8):: tmax

    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(rt(n),rtmax(n),rttmp(n))
    rt = t
    call MPI_BCAST(rt,n,MPI_DOUBLE_COMPLEX,0,comm,ierr)
    rttmp = abs(t-rt)
    call MPI_REDUCE(rttmp,rtmax,n,MPI_DOUBLE_PRECISION,MPI_MAX,0,comm,ierr)
    if (myrank.eq.0) then
       tmax = maxval(rtmax)
       if (tmax.gt.0.0d0) then
          write(6,'(1x,''WARN: inconsistent '',a,e12.5)') name, tmax
       end if
    end if
    deallocate(rt,rtmax,rttmp)
    return
  end subroutine zchkcons
end module utilmpi
