!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
program wfn2chg
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use chain_wfn
  use util
  use iounit
  use bsplpot
  use uspp_scrdd
  use vb_chrg_ovlp
  use types
  use meshgrid
  use chkinput
  use subr_rddata
  use chain_atom_cg
#ifdef ESM
  use esm_manage
#endif
  implicit none
  !
  character(len=7),parameter:: prgnam = 'wfn2chg'
#include "config90.h"
  integer:: xctp
  real(kind=8):: bb(3,3),va
  integer:: nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz

  integer:: nkr,nkfxa,nng
  integer:: laprj
  !
  integer:: nkc
  !
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3)
  real(kind=8):: uptime
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  !
  real(kind=8):: qwf
  !
  complex(kind=8):: pby(:,:,:,:,:)
  allocatable pby
  !
  real(kind=8),allocatable:: ee(:,:,:)
  real(kind=8),allocatable:: wgtk(:,:,:)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer:: i
  integer:: ik,is
  logical:: ifrgpbyee, ifinvsym
  real(kind=8):: clk
  integer:: mssg,ierr
  !
  type (wfn_for_k),allocatable:: dp(:,:)
  !
  integer:: myrank,nproc,comm
  integer:: bkm,ekm

  type (meshgrid_struct):: mgs

  integer:: ifbunp,ifrho,nbba1,nbba2,nsph
  integer:: isph

  integer,pointer:: ndiv(:),iaxis(:)
  real(kind=8),pointer:: rmin(:),rmax(:),rzero(:,:)
  integer:: itwav
  logical:: ifoutkmsh
  integer:: ndx(3)

  real(kind=8),pointer:: udirchr(:,:)
  !
  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  if (nproc.ne.1) then
     write(6,*) 'wfn2chg does not support nproc > 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*) prgnam
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdwfn2chgdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode &
       ,uptime,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero &
       ,rmin,rmax,ndiv,itwav,NFINPUT,mssg)
  !
  if (iopt(5).ne.1) then
     allocate(udirchr(3,nsi))
     udirchr = 0.0d0
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
     deallocate(udirchr)
  end if
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
  !
  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if
  !
  allocate(ee(nband,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'ee',ierr)
  !
  ! WARNING: ee must be cleared by 0 to notify
  !          ps2vcg() the SCF loop is initializing.
  !
  ee = 0.0d0
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc(PRGNAM,'bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)
  !
  allocate(dp(nks,nspin),stat=ierr)
  call chkalloc(prgnam,'dp',ierr)
  do ik=1,nks
     bkm=mgs%g_displ(myrank,ik)+1
     ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
     do is=1,nspin
        allocate(dp(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
        call chkalloc(prgnam,'dp%wf',ierr)
     end do
  end do

  laprj = getlaprj(nkd,bspots)
  allocate(pby(laprj,nsi,nw,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'pby',ierr)
  !
  if (iopt(2).ne.1) then
     write(6,'(1x,a)') 'initial_wfn must be 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  allocate(wgtk(nb1:nb2,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'wgtk',ierr)
  call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
       dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
       nband,ee,nkd,bspots,mssg,myrank,nproc,comm)

  if (ifrgpbyee) then
     write(6,'(1x,a,a)') prgnam, &
          ' requires an appropriate <beta|psi> data in wave func. file.'
     call MPI_FINALIZE(ierr)
     stop
  end if
  !
  ! In some architecture, two different I/O unit can not
  ! share the same file. Therefore NFWFI must be closed at this
  ! point for NFWFO.
  !
  close(NFWFI)
  !
  ! For charged state
  if (mssg.gt.-1) write(6,10) znext
10 format('CHARGE STATE CALCULATION:' &
       ,f8.3,' EXTRA ELECTRON IS ADDED')
  !
  if (ifbunp.gt.0) then
     call xrewind(nfout)
     write(nfout,'(6(1x,i4),1x,i6,1x,i1)') &
          nks,nw,ifbunp,nspin,nbba1,nbba2,nkmesh,ncomp
  end if
  if (ifrho.eq.1) then
     write(6,*) ' output rho(k;i)'
     call xrewind(nfrok)
     write(nfrok) 'TAPP:rok'
     write(nfrok) aa,nspin
     write(nfrok) nrx*2,nry*2,nrz*2
     write(nfrok) nks,nbba1,nbba2,nkmesh
  end if
  if ((ifbunp.eq.1).or.(ifbunp.eq.2)) then
     write(nfout,'(1x,i4)') nsph
     do isph=1,nsph
        if (ifbunp .eq. 1) then
           ! One-dimensional distribution
           write(nfout,5120) iaxis(isph), &
                rmin(isph),rmax(isph),ndiv(isph)
5120       format(i5,2e15.7,i5)
        else if (ifbunp .eq. 2) then
           ! Radial distribution around the point rzero(3)
           write(nfout,5130) (rzero(i,isph),i=1,3), &
                rmin(isph),rmax(isph),ndiv(isph)
5130       format(5e14.6,i5)
        end if
     end do
  end if
  !
  ifoutkmsh = .false.
  ndx = 0
  write(6,*) mmm
  if ( (ifbunp.ge.1 .and. ifbunp.le.4) .and. nsk.eq.1 .and. & 
       mod(mmm(1,1,1),mmm(1,2,1)).eq.0 .and. &
       mod(mmm(2,1,1),mmm(2,2,1)).eq.0 .and. &
       mod(mmm(3,1,1),mmm(3,2,1)).eq.0 ) then
     if (mssg.gt.-1) then
        write(6,*) 'single k point mesh which contains the Gamma point.'
        write(6,*) 'enable output on the mesh for the tetrahedron method.'
     end if
     ifoutkmsh = .true.
     ndx(:) = nk/mmm(:,2,1)
  end if
  !
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'WFN2CHG START',clk

  call vbchrg(dp,pby,nband,laprj,nsi,nkd &
       ,nspin,ncomp &
       ,nw,nbba1,nbba2 &
       ,va,aa,bb,asi,kd,ee,mgs,bspotfc,bspots,ifinvsym &
       ,ifbunp,ifrho,nfrok,nfwfk,nfout,nsph,iaxis,rmin,rmax,ndiv,rzero &
       ,ng,rg,nnp,pg,srg,itwav,ifoutkmsh,ndx,nfoutkmsh &
       ,nb1,nb2,wgtk &
       ,mssg,myrank,nproc,comm)
  !
  call dealloc_mesh(mgs)
  deallocate(ee)
  deallocate(pby)
  deallocate(wgtk)
  do is=1,nspin
     do ik=1,nks
        deallocate(dp(ik,is)%wf)
     end do
  end do
  deallocate(dp)
  deallocate(kd,asi,mmm)
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'WFN2CHG is over. Time=', clk
  call MPI_FINALIZE(ierr)
  stop
end program wfn2chg
