/*
  Copyright (c) 2012-2013 Yoshihide Yoshimoto
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>

#include "exchange_kern_opencl.h"

#undef DEBUG_MODE

#ifdef DEBUG_MODE
#define CHECKSTAT(place,stat) { printf(place,stat); }
#else
#define CHECKSTAT(place,stat)
#endif

int main(int argc, char *argv[]) {

	int nw = 20;
    int nwx = 48, nwy = 48, nwz = 48;
    int nwx2 = nwx*2, nwy2 = nwy*2, nwz2 = nwz*2;
	int nlx, nly, nlz, nlxyz;
	const double bb[3][3] = {{1.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0.0, 0.0, 1.0}};
	const double q[3] = {0.0, 0.0, 0.0};

	const int nkmesh = 1;
	double_complex *wfi, *wfj, *vxp;
	double_complex *wf1;
#ifdef DEBUG_MODE
    double_complex *vxp1;
#endif
	double *wgt;

	int ix,iy,iz,ig,ib;

	struct timeval t1,t2,t21;
	struct timezone tz;

	printf("nw[xyz] = %d %d %d\n", nwx, nwy, nwz);
	printf("nw = %d\n", nw);

	init_exchange_(&nw,&nw,&nwx,&nwy,&nwz);

	get_exchange_nlxyz_(&nlx,&nly,&nlz);
	nlxyz = nlx*nly*nlz;

	wfi = (double_complex *)malloc(nw*nlxyz*sizeof(double_complex));
	wfj = (double_complex *)malloc(nw*nlxyz*sizeof(double_complex));
	vxp = (double_complex *)malloc(nw*nlxyz*sizeof(double_complex));
	wgt = (double *)malloc(nw*sizeof(double));

	CHECKSTAT("allocate %d\n", 1);

	memset(wfj, 0, sizeof(double_complex)*nlxyz*nw);
	wf1 = wfj; 
	for(ib=0; ib<nw; ib++) {
		for(iz=0;iz<nwz2; iz++) {
			for(iy=0;iy<nwy2; iy++) {
				for(ix=0;ix<nwx2; ix++) {
					ig = ix + nlx*iy + nlx*nly*iz;
					wf1[ig].re = 1.0;
					wf1[ig].im = 0.0;
				}
			}
		}
		wf1 = wf1 + nlxyz;
	}
	CHECKSTAT("allocate %d\n", 2);

	memset(wfi, 0, sizeof(double_complex)*nlxyz*nw);
	wf1 = wfi;
	for(ib=0; ib<nw; ib++) {
		for(iz=0; iz<nwz2; iz++) {
			for(iy=0; iy<nwy2; iy++) {
				for(ix=0; ix<nwx2; ix++) {
					ig = ix + nlx*iy + nlx*nly*iz;
					wf1[ig].re = 1.0;
					wf1[ig].im = 0.0;
				}
			}
		}
		wf1 = wf1 + nlxyz;
	}

	memset(vxp, 0, sizeof(double_complex)*nlxyz*nw);

	for(ib=0; ib<nw; ib++) {
		wgt[ib] = 1.5;
	}

	CHECKSTAT("initwf %d\n",0);

	setup_exchange_clkern_(&nkmesh,bb,q);

	gettimeofday(&t1,&tz);

	compute_exchange_(&nw, &nw, wfi, wfj, vxp, wgt);

	gettimeofday(&t2,&tz);
	timersub(&t2,&t1,&t21);

	printf("time = %e\n", t21.tv_sec + 1e-6*t21.tv_usec);

#ifdef DEBUG_MODE
	vxp1 = vxp;
	for(ib=0; ib<nw; ib++) {
		for(iz=0; iz<nwz2; iz++) {
			for(iy=0; iy<nwy2; iy++) {
				for(ix=0; ix<nwx2; ix++) {
					ig = ix + nlx*iy + nlx*nly*iz;
					printf("%e %e\n",vxp1[ig].re, vxp1[ig].im);
				}
			}
		}
		vxp1 = vxp1 + nlxyz;
	}
#endif

	teardown_exchange_();

	fflush(stdout);

	exit(0);
}
