!
! Copyright (c) 2012-2013 Yoshihide Yoshimoto
!

#include "fftgrid.h"

program potexchdemo
  use fft_3d
  use util
  implicit none

  !integer,parameter:: nw = 4
  !integer,parameter:: nwx = 9, nwy = 9, nwz = 9
  integer,parameter:: nw = 20
  integer,parameter:: nwx = 48, nwy = 48, nwz = 48
  integer,parameter:: nwx2 = nwx*2, nwy2 = nwy*2, nwz2 = nwz*2
  real(kind=8),parameter:: bb(3,3) = reshape( (/1.0d0, 0.0d0, 0.0d0, 0.0d0, 1.0d0, 0.0d0, 0.0d0, 0.0d0, 1.0d0 /), (/ 3, 3 /) )
  integer,parameter:: nkmesh = 1
  real(kind=8):: wff(2,nwx2,nwy2,nwz2,nw)
  real(kind=8):: wf(2,nwx2,nwy2,nwz2,nw)
  real(kind=8):: vxp(2,nwx2,nwy2,nwz2,nw)

  real(kind=8),parameter:: pi = 3.14159265358979323846d0
  real(kind=8),allocatable:: ztmp(:,:,:,:),ztmp2(:,:)
  real(kind=8),allocatable:: pfxi(:,:,:,:),pfxj(:,:,:,:,:),pxj(:,:,:,:,:)
  real(kind=8),allocatable:: vsum(:,:,:,:)
  real(kind=8),allocatable:: clkern(:,:,:)

  integer:: nlx,nly,nlz,nlxyz
  integer:: ix,iy,iz,i1,i2,i3
  integer:: ib,jb
  real(kind=8):: fac, g(3)
  real(kind=8):: vb, q0
  integer:: count,count_rate
  real(kind=8):: t1,t2

  type (fft3_struct):: fs

  write(6,*) 'begin'

  write(6,*) 'nw[xyz] = ', nwx, nwy, nwz
  write(6,*) 'nw = ', nw

  nlx = FFT_PAD_NX(nwx2)
  nly = FFT_PAD_NY(nwy2)
  nlz = FFT_PAD_NZ(nwz2)
  nlxyz = nlx*nly*nlz

  call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)

  allocate(pfxi(2,nlx,nly,nlz),pfxj(2,nlx,nly,nlz,nw), &
       pxj(2,nlx,nly,nlz,nw), &
       ztmp(2,nlx,nly,nlz),ztmp2(2,nlxyz), &
       vsum(2,nlx,nly,nlz), &
       clkern(nwx2,nwy2,nwz2))

  write(6,*) 'allocate'

  wf = 0.0d0
  do ib=1,nw
     do iz = 1,nwz2
        do iy = 1,nwy2
           do ix = 1,nwx2
              wf(1,ix,iy,iz,ib) = 1.0d0
              wf(2,ix,iy,iz,ib) = 0.0d0
           end do
        end do
     end do
  end do
  wff = 0.0d0
  do ib=1,nw
     do iz = 1,nwz2
        do iy = 1,nwy2
           do ix = 1,nwx2
              wff(1,ix,iy,iz,ib) = 1.0d0
              wff(2,ix,iy,iz,ib) = 0.0d0
           end do
        end do
     end do
  end do

  vxp = 0.0d0 

  write(6,*) 'initwf'

  clkern = 0.0d0
  do iz = -nwz+1,nwz-1
     do iy = -nwy+1,nwy-1
        do ix = -nwx+1,nwx-1
           i1=mod(nwx2+ix,nwx2)+1
           i2=mod(nwy2+iy,nwy2)+1
           i3=mod(nwz2+iz,nwz2)+1

           if (ix .ne. 0 .or. iy .ne. 0 .or. iz .ne. 0) then
              g(1) = bb(1,1)*ix+bb(1,2)*iy+bb(1,3)*iz
              g(2) = bb(2,1)*ix+bb(2,2)*iy+bb(2,3)*iz
              g(3) = bb(3,1)*ix+bb(3,2)*iy+bb(3,3)*iz
              clkern(i1,i2,i3) = 4.0d0*pi/( g(1)**2 + g(2)**2 + g(3)**2 )
           else
              vb = cellvol(bb)
              ! vb/nkmesh = 4.0d0*pi/3.0d0*q_0**3
              q0 = ( 3.0d0*vb/(4.0d0*pi*nkmesh) )**(1.0d0/3.0d0)
              !
              ! \int_0^{q_0} 4\pi/q^2 4\pi q^2 dq = \int_0^{q_0} 16\pi^2 dq
              ! = 16\pi^2 q_0
              !
              clkern(i1,i2,i3) = 16.0d0*(pi**2)*q0
           end if
        end do
     end do
  end do

  write(6,*) 'initclkern'

  call system_clock(count,count_rate)
  t1 = dble(count)/dble(count_rate)

  pfxj = 0.0d0
  do jb=1,nw
     pfxj(1,1:nwx2,1:nwy2,1:nwz2,jb) = wff(1,:,:,:,jb)
     pfxj(2,1:nwx2,1:nwy2,1:nwz2,jb) = wff(2,:,:,:,jb)

     call fft3_bw(fs,pfxj(:,:,:,:,jb),ztmp2)
  end do

  pxj = 0.0d0
  do jb=1,nw
     pxj(1,1:nwx2,1:nwy2,1:nwz2,jb) = wf(1,:,:,:,jb)
     pxj(2,1:nwx2,1:nwy2,1:nwz2,jb) = wf(2,:,:,:,jb)

     call fft3_bw(fs,pxj(:,:,:,:,jb),ztmp2)
  end do

  ztmp = 0.0d0
  ztmp2 = 0.0d0
  pfxi = 0.0d0

  do ib=1,nw

     pfxi(1,1:nwx2,1:nwy2,1:nwz2) = wff(1,:,:,:,ib)
     pfxi(2,1:nwx2,1:nwy2,1:nwz2) = wff(2,:,:,:,ib)

     call fft3_bw(fs,pfxi(:,:,:,:),ztmp2)

     vsum = 0.0d0
     do jb=1,nw
        do iz = 1,nwz2
           do iy = 1,nwy2
              do ix = 1,nwx2
                 ztmp(1,ix,iy,iz) = &
                      pfxj(1,ix,iy,iz,jb)*pfxi(1,ix,iy,iz) &
                      + pfxj(2,ix,iy,iz,jb)*pfxi(2,ix,iy,iz)
                 ztmp(2,ix,iy,iz) = &
                      pfxj(1,ix,iy,iz,jb)*pfxi(2,ix,iy,iz) &
                      - pfxj(2,ix,iy,iz,jb)*pfxi(1,ix,iy,iz)
              end do
           end do
        end do
        call fft3_fw(fs,ztmp,ztmp2)

        do iz = 1,nwz2
           do iy = 1,nwy2
              do ix = 1,nwx2
                 ztmp(1,ix,iy,iz) = clkern(ix,iy,iz)*ztmp(1,ix,iy,iz)
                 ztmp(2,ix,iy,iz) = clkern(ix,iy,iz)*ztmp(2,ix,iy,iz)
              end do
           end do
        end do
        call fft3_bw(fs,ztmp,ztmp2)

        do iz = 1,nwz2
           do iy = 1,nwy2
              do ix = 1,nwx2
                 vsum(1,ix,iy,iz) = vsum(1,ix,iy,iz) &
                      + ztmp(1,ix,iy,iz)*pxj(1,ix,iy,iz,jb) &
                      - ztmp(2,ix,iy,iz)*pxj(2,ix,iy,iz,jb)
                 vsum(2,ix,iy,iz) = vsum(2,ix,iy,iz) &
                      + ztmp(1,ix,iy,iz)*pxj(2,ix,iy,iz,jb) &
                      + ztmp(2,ix,iy,iz)*pxj(1,ix,iy,iz,jb)
              end do
           end do
        end do
     end do
     call fft3_fw(fs,vsum,ztmp2)

     fac = 1.5d0 ! demo value

     vxp(1,:,:,:,ib) = vxp(1,:,:,:,ib) - fac*vsum(1,1:nwx2,1:nwy2,1:nwz2)
     vxp(2,:,:,:,ib) = vxp(2,:,:,:,ib) - fac*vsum(2,1:nwx2,1:nwy2,1:nwz2)

  end do

  call system_clock(count,count_rate)
  t2 = dble(count)/dble(count_rate)

  write(6,*) 'time = ', t2-t1

  do ib=1,nw
     do iz=1,nwz2
        do iy=1,nwy2
           do ix=1,nwx2
              write(6,*) vxp(1,ix,iy,iz,ib),vxp(2,ix,iy,iz,ib)
           end do
        end do
     end do
  end do

  call fft3_done(fs)

end program potexchdemo
