!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
program testsarch
  implicit none
  integer:: m,m1,m2,b,e,ierr
  logical:: check, initran

  read(5,*) check, initran
  write(6,*) 'sub dt[s] p[GFlops]'
  do while(.true.)
     read(5,*,IOSTAT=ierr) m, m1, m2, b, e
     if (ierr.ne.0) exit

     write(6,'(4(1x,a,i5),1x,a,i7)') 'm = ', m, 'm1 = ', m1, 'm2 = ', m2, &
          'b = ', b, 'e = ', e

     write(6,*) 'testing arch_rotwfn'
     call testrotwfn(m,b,e,check,initran)
     write(6,*) 'testing arch_nlprj'
     call testnlprj(m1,m2,b,e,check,initran)
     write(6,*) 'testing arch_sorthcl'
     call testsorthcl(m1,m2,b,e,check,initran)

     write(6,*) 'testing arch_rotwfni'
     call testrotwfni(m,b,e,check,initran)
     write(6,*) 'testing arch_nlprji'
     call testnlprji(m1,m2,b,e,check,initran)
     write(6,*) 'testing arch_sorthcli'
     call testsorthcli(m1,m2,b,e,check,initran)

     write(6,*) 'testing arch_rotwfng, b=2'
     call testrotwfng(m,2,e,check,initran)
     write(6,*) 'testing arch_rotwfng, b=1'
     call testrotwfng(m,1,e,check,initran)
     write(6,*) 'testing arch_nlprjg, b=2'
     call testnlprjg(m1,m2,2,e,check,initran)
     write(6,*) 'testing arch_nlprjg, b=1'
     call testnlprjg(m1,m2,1,e,check,initran)
     write(6,*) 'testing arch_sorthclg, b=2'
     call testsorthclg(m1,m2,2,e,check,initran)
     write(6,*) 'testing arch_sorthclg, b=1'
     call testsorthclg(m1,m2,1,e,check,initran)

  end do
contains
  subroutine testrotwfn(m,b,e,check,initran)
    use arch_rotwfn
    implicit none
    integer:: m, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    complex(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m),y(b:e,2,m),yr(b:e,2,m), &
         a(m,m),g(m,m),gr(m,m))

    iseed = 1234567

    if (initran) then
       do k=1,m
          do j=1,m
             a(j,k) = dcmplx(ran0(iseed),ran0(iseed))
          end do
       end do

       do j=1,m
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
    else
       a = (1.0d0, 1.0d0)
       x = 1.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotvecs(b,e,m,a,x,y)
    call timer(et,1)
    write(6,*) 'rotwfn:rotvecs ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m
          do j=1,m
             do i=b,e
                yr(i,1,k) = yr(i,1,k) &
                     + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
                yr(i,2,k) = yr(i,2,k) &
                     + dble(a(j,k))*x(i,2,j) + aimag(a(j,k))*x(i,1,j)
             end do
          end do
       end do

       do j=1,m
          do i=b,e
             if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
                  (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
                write(6,*) i,j,y(i,1,j),y(i,2,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfn:ovlmat ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m
          do j=1,m
             do i=b,e
                gr(k,j) = dcmplx( &
                     dble(gr(k,j)) &
                     + x(i,1,k)*y(i,1,j) + x(i,2,k)*y(i,2,j), &
                     aimag(gr(k,j)) &
                     + x(i,1,k)*y(i,2,j) - x(i,2,k)*y(i,1,j) )
             end do
          end do
       end do

       do k=1,m
          do j=1,m
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ltrovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfn:ltrovlmat ',et-bt, &
         8.0d0*dble(e-b+1)*0.5d0*dble(m+1)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m
          do j=k,m
             do i=b,e
                gr(j,k) = dcmplx( &
                     dble(gr(j,k)) &
                     + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k), &
                     aimag(gr(j,k)) &
                     + x(i,1,j)*y(i,2,k) - x(i,2,j)*y(i,1,k) )
             end do
          end do
       end do

       do k=1,m
          do j=k,m
             if (abs(g(j,k)-gr(j,k)) .gt. eps ) then
                write(6,*) j,k,g(j,k)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testrotwfn

  subroutine testnlprj(m1,m2,b,e,check,initran)
    use arch_nlprj
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    complex(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m1),y(b:e,2,m2),yr(b:e,2,m2), &
         a(m1,m2),g(m1,m2),gr(m1,m2))

    if (initran) then 
       iseed = 1234567
       do k=1,m2
          do j=1,m1
             a(j,k) = dcmplx(ran0(iseed),ran0(iseed))
          end do
       end do

       do j=1,m1
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
    else
       a = (1.0d0, -1.0d0)
       x = 1.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotprjs(b,e,m1,m2,a,x,y)
    call timer(et,1)
    write(6,*) 'nlprj:rotprjs ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m2
          do j=1,m1
             do i=b,e
                yr(i,1,k) = yr(i,1,k) &
                     + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
                yr(i,2,k) = yr(i,2,k) &
                     + dble(a(j,k))*x(i,2,j) + aimag(a(j,k))*x(i,1,j)
             end do
          end do
       end do

       do j=1,m2
          do i=b,e
             if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
                  (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
                write(6,*) i,j,y(i,1,j),y(i,2,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'nlprj:ovlmat ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m1
          do j=1,m2
             do i=b,e
                gr(k,j) = dcmplx( &
                     dble(gr(k,j)) &
                     + x(i,1,k)*yr(i,1,j) + x(i,2,k)*yr(i,2,j), &
                     aimag(gr(k,j)) &
                     + x(i,1,k)*yr(i,2,j) - x(i,2,k)*yr(i,1,j) )
             end do
          end do
       end do

       do k=1,m1
          do j=1,m2
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testnlprj

  subroutine testsorthcl(m1,m2,b,e,check,initran)
    use arch_sorthcl
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    complex(kind=8),allocatable:: g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: tr,ti
    real(kind=8):: bt,et

    allocate(x(b:e,2,m1),y(b:e,2,m2),yr(b:e,2,m2), &
         g(m1,m2),gr(m1,m2))

    if (initran) then
       iseed = 1234567
       do j=1,m1
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
       do j=1,m2
          call ran0v(iseed,b,e,y(:,1,j))
          call ran0v(iseed,b,e,y(:,2,j))
       end do
    else
       x = 1.0d0
       y = -1.0d0
    end if

    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'sorthcl:ovlmat ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
    gr = (0.0d0, 0.0d0)
    do k=1,m1
       do j=1,m2
          do i=b,e
             tr = dble(gr(k,j)) + x(i,1,k)*y(i,1,j) + x(i,2,k)*y(i,2,j)
             ti = aimag(gr(k,j)) + x(i,1,k)*y(i,2,j) - x(i,2,k)*y(i,1,j)
             gr(k,j) = dcmplx( tr, ti )
          end do
       end do
    end do

    do k=1,m1
       do j=1,m2
          if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
             write(6,*) j,k,g(k,j)
          end if
       end do
    end do
    end if

    yr = y

    call timer(bt,0)
    call subovlpt(b,e,m1,m2,g,x,y)
    call timer(et,1)
    write(6,*) 'sorthcl:subovlpt ',et-bt, &
         8.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
    do k=1,m2
       do j=1,m1
          do i=b,e
             yr(i,1,k) = yr(i,1,k) &
                  - dble(gr(j,k))*x(i,1,j) + aimag(gr(j,k))*x(i,2,j)
             yr(i,2,k) = yr(i,2,k) &
                  - dble(gr(j,k))*x(i,2,j) - aimag(gr(j,k))*x(i,1,j)
          end do
       end do
    end do

    do j=1,m2
       do i=b,e
          if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
               (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
             write(6,*) i,j,y(i,1,j),y(i,2,j)
          end if
       end do
    end do
    end if

    deallocate(x,y,yr,g,gr)
    return
  end subroutine testsorthcl

  subroutine testrotwfni(m,b,e,check,initran)
    use arch_rotwfni
    implicit none
    integer:: m, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:),y(:,:),yr(:,:)
    real(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,m),y(b:e,m),yr(b:e,m), &
         a(m,m),g(m,m),gr(m,m))

    iseed = 1234567

    if (initran) then
       do k=1,m
          do j=1,m
             a(j,k) = ran0(iseed)
          end do
       end do

       do j=1,m
          call ran0v(iseed,b,e,x(:,j))
       end do
    else
       a = 1.0d0
       x = 1.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotvecs(b,e,m,a,x,y)
    call timer(et,1)
    write(6,*) 'rotwfni:rotvecs ',et-bt, &
         2.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m
          do j=1,m
             do i=b,e
                yr(i,k) = yr(i,k) + a(j,k)*x(i,j)
             end do
          end do
       end do

       do j=1,m
          do i=b,e
             if ( abs(y(i,j)-yr(i,j)).gt.eps ) then
                write(6,*) i,j,y(i,j)
             end if
          end do
       end do
    end if

    g = 0.0d0
    call timer(bt,0)
    call ovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfni:ovlmat ',et-bt, &
         2.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = 0.0d0
       do k=1,m
          do j=1,m
             do i=b,e
                gr(k,j) = gr(k,j) + x(i,k)*y(i,j)
             end do
          end do
       end do

       do k=1,m
          do j=1,m
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    g = 0.0d0
    call timer(bt,0)
    call ltrovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfni:ltrovlmat ',et-bt, &
         2.0d0*dble(e-b+1)*0.5d0*dble(m+1)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = 0.0d0
       do k=1,m
          do j=k,m
             do i=b,e
                gr(j,k) = gr(j,k) + x(i,j)*y(i,k)
             end do
          end do
       end do

       do k=1,m
          do j=k,m
             if (abs(g(j,k)-gr(j,k)) .gt. eps ) then
                write(6,*) j,k,g(j,k)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testrotwfni

  subroutine testnlprji(m1,m2,b,e,check,initran)
    use arch_nlprji
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:),yr(:,:)
    complex(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m1),y(b:e,m2),yr(b:e,m2), &
         a(m1,m2),g(m1,m2),gr(m1,m2))

    if (initran) then 
       iseed = 1234567
       do k=1,m2
          do j=1,m1
             a(j,k) = dcmplx(ran0(iseed),ran0(iseed))
          end do
       end do

       do j=1,m1
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
    else
       a = (1.0d0, -1.0d0)
       x = 1.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotprjs(b,e,m1,m2,a,x,y)
    call timer(et,1)
    write(6,*) 'nlprji:rotprjs ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m2
          do j=1,m1
             do i=b,e
                yr(i,k) = yr(i,k) &
                     + dble(a(j,k))*x(i,1,j) - aimag(a(j,k))*x(i,2,j)
             end do
          end do
       end do

       do j=1,m2
          do i=b,e
             if ( (abs(y(i,j)-yr(i,j)).gt.eps) .or. &
                  (abs(y(i,j)-yr(i,j)).gt.eps) ) then
                write(6,*) i,j,y(i,j),y(i,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'nlprji:ovlmat ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m1
          do j=1,m2
             do i=b,e
                gr(k,j) = dcmplx( &
                     dble(gr(k,j)) + x(i,1,k)*yr(i,j), &
                     aimag(gr(k,j)) - x(i,2,k)*yr(i,j) )
             end do
          end do
       end do

       do k=1,m1
          do j=1,m2
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testnlprji

  subroutine testsorthcli(m1,m2,b,e,check,initran)
    use arch_sorthcli
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:),y(:,:),yr(:,:)
    real(kind=8),allocatable:: g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,m1),y(b:e,m2),yr(b:e,m2), &
         g(m1,m2),gr(m1,m2))

    if (initran) then
       iseed = 1234567
       do j=1,m1
          call ran0v(iseed,b,e,x(:,j))
       end do
       do j=1,m2
          call ran0v(iseed,b,e,y(:,j))
       end do
    else
       x = 1.0d0
       y = -1.0d0
    end if

    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'sorthcli:ovlmat ',et-bt, &
         2.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
    gr = 0.0d0
    do k=1,m1
       do j=1,m2
          do i=b,e
             gr(k,j) = gr(k,j) + x(i,k)*y(i,j)
          end do
       end do
    end do

    do k=1,m1
       do j=1,m2
          if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
             write(6,*) j,k,g(k,j)
          end if
       end do
    end do
    end if

    yr = y

    call timer(bt,0)
    call subovlpt(b,e,m1,m2,g,x,y)
    call timer(et,1)
    write(6,*) 'sorthcli:subovlpt ',et-bt, &
         2.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
    do k=1,m2
       do j=1,m1
          do i=b,e
             yr(i,k) = yr(i,k) - gr(j,k)*x(i,j)
          end do
       end do
    end do

    do j=1,m2
       do i=b,e
          if ( abs(y(i,j)-yr(i,j)).gt.eps ) then
             write(6,*) i,j,y(i,j)
          end if
       end do
    end do
    end if

    deallocate(x,y,yr,g,gr)
    return
  end subroutine testsorthcli

  subroutine testrotwfng(m,b,e,check,initran)
    use arch_rotwfng
    implicit none
    integer:: m, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    real(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m),y(b:e,2,m),yr(b:e,2,m), &
         a(m,m),g(m,m),gr(m,m))

    iseed = 1234567

    if (initran) then
       do k=1,m
          do j=1,m
             a(j,k) = ran0(iseed)
          end do
       end do

       do j=1,m
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
    else
       a = 1.0d0
       x = 1.0d0
    end if
    if (b .eq. 1) then
       x(1,2,:) = 0.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotvecs(b,e,m,a,x,y)
    call timer(et,1)
    write(6,*) 'rotwfng:rotvecs ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m
          do j=1,m
             do i=b,e
                yr(i,1,k) = yr(i,1,k) + a(j,k)*x(i,1,j)
                yr(i,2,k) = yr(i,2,k) + a(j,k)*x(i,2,j)
             end do
          end do
       end do

       do j=1,m
          do i=b,e
             if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
                  (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
                write(6,*) i,j,y(i,1,j),y(i,2,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfng:ovlmat ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m
          do j=1,m
             do i=max(b,2),e
                gr(k,j) = gr(k,j) + x(i,1,k)*y(i,1,j) + x(i,2,k)*y(i,2,j)
             end do
          end do
       end do
       gr = 2.0d0*gr
       if (b.eq.1) then
          do k=1,m
             do j=1,m
                gr(k,j) = gr(k,j) + x(1,1,k)*y(1,1,j)
             end do
          end do
       end if

       do k=1,m
          do j=1,m
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ltrovlmat(b,e,m,x,y,g)
    call timer(et,1)
    write(6,*) 'rotwfng:ltrovlmat ',et-bt, &
         4.0d0*dble(e-b+1)*0.5d0*dble(m+1)*dble(m)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m
          do j=k,m
             do i=max(b,2),e
                gr(j,k) = gr(j,k) + x(i,1,j)*y(i,1,k) + x(i,2,j)*y(i,2,k)
             end do
          end do
       end do
       gr = 2.0d0*gr
       if (b.eq.1) then
          do k=1,m
             do j=k,m
                gr(j,k) = gr(j,k) + x(1,1,j)*y(1,1,k)
             end do
          end do
       end if

       do k=1,m
          do j=k,m
             if (abs(g(j,k)-gr(j,k)) .gt. eps ) then
                write(6,*) j,k,g(j,k)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testrotwfng

  subroutine testnlprjg(m1,m2,b,e,check,initran)
    use arch_nlprjg
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    real(kind=8),allocatable:: a(:,:),g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m1),y(b:e,2,m2),yr(b:e,2,m2), &
         a(m1,m2),g(m1,m2),gr(m1,m2))

    if (initran) then 
       iseed = 1234567
       do k=1,m2
          do j=1,m1
             a(j,k) = dcmplx(ran0(iseed),ran0(iseed))
          end do
       end do

       do j=1,m1
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
    else
       a = (1.0d0, -1.0d0)
       x = 1.0d0
    end if
    if (b .eq. 1) then
       x(1,2,:) = 0.0d0
    end if

    y = 0.0d0
    call timer(bt,0)
    call rotprjs(b,e,m1,m2,a,x,y)
    call timer(et,1)
    write(6,*) 'nlprjg:rotprjs ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       yr = 0.0d0
       do k=1,m2
          do j=1,m1
             do i=b,e
                yr(i,1,k) = yr(i,1,k) + a(j,k)*x(i,1,j)
                yr(i,2,k) = yr(i,2,k) + a(j,k)*x(i,2,j)
             end do
          end do
       end do

       do j=1,m2
          do i=b,e
             if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
                  (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
                write(6,*) i,j,y(i,1,j),y(i,2,j)
             end if
          end do
       end do
    end if

    g = (0.0d0, 0.0d0)
    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'nlprjg:ovlmat ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m1
          do j=1,m2
             do i=max(b,2),e
                gr(k,j) = gr(k,j) + x(i,1,k)*yr(i,1,j) + x(i,2,k)*yr(i,2,j)
             end do
          end do
       end do
       gr = gr*2.0d0
       if (b.eq.1) then
          do k=1,m1
             do j=1,m2
                gr(k,j) = gr(k,j) + x(1,1,k)*yr(1,1,j)
             end do
          end do
       end if

       do k=1,m1
          do j=1,m2
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,a,g,gr)
    return
  end subroutine testnlprjg

  subroutine testsorthclg(m1,m2,b,e,check,initran)
    use arch_sorthclg
    implicit none
    integer:: m1, m2, b, e
    logical:: check, initran

    real(kind=8),parameter:: eps = 1.0d-8

    real(kind=8),allocatable:: x(:,:,:),y(:,:,:),yr(:,:,:)
    real(kind=8),allocatable:: g(:,:),gr(:,:)

    integer:: iseed,i,j,k
    real(kind=8):: bt,et

    allocate(x(b:e,2,m1),y(b:e,2,m2),yr(b:e,2,m2), &
         g(m1,m2),gr(m1,m2))

    if (initran) then
       iseed = 1234567
       do j=1,m1
          call ran0v(iseed,b,e,x(:,1,j))
          call ran0v(iseed,b,e,x(:,2,j))
       end do
       do j=1,m2
          call ran0v(iseed,b,e,y(:,1,j))
          call ran0v(iseed,b,e,y(:,2,j))
       end do
    else
       x = 1.0d0
       y = -1.0d0
    end if
    if (b .eq. 1) then
       x(1,2,:) = 0.0d0
       y(1,2,:) = 0.0d0
    end if

    call timer(bt,0)
    call ovlmat(b,e,m1,m2,x,y,g)
    call timer(et,1)
    write(6,*) 'sorthclg:ovlmat ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       gr = (0.0d0, 0.0d0)
       do k=1,m1
          do j=1,m2
             do i=max(b,2),e
                gr(k,j) = gr(k,j) + x(i,1,k)*y(i,1,j) + x(i,2,k)*y(i,2,j)
             end do
          end do
       end do
       gr = 2.0d0*gr
       if (b.eq.1) then
          do k=1,m1
             do j=1,m2
                gr(k,j) = gr(k,j) + x(1,1,k)*y(1,1,j)
             end do
          end do
       end if

       do k=1,m1
          do j=1,m2
             if ( abs(g(k,j)-gr(k,j)) .gt. eps ) then
                write(6,*) j,k,g(k,j)
             end if
          end do
       end do
    end if

    yr = y

    call timer(bt,0)
    call subovlpt(b,e,m1,m2,g,x,y)
    call timer(et,1)
    write(6,*) 'sorthclg:subovlpt ',et-bt, &
         4.0d0*dble(e-b+1)*dble(m1)*dble(m2)/(et-bt)*1.0d-9

    if (check) then
       do k=1,m2
          do j=1,m1
             do i=b,e
                yr(i,1,k) = yr(i,1,k) - gr(j,k)*x(i,1,j)
                yr(i,2,k) = yr(i,2,k) - gr(j,k)*x(i,2,j)
             end do
          end do
       end do

       do j=1,m2
          do i=b,e
             if ( (abs(y(i,1,j)-yr(i,1,j)).gt.eps) .or. &
                  (abs(y(i,2,j)-yr(i,2,j)).gt.eps) ) then
                write(6,*) i,j,y(i,1,j),y(i,2,j)
             end if
          end do
       end do
    end if

    deallocate(x,y,yr,g,gr)
    return
  end subroutine testsorthclg

  real(kind=8) function ran0(idum)
    implicit none
    integer idum,IA,IM,IQ,IR
    real(kind=8) AM
    parameter (IA=16807,IM=2147483647,AM=1./IM,IQ=127773,IR=2836)
    integer k
    if (idum .eq. 0) then
       write(6,*) 'ran0: idum = 0'
       stop
    end if
    k=idum/IQ
    idum=IA*(idum-k*IQ)-IR*k
    if (idum.lt.0) idum=idum+IM
    ran0=2.0d0*(AM*idum-0.5d0)
    return
  end function ran0

  subroutine ran0v(idum,b,e,vec)
    implicit none
    integer idum,b,e,IA,IM,IQ,IR
    real(kind=8) vec(b:e),AM
    parameter (IA=16807,IM=2147483647,AM=1./IM,IQ=127773,IR=2836)
    integer k,j
    if (idum .eq. 0) then
       write(6,*) 'ran0: idum = 0'
       stop
    end if
    do j=b,e
       k=idum/IQ
       idum=IA*(idum-k*IQ)-IR*k
       if (idum.lt.0) idum=idum+IM
       vec(j)=2.0d0*(AM*idum-0.5d0)
    end do
    return
  end subroutine ran0v

  subroutine timer(t,i)
    implicit none
    real(kind=8):: t
    integer:: i
    integer:: c,r
    integer,save:: cs
    if (i .eq. 0) then
       call system_clock(c,r)
       cs = c
    else if (i .eq. 1) then
       call system_clock(c,r)
    else
       stop
    end if
    t = dble(c-cs)/dble(r)
    return
  end subroutine timer
end program testsarch


