!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
program testfft3d
  implicit none

  real(kind=8),parameter:: eps = 1.0d-10
  logical:: check, padding
  integer:: nx,ny,nz,niter,ierr

  read(5,*) check, padding
  write(6,*) 'sub dt[s] p[GFlops]'
  do while(.true.)
     read(5,*,IOSTAT=ierr) nx,ny,nz,niter
     if (ierr.ne.0) then
        exit
     end if
     if (niter.eq.1) then
        call testone(nx,ny,nz, padding, check,eps)
     else
        call testiter(nx,ny,nz, padding, niter)
     end if
  end do

  stop

contains

  subroutine testone(nx,ny,nz,padding,check,eps)
    !use fft_3d
    use pfft_3d
    implicit none
    integer,intent(in):: nx,ny,nz
    logical,intent(in):: padding,check
    real(kind=8),intent(in):: eps

    integer:: lx,ly,lz
    real(kind=8),allocatable,dimension(:,:,:,:):: dat,dat2,tmp
    type (fft3_struct):: fs
    integer:: i,j,k
    real(kind=8):: delta, nxyz

    if (padding) then
       lx = nx+1
       ly = ny+1
       lz = nz+1
    else
       lx = nx
       ly = ny
       lz = nz
    end if
    allocate(dat(lx,ly,lz,2),dat2(lx,ly,lz,2),tmp(lx,ly,lz,2))

    call fft3_init(nx,ny,nz,lx,ly,lz,fs)

    write(6,'(1x,a,3(1x,i3))') 'testing forward',nx,ny,nz

    call random_seed

    tmp = 0.0d0
    call prepdat(nx,ny,nz,lx,ly,lz,dat)
    dat2 = dat

    call timer(delta,0)
    call fft3_fw(fs,dat,tmp)
    call timer(delta,1)
    nxyz = nx*ny*nz
    write(6,'(1x,e12.5,1x,f9.4)') delta, &
         (6.0d0*nxyz*log(nxyz)/log(2.0d0)-6.0d0*nxyz)/delta*1.0d-9

    if (check) then
       call ft3(nx,ny,nz,lx,ly,lz,1,dat2,tmp)
       do i=1,nx
          do j=1,ny
             do k=1,nz
                if ((abs(dat(i,j,k,1)-dat2(i,j,k,1)) .gt. eps) .or. &
                     (abs(dat(i,j,k,2)-dat2(i,j,k,2)) .gt. eps)) then
                   write(6,*) i,j,k, &
                        dat(i,j,k,1)-dat2(i,j,k,1), &
                        dat(i,j,k,2)-dat2(i,j,k,2), dat(i,j,k,1), dat(i,j,k,2)
                end if
                !write(6,*) dat(i,j,k,1),dat(i,j,k,2)
             end do
          end do
       end do
    end if

    write(6,'(1x,a,3(1x,i3))') 'testing backward',nx,ny,nz

    tmp = 0.0d0
    call prepdat(nx,ny,nz,lx,ly,lz,dat)
    dat2 = dat

    call timer(delta,0)
    call fft3_bw(fs,dat,tmp)
    call timer(delta,1)
    nxyz = nx*ny*nz
    write(6,'(1x,e12.5,1x,f9.4)') delta, &
         (6.0d0*nxyz*log(nxyz)/log(2.0d0)-6.0d0*nxyz)/delta*1.0d-9

    if (check) then
       call ft3(nx,ny,nz,lx,ly,lz,-1,dat2,tmp)

       do i=1,nx
          do j=1,ny
             do k=1,nz
                if ((abs(dat(i,j,k,1)-dat2(i,j,k,1)) .gt. eps) .or. &
                     (abs(dat(i,j,k,2)-dat2(i,j,k,2)) .gt. eps)) then
                   write(6,*) i,j,k, &
                        dat(i,j,k,1)-dat2(i,j,k,1), &
                        dat(i,j,k,2)-dat2(i,j,k,2), dat(i,j,k,1), dat(i,j,k,2)
                end if
                !write(6,*) dat(i,j,k,1),dat(i,j,k,2)
             end do
          end do
       end do
    end if

    call fft3_done(fs)

    deallocate(dat,dat2,tmp)

    return

  end subroutine testone

  subroutine testiter(nx,ny,nz,padding,iter)
    !use fft_3d
    use pfft_3d
    implicit none
    integer,intent(in):: nx,ny,nz,iter
    logical,intent(in):: padding

    integer:: lx,ly,lz
    real(kind=8),allocatable,dimension(:,:,:,:):: dat,tmp
    type (fft3_struct):: fs
    integer:: i
    real(kind=8):: delta,nxyz

    if (padding) then
       lx = nx+1
       ly = ny+1
       lz = nz+1
    else
       lx = nx
       ly = ny
       lz = nz
    end if
    allocate(dat(lx,ly,lz,2),tmp(lx,ly,lz,2))

    call fft3_init(nx,ny,nz,lx,ly,lz,fs)

    write(6,'(1x,a,4(1x,i3))') 'testing forward and backward',nx,ny,nz,iter

    call random_seed

    tmp = 0.0d0
    call prepdat(nx,ny,nz,lx,ly,lz,dat)

    call timer(delta,0)
    do i=1,iter
       call fft3_fw(fs,dat,tmp)
       call fft3_bw(fs,dat,tmp)
    end do
    call timer(delta,1)
    nxyz = nx*ny*nz
    write(6,'(1x,e12.5,1x,f9.4)') delta, &
         (6.0d0*nxyz*log(nxyz)/log(2.0d0)-6.0d0*nxyz)*2.0d0*iter/delta*1.0d-9

    call fft3_done(fs)

    deallocate(dat,tmp)

    return

  end subroutine testiter

  subroutine timer(t,i)
    implicit none
    real(kind=8):: t
    integer:: i
    if (i .eq. 0) then
       call xclock(t,7)
    else if (i .eq. 1) then
       call xclock(t,8)
    else
       stop
    end if
    return
  end subroutine timer

  subroutine prepdat(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer:: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(lx,ly,lz,2)
    integer:: i

    call random_number(dat)
    dat = 2.0d0*dat - 1.0d0

    do i=nx+1,lx
       dat(i,:,:,:) = 0.0d0
    end do
    do i=ny+1,ly
       dat(:,i,:,:) = 0.0d0
    end do
    do i=nz+1,lz
       dat(:,:,i,:) = 0.0d0
    end do

    return
  end subroutine prepdat

  subroutine ft3(nx,ny,nz,lx,ly,lz,s,dat,tmp)
    implicit none
    integer:: nx,ny,nz,lx,ly,lz,s
    real(kind=8):: dat(lx,ly,lz,2),tmp(lx,ly,lz,2)
    integer:: i1,i2,j1,j2,k1,k2
    real(kind=8):: w1,w2,ph,sign
    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    if (abs(s).ne.1) then
       write(6,*) 'ft3: invalid s'
       stop
    end if

    if (s.eq.1) then
       sign = -1.0d0
    else if (s.eq.-1) then
       sign = 1.0d0
    end if

    tmp = 0.0d0

    do k2=0,nz-1
       do j2=0,ny-1
          do i2=0,nx-1
             do k1=0,nz-1
                do j1=0,ny-1
                   do i1=0,nx-1
                      ph = sign*2.0d0*pi*(dble(i1*i2)/nx+dble(j1*j2)/ny+ &
                           dble(k1*k2)/nz)
                      w1 = cos(ph)
                      w2 = sin(ph)
                      tmp(i2+1,j2+1,k2+1,1) = tmp(i2+1,j2+1,k2+1,1) &
                           + dat(i1+1,j1+1,k1+1,1)*w1 &
                           - dat(i1+1,j1+1,k1+1,2)*w2
                      tmp(i2+1,j2+1,k2+1,2) = tmp(i2+1,j2+1,k2+1,2) &
                           + dat(i1+1,j1+1,k1+1,1)*w2 &
                           + dat(i1+1,j1+1,k1+1,2)*w1
                   end do
                end do
             end do
          end do
       end do
    end do
    if (s.eq.1) then
       dat = tmp/(nx*ny*nz)
    else if (s.eq.-1) then
       dat = tmp
    end if
    return
  end subroutine ft3
end program testfft3d
