!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module xcgga
  private Gcor2, Ecpbe, Expbe, Hpbe, Apbe
contains
  subroutine Gcor2(np,vrs,a,alph1,bet1,bet2,bet3,bet4,gg,ggrs)
    implicit none
    integer:: np
    real(kind=8),intent(in):: vrs(np),a,alph1,bet1,bet2,bet3,bet4
    real(kind=8),intent(out):: gg(np),ggrs(np)
    real(kind=8):: rs,rtrs,q0,q1,q2,q1rs
    integer:: i
    !
    do i=1,np
       rs = vrs(i)
       rtrs = sqrt(rs)
       q0 = -2.0d0*a*(1.0d0 + alph1*rs)
       q1 = 2.0d0*a*(bet1*rtrs + bet2*rs + bet3*rs*rtrs + bet4*rs*rs)
       q2 = log(1.0d0 + 1.0d0/q1)
       gg(i) = q0*q2
       q1rs = a*(bet1/rtrs + 2.0d0*bet2 + 3.0d0*bet3*rtrs + 4.0d0*bet4*rs)
       ggrs(i) = -2.0d0*a*alph1*q2 - q0*q1rs/(q1*(1.0d0+q1))
    end do
    !
    return
  end subroutine Gcor2
  !
  subroutine Ecpbe(np,rs,zeta,ec,ecrs,eczet)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, zeta, ec, ecrs, eczet
    !
    real(kind=8):: fzz
    ! parameter (fzz = 8.0d0/(9.0d0*(-2.0d0+2.0d0**(4.0d0/3.0d0))))
    parameter (fzz = 1.70992093416137d0)
    real(kind=8),allocatable,dimension(:):: eu,eurs,ep,eprs,alf,alfrs
    real(kind=8),allocatable:: f(:),fz(:)
    !
    allocate(eu(np),eurs(np),ep(np),eprs(np),alf(np),alfrs(np))

    call Gcor2(np, rs, 0.031091d0, 0.21370d0, 7.5957d0, 3.5876d0 &
         , 1.6382d0, 0.49294d0, eu, eurs)
    call Gcor2(np, rs, 0.015545d0, 0.20548d0, 14.1189d0, 6.1977d0 &
         , 3.3662d0, 0.62517d0, ep, eprs)
    call Gcor2(np, rs, 0.016887d0, 0.11125d0, 10.357d0, 3.6231d0 &
         , 0.88026d0, 0.49671d0, alf, alfrs)

    allocate(f(np), fz(np))
    f = (-2.0d0 + (1.0d0-zeta)**(4.0d0/3.0d0) &
         + (1.0d0+zeta)**(4.0d0/3.0d0))/ &
         (-2.0d0 + 2.0d0**(4.0d0/3.0d0))

    alf = - alf
    ec = eu + (ep - eu)*f*zeta**4 + &
         alf*f*(1.0d0 - zeta**4)/fzz
    alfrs = - alfrs
    ecrs = eurs + (eprs - eurs)*f*zeta**4 &
         + alfrs*f*(1.0d0 - zeta**4)/fzz
    fz = (4.0d0/3.0d0)*((1.0d0+zeta)**(1.0d0/3.0d0) &
         - (1.0d0-zeta)**(1.0d0/3.0d0))/ &
         (-2.0d0 + 2.0d0**(4.0d0/3.0d0))
    eczet = 4.0d0*zeta**3*f*(ep-eu-alf/fzz) + fz*(zeta**4*(ep-eu) &
         + (1.0d0-zeta**4)*alf/fzz)
    deallocate(eu,ep,alf,eurs,eprs,alfrs,f,fz)
    !
    return
  end subroutine Ecpbe
  !
  subroutine Expbe(np, rs, s, ex, exrs, exs)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, s, ex, exrs, exs
    !
    real(kind=8):: q1,q3,kappa, mu
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(q1=(9.0d0*Pi/4.0d0)**(1.0d0/3.0d0))
    parameter(q1=1.91915829267751d0)
    ! parameter(q3=-3.0d0/(4.0d0*Pi))
    parameter(q3=-0.238732414637843d0)
    parameter(kappa = 0.8040d0, mu = 0.21951d0)
    !
    real(kind=8),dimension(:),allocatable:: fx
    !
    allocate(fx(np))
    fx = 1.0d0 + kappa - kappa/(1.0d0 + mu*s*s/kappa)
    ex = q3*(q1/rs)*fx
    exrs = -q3*q1*fx/rs**2
    exs = q3*(q1/rs)*2.0d0*mu*s/(1.0d0 + mu*s*s/kappa)**2
    deallocate(fx)
    return
  end subroutine Expbe

  subroutine Hpbe(np, vA, vt, vphi, vh, vha, vht, vhphi)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: vA, vt, vphi, vh, vha, vht, vhphi
    !
    real(kind=8):: gamma,beta,delta
    parameter(gamma=0.031091d0)
    parameter(beta=0.066725d0,delta=beta/gamma)
    !
    real(kind=8):: p1,p2,p3
    real(kind=8):: A, t, phi, h
    integer:: i
    !
    do i=1,np
       A = vA(i)
       t = vt(i)
       phi = vphi(i)
       p1 = t**2 + A*t**4
       p2 = 1.0d0 + A*t**2 + A**2*t**4
       h = gamma*phi**3*log(1.0d0 + delta*p1/p2)
       p3 = p2**2 + delta*p1*p2
       vha(i) = -beta*phi**3*t**6*a*(2.0d0+a*t**2)/p3
       vht(i) = beta*phi**3*2.0d0*t*(1.0d0+2.0d0*a*t**2)/p3
       vhphi(i) = 3.0d0*h/phi
       vh(i) = h
    end do
    return
  end subroutine Hpbe

  subroutine Apbe(np,ecl,phi,a,aecl,aphi)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: ecl,phi,a,aecl,aphi
    !
    real(kind=8):: gamma,beta,delta
    parameter(gamma=0.031091d0)
    parameter(beta=0.066725d0,delta=beta/gamma)
    !
    a = delta/(exp(-ecl/(gamma*phi**3)) - 1.0d0)
    aecl = a*(1.0d0+a/delta)/(gamma*phi**3)
    aphi = -a*(1.0d0+a/delta)*3.0d0*ecl/(gamma*phi**4)
    return
  end subroutine Apbe
  !
  subroutine exchcor(np, n, dn, exc, fxcn, fxcdn)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: exc, fxcn, fxcdn
    !
    real(kind=8):: Pi,neps,eta
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-26)
    parameter(eta=1.0d-24) ! stop phizeta from blowing up at |zeta| = 1
    !
    real(kind=8),dimension(:),allocatable:: rs,kf,s,zeta,ks,t
    real(kind=8),dimension(:),allocatable:: phi,phizeta
    real(kind=8),dimension(:),allocatable:: ex, exrs, exs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,aphi
    real(kind=8),dimension(:),allocatable:: h,ha,ht,hphi
    !
    write(6,*)'using PBE'
    allocate(rs(np),kf(np),s(np),zeta(np),ks(np),t(np), &
         phi(np), &
         ex(np),exrs(np),exs(np), &
         ecl(np),eclrs(np),eclzet(np))
    zeta = 0.0d0
    rs = (3.0d0/(4.0d0*Pi*max(n,neps)))**(1.0d0/3.0d0)
    kf = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs
    s = dn/(2.0d0*kf*max(n,neps))
    ks = sqrt(4.0d0*kf/Pi)
    phi = ((1.0d0+zeta)**(2.0d0/3.0d0) + (1.0d0-zeta)**(2.0d0/3.0d0))/2.0d0
    t = dn/(2.0d0*phi*ks*max(n,neps))
    !
    call Expbe(np,rs,s,ex,exrs,exs)
    call Ecpbe(np,rs,zeta,ecl,eclrs,eclzet)
    !
    allocate(a(np),aecl(np),aphi(np),h(np),ha(np),ht(np),hphi(np))
    !
    call Apbe(np,ecl,phi,a,aecl,aphi)
    call Hpbe(np,a,t,phi,h,ha,ht,hphi)
    !
    allocate(phizeta(np))
    phizeta = (((1+zeta)**2+eta)**(-1.0d0/6.0d0) &
         - ((1-zeta)**2 + eta)**(-1.0d0/6.0d0))/3.0d0
    !
    exc = ex + ecl + h
    fxcn = exc-(exrs + eclrs + ha*aecl*eclrs)/3.0d0*rs &
         - 4.0d0/3.0d0*s*exs - 7.0d0/6.0d0*t*ht
    fxcdn = exs/(2.0d0*kf) + ht/(2.0d0*phi*ks)
    !
    deallocate(rs,kf,s,zeta,ks,t,phi,phizeta,ex,exrs,exs, &
         ecl,eclrs,eclzet,a,aecl,aphi,h,ha,ht,hphi)
    return
  end subroutine exchcor

  subroutine exchcorsp(np, pn, pdn, dn, exc, pfxcn, pfxdn, fcdn)
    implicit none
    integer:: np
    real(kind=8),intent(in):: pn(np,2), pdn(np,2), dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8),intent(out):: pfxcn(np,2), pfxdn(np,2), fcdn(np)
    !
    real(kind=8):: Pi,neps,eta
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-26)
    parameter(eta=1.0d-24) ! stop phizeta from blowing up at |zeta| = 1
    !
    real(kind=8),dimension(:),allocatable:: n,rs,kf,s,zeta,ks,t
    real(kind=8),dimension(:),allocatable:: phi,phizeta
    real(kind=8),dimension(:,:),allocatable:: pex, pexrs, pexs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,aphi
    real(kind=8),dimension(:),allocatable:: h,ha,ht,hphi
    real(kind=8),dimension(:,:),allocatable:: qrs,qkf,qs
    integer:: i
    real(kind=8):: fcn,fczeton,pfxn(2)
    !
    write(6,*)'using PBE (LSDA)'
    allocate(n(np),rs(np),kf(np),s(np),zeta(np),ks(np),t(np), &
         phi(np), &
         pex(np,2),pexrs(np,2),pexs(np,2), &
         ecl(np),eclrs(np),eclzet(np), &
         qrs(np,2),qkf(np,2),qs(np,2))
    n = pn(:,1)+pn(:,2)
    zeta = (pn(:,1)-pn(:,2))/(max(n,neps))
    rs = (3.0d0/(4.0d0*Pi*max(n,neps)))**(1.0d0/3.0d0)
    kf = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs
    s = dn/(2.0d0*kf*max(n,neps))
    ks = sqrt(4.0d0*kf/Pi)
    phi = ((1.0d0+zeta)**(2.0d0/3.0d0) + (1.0d0-zeta)**(2.0d0/3.0d0))/2.0d0
    t = dn/(2.0d0*phi*ks*max(n,neps))
    !
    ! qn := 1/2*pn, qdn := 1/2*pdn
    ! qn, qdn -> qrs, qkf, qs
    !
    qrs = (3.0d0/(4.0d0*Pi*max(2.0d0*pn,neps)))**(1.0d0/3.0d0)
    qkf = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs
    qs = (2.0d0*pdn)/(2.0d0*qkf*max(2.0d0*pn,neps))

    call Expbe(np*2,qrs,qs,pex,pexrs,pexs)
    call Ecpbe(np,rs,zeta,ecl,eclrs,eclzet)
    !
    allocate(a(np),aecl(np),aphi(np),h(np),ha(np),ht(np),hphi(np))
    !
    call Apbe(np,ecl,phi,a,aecl,aphi)
    call Hpbe(np,a,t,phi,h,ha,ht,hphi)
    !
    allocate(phizeta(np))
    phizeta = (((1+zeta)**2+eta)**(-1.0d0/6.0d0) &
         - ((1-zeta)**2 + eta)**(-1.0d0/6.0d0))/3.0d0
    !
    exc = 0.0d0
    do i=1,np
       exc = exc + pn(i,1)*pex(i,1) + pn(i,2)*pex(i,2) &
            + n(i)*(ecl(i) + h(i))
       pfxn(1) = pex(i,1) - pexrs(i,1)/3.0d0*qrs(i,1) &
            - 4.0d0/3.0d0*qs(i,1)*pexs(i,1)
       pfxn(2) = pex(i,2) - pexrs(i,2)/3.0d0*qrs(i,2) &
            - 4.0d0/3.0d0*qs(i,2)*pexs(i,2)
       fcn = ecl(i) + h(i) &
            - (eclrs(i) + ha(i)*aecl(i)*eclrs(i))/3.0d0*rs(i) &
            - 7.0d0/6.0d0*t(i)*ht(i)
       fczeton = eclzet(i) + ha(i)*(aecl(i)*eclzet(i)+aphi(i)*phizeta(i)) &
         + (ht(i)*(-t(i)/phi(i)) + hphi(i))*phizeta(i)
       pfxcn(i,1) = pfxn(1) + fcn + fczeton*(1.0d0-zeta(i))
       pfxcn(i,2) = pfxn(2) + fcn - fczeton*(1.0d0+zeta(i))
    end do
    fcdn = ht/(2.0d0*phi*ks)
    pfxdn = pexs/(2.0d0*qkf)
    !
    deallocate(n,rs,kf,s,zeta,ks,t,phi,phizeta,pex,pexrs,pexs, &
         ecl,eclrs,eclzet,a,aecl,aphi,h,ha,ht,hphi,qrs,qkf,qs)
    return
  end subroutine exchcorsp
end module xcgga
